/*
 * Copyright (c) 2014 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <maps_coordinates.h>

#include <stdlib.h>

#include "mapzen_jsonparser.hpp"
#include "rapidjson/document.h"
#include "rapidjson/error/error.h"
#include "rapidjson/error/en.h"

extern "C" {
#include "mapzen_queue.h"
#include "mapzen_debug.h"
#include "mapzen_types.h"
#include "mapzen_util.h"
}

#include <string>
#include <vector>

#ifdef __cplusplus
extern "C" {
#endif

#define ROUTE_UNIT_CONVERSION_MILE_TO_M(x)	(1609.34 * (x))
#define ROUTE_UNIT_CONVERSION_MILE_TO_KM(x)	(1.60934 * (x))
#define ROUTE_UNIT_CONVERSION_MILE_TO_FT(x)	(5280 * (x))
#define ROUTE_UNIT_CONVERSION_MILE_TO_YD(x)	(1760 * (x))

static route_unit __route_unit = UNIT_KM;
static route_type __route_type = COSTING_AUTO;
constexpr double kPolylinePrecision = 1E6;
constexpr double kInvPolylinePrecision = 1.0 / kPolylinePrecision;


//TODO: port this to c? or port the whole project to c++ or add all the proper externs
void decode_shape(const std::string& encoded, std::vector<coords_s>& output) {
        // what byte are we looking at
        size_t i = 0;

        // Handy lambda to turn a few bytes of an encoded string into an integer
        auto deserialize = [&encoded, &i](const int previous) {
                // Grab each 5 bits and mask it in where it belongs using the shift
                int byte, shift = 0, result = 0;
                do {
                        byte = static_cast<int>(encoded[i++]) - 63;
                        result |= (byte & 0x1f) << shift;
                        shift += 5;
                } while (byte >= 0x20);

                // Undo the left shift from above or the bit flipping and add
                // to previous since its an offset
                return previous + (result & 1 ? ~(result >> 1) : (result >> 1));
        };

        // Iterate over all characters of the encoded shape string
        int last_lon = 0, last_lat = 0;
        while (i < encoded.length())
        {
                // Decode the coordinates, lat first
                int lat = deserialize(last_lat);
                int lon = deserialize(last_lon);

                // Shift the decimal point 5 places to the left and add to the output
                coords_s ll;
                ll.latitude  = static_cast<double>(lat) * kInvPolylinePrecision;
                ll.longitude = static_cast<double>(lon) * kInvPolylinePrecision;
                output.emplace_back(std::move(ll));

                // Remember the last one we encountered
                last_lat = lat;
                last_lon = lon;
        }
}


static mapzen_error_e __convert_status(int status)
{
	mapzen_error_e error = MAPZEN_ERROR_UNKNOWN;
	switch (status) {
	case 0:
		{
			/* Successful Geocode call */
			error = MAPZEN_ERROR_NONE;
			break;
		}
	case 400:
		{
			/* Error with input - Illegal argument from request */
			error = MAPZEN_ERROR_INVALID_PARAMETER;
			break;
		}
	case 403:
		{
			/* Key related error - Invalid key */
			error = MAPZEN_ERROR_KEY_NOT_AVAILABLE;
			break;
		}
	case 500:
		{
			/* Unknown error */
			error = MAPZEN_ERROR_UNKNOWN;
			break;
		}
	case -1:
		{
			/* Network error */
			error = MAPZEN_ERROR_NETWORK_UNREACHABLE;
			break;
		}
	}

	return error;
}

bool __get_string(const rapidjson::Value& obj, const char* key, char** value)
{
	//if we have the key and its not an empty value
	rapidjson::Value::ConstMemberIterator itr = obj.FindMember(key);
	if(itr != obj.MemberEnd() && itr->value.GetStringLength()){
		(*value) = (gchar *)g_malloc(itr->value.GetStringLength() + sizeof(rapidjson::Document::Ch));
		strncpy((*value), itr->value.GetString(), itr->value.GetStringLength());
		memset((*value) + itr->value.GetStringLength(), 0, sizeof(rapidjson::Document::Ch));
	//we had nothing
	} else {
		(*value) = NULL;
	}
	//is this something
	return *value != NULL;
}

/************ GEOCODE ***************/


static void __parse_geocode_response(char *response, int size, int *status, GList **coordsList)
{
	MAP_DEBUG("Inside __parse_geocode_response..");


	if (!response || !status || !coordsList) return;

	//so far we have nothing
	*coordsList = NULL;
	*status = 500;

	//crack open that json
	rapidjson::Document document;
	document.Parse(std::string(response, size).c_str());

	if (document.HasParseError()) {
		MAP_DEBUG("Error Parsing JSON Response: %s (%u)", GetParseError_En(document.GetParseError()), document.GetErrorOffset());
		return;
	}

	rapidjson::Value::ConstMemberIterator features = document.FindMember("features");
	if(features != document.MemberEnd() && features->value.IsArray()) {
		//for each feature
		for (rapidjson::Value::ConstValueIterator f = features->value.Begin(); f != features->value.End(); ++f) {
			//it has to have geometry
			rapidjson::Value::ConstMemberIterator geom = f->FindMember("geometry");
			if(geom != f->MemberEnd()) {
				//skip non-points
				rapidjson::Value::ConstMemberIterator type = geom->value.FindMember("type");
				if(type == geom->value.MemberEnd() || strcmp(type->value.GetString(), "Point"))
					continue;
				//get actual points
				rapidjson::Value::ConstMemberIterator coords = geom->value.FindMember("coordinates");
				if(coords == geom->value.MemberEnd() || !coords->value.IsArray() || coords->value.Size() != 2)
					continue;

				coords_s *coord = (coords_s *)g_malloc0(sizeof(coords_s));
				if (coord != NULL) {
					coord->longitude = coords->value[0].GetDouble();
					coord->latitude = coords->value[1].GetDouble();
				}

				if (*coordsList == NULL)
					*coordsList = g_list_append(*coordsList, coord);
				else
					*coordsList = g_list_insert_before(*coordsList, NULL, coord);
			}
		}
	}
	*status = 0;
}

/****************** REVERSE GEOCODE *********************/

static void __parse_revgeocode_response(char *response, int size, int *status, mapzen_address_resp_s **respAddr)
{
	if (!response || !status || !respAddr) return;

	//so far we have nothing
	*respAddr = NULL;
	*status = 500;

	//crack open that json
	rapidjson::Document document;
	document.Parse(std::string(response, size).c_str());

	if (document.HasParseError()) {
		MAP_DEBUG("Error Parsing JSON Response: %s (%u)", GetParseError_En(document.GetParseError()), document.GetErrorOffset());
		return;
	}

	rapidjson::Value::ConstMemberIterator features = document.FindMember("features");
	if(features != document.MemberEnd() && features->value.IsArray()) {
		//for each feature
		for (rapidjson::Value::ConstValueIterator f = features->value.Begin(); f != features->value.End(); ++f) {
			//it has to have geometry
			rapidjson::Value::ConstMemberIterator geom = f->FindMember("geometry");
			if(geom != f->MemberEnd()) {
				//skip non-points
				rapidjson::Value::ConstMemberIterator type = geom->value.FindMember("type");
				if(type == geom->value.MemberEnd() || strcmp(type->value.GetString(), "Point"))
					continue;
				//get actual points
				rapidjson::Value::ConstMemberIterator coords = geom->value.FindMember("coordinates");
				if(coords == geom->value.MemberEnd() || !coords->value.IsArray() || coords->value.Size() != 2)
					continue;

				//NOTE: it seems as though the tizen maps plugin api assumes that reverse geocode results will
				//actually be at the coordinates where you requested the reverse geocode. the mapzen api can return
				//POIS which are some distance away from the location and currently this is not accounted for :o(

				//get out the address information
				rapidjson::Value::ConstMemberIterator properties = f->FindMember("properties");
				if(properties != f->MemberEnd()) {
					//fill this out as we go
					*respAddr = (mapzen_address_resp_s *)g_malloc(sizeof(mapzen_address_resp_s));
					bool something = false;
					something = __get_string(properties->value, "street", &(*respAddr)->street) || something;
					something = __get_string(properties->value, "neighbourhood", &(*respAddr)->neighbourhood) || something;
					something = __get_string(properties->value, "housenumber", &(*respAddr)->housenumber) || something;
					something = __get_string(properties->value, "localadmin", &(*respAddr)->localadmin) || something;
					something = __get_string(properties->value, "county", &(*respAddr)->county) || something;
					something = __get_string(properties->value, "region", &(*respAddr)->region) || something;
					something = __get_string(properties->value, "country", &(*respAddr)->country) || something;
					something = __get_string(properties->value, "country_a", &(*respAddr)->country_a) || something;
					something = __get_string(properties->value, "postalcode", &(*respAddr)->postalcode) || something;
					//if everything was null thats pretty much unusable
					//TODO: these are pretty weak criteria..
					if(!something) {
						g_free(*respAddr);
						(*respAddr) = NULL;
					//forget the rest, we have one with something in it
					} else {
						break;
					}
				}
			}
		}
	}
	*status = 0;
}

/****************  PLACE SEARCH ********************/

static void __parse_place_response(char *response, int size, GList **placeList)
{
	MAP_DEBUG("Inside __parse_place_response.");

	if (!response || !placeList) return;

	*placeList = NULL;

	//crack open that json
	rapidjson::Document document;
	document.Parse(std::string(response, size).c_str());

	if (document.HasParseError()) {
		MAP_DEBUG("Error Parsing JSON Response: %s (%u)", GetParseError_En(document.GetParseError()), document.GetErrorOffset());
		return;
	}

	rapidjson::Value::ConstMemberIterator features = document.FindMember("features");
	if(features != document.MemberEnd() && features->value.IsArray()) {
		//for each feature
		for (rapidjson::Value::ConstValueIterator f = features->value.Begin(); f != features->value.End(); ++f) {
			//it has to have geometry
			rapidjson::Value::ConstMemberIterator geom = f->FindMember("geometry");
			if(geom != f->MemberEnd()) {
				//skip non-points
				rapidjson::Value::ConstMemberIterator type = geom->value.FindMember("type");
				if(type == geom->value.MemberEnd() || strcmp(type->value.GetString(), "Point"))
					continue;
				//get actual points
				rapidjson::Value::ConstMemberIterator coords = geom->value.FindMember("coordinates");
				if(coords == geom->value.MemberEnd() || !coords->value.IsArray() || coords->value.Size() != 2)
					continue;

				coords_s coordinate;
				coordinate.longitude = coords->value[0].GetDouble();
				coordinate.latitude = coords->value[1].GetDouble();

				//get out the address information
				rapidjson::Value::ConstMemberIterator properties = f->FindMember("properties");
				if(properties != f->MemberEnd()) {
					//fill this out as we go
					mapzen_place_resp_s *respPlaces = (mapzen_place_resp_s *)g_malloc0(sizeof(mapzen_place_resp_s));
					if (!respPlaces)
						return;

					respPlaces->place_id = NULL;
					respPlaces->categories = NULL;
					respPlaces->display_name = NULL;
					respPlaces->address = NULL;

					bool something = false;
					something = __get_string(properties->value, "gid", &respPlaces->place_id) || something;
					something = __get_string(properties->value, "name", &respPlaces->display_name) || something;

					respPlaces->address = (mapzen_address_resp_s *)g_malloc(sizeof(mapzen_address_resp_s));

					something = __get_string(properties->value, "housenumber", &respPlaces->address->housenumber) || something;
					something = __get_string(properties->value, "street", &respPlaces->address->street) || something;
					something = __get_string(properties->value, "neighbourhood", &respPlaces->address->neighbourhood) || something;
					something = __get_string(properties->value, "county", &respPlaces->address->county) || something;
					something = __get_string(properties->value, "region", &respPlaces->address->region) || something;
					something = __get_string(properties->value, "country", &respPlaces->address->country) || something;
					something = __get_string(properties->value, "country_a", &respPlaces->address->country_a) || something;
					something = __get_string(properties->value, "locality", &respPlaces->address->localadmin) || something;
					something = __get_string(properties->value, "postalcode", &respPlaces->address->postalcode) || something;

					respPlaces->coordinates = coordinate;

					respPlaces->categories = NULL;
					rapidjson::Value::ConstMemberIterator categories = properties->value.FindMember("category");
					if (categories != geom->value.MemberEnd() && categories->value.IsArray()) {
						for (rapidjson::SizeType i = 0; i < categories->value.Size(); i++) {
							if (respPlaces->categories == NULL)
								respPlaces->categories = g_list_append(respPlaces->categories, g_strdup(categories->value[i].GetString()));
							else
								respPlaces->categories = g_list_insert_before(respPlaces->categories, NULL, g_strdup(categories->value[i].GetString()));
						}
					}

					if(!something) {
						g_free(respPlaces);
						respPlaces = NULL;
					} else {
						if (*placeList == NULL)
							*placeList = g_list_append(*placeList, respPlaces);
						else
							*placeList = g_list_insert_before(*placeList, NULL, respPlaces);
					}
				}
			}
		}
	}
}

/********************* ROUTE RESPONSE ***********************/

static void __process_shape(std::vector<coords_s>& decoded_shape, GList **shapePoints)
{
	MAP_DEBUG(">>>>> START __process_shape");

	for (auto& shape_pt : decoded_shape) {
		coords_s *coords = (coords_s *)g_malloc0(sizeof(coords_s));
		if (coords) {
			coords->latitude = shape_pt.latitude;
			coords->longitude = shape_pt.longitude;
			if ((*shapePoints) == NULL)
				(*shapePoints) = g_list_append((*shapePoints), (gpointer)coords);
			else
				(*shapePoints) = g_list_insert_before((*shapePoints), NULL, (gpointer)coords);
	//		MAP_DEBUG(">>>>> PROCESS __process_shape: lat,lon=%f,%f", coords->latitude, coords->longitude);
		}
	}
	MAP_DEBUG(">>>>> END __process_shape");
}

static void __set_bounding_box(rapidjson::Value::ConstMemberIterator& summary, rectangle_s& bounding_box)
{
	rapidjson::Value::ConstMemberIterator min_lat = summary->value.FindMember("min_lat");
	rapidjson::Value::ConstMemberIterator min_lon = summary->value.FindMember("min_lon");
	rapidjson::Value::ConstMemberIterator max_lat = summary->value.FindMember("max_lat");
	rapidjson::Value::ConstMemberIterator max_lon = summary->value.FindMember("max_lon");
	if ((min_lat != summary->value.MemberEnd())
			&& (min_lon != summary->value.MemberEnd())
			&& (max_lat != summary->value.MemberEnd())
			&& (max_lon != summary->value.MemberEnd())) {
		bounding_box.top_left.latitude = max_lat->value.GetDouble();
		bounding_box.top_left.longitude = min_lon->value.GetDouble();
		bounding_box.bottom_right.latitude = min_lat->value.GetDouble();
		bounding_box.bottom_right.longitude = max_lon->value.GetDouble();
	}
}

static void __parse_maneuvers(rapidjson::Value::ConstMemberIterator maneuvers, mapzen_route_segment *segment_resp, std::vector<coords_s>& segment_shape)
{
	MAP_DEBUG(">>>>> START __parse_maneuvers");

	rapidjson::Value::ConstMemberIterator begin_shape_index, end_shape_index, maneuver_time,
		maneuver_distance, type, travel_type, travel_mode, street_names_resp;
	rapidjson::Value::ConstValueIterator street_name;

	for (rapidjson::Value::ConstValueIterator maneuver =
			maneuvers->value.Begin(); maneuver != maneuvers->value.End();
			++maneuver) {
		mapzen_route_maneuver *maneuver_resp = (mapzen_route_maneuver *)g_malloc(sizeof(mapzen_route_maneuver));


		// Set maneuver start point
		begin_shape_index = maneuver->FindMember("begin_shape_index");
		if (begin_shape_index != maneuver->MemberEnd()) {
			maneuver_resp->start_point = segment_shape.at(begin_shape_index->value.GetInt());
			MAP_DEBUG(">>>>> PROCESS __parse_maneuvers: maneuver_resp->start_point=%f,%f", maneuver_resp->start_point.latitude, maneuver_resp->start_point.longitude);
		}

		// Set maneuver end point
		end_shape_index = maneuver->FindMember("end_shape_index");
		if (end_shape_index != maneuver->MemberEnd()) {
			maneuver_resp->end_point = segment_shape.at(end_shape_index->value.GetInt());
			MAP_DEBUG(">>>>> PROCESS __parse_maneuvers: maneuver_resp->end_point=%f,%f", maneuver_resp->end_point.latitude, maneuver_resp->end_point.longitude);
		}

		// Set maneuver distance
		maneuver_distance = maneuver->FindMember("length");
		if (maneuver_distance != maneuver->MemberEnd()) {
			MAP_DEBUG(">>>>> __parse_maneuvers: PROCESS maneuver_distance: %f", maneuver_distance->value.GetDouble());
			maneuver_resp->distance = maneuver_distance->value.GetDouble();
		}

		// Set maneuver time
		maneuver_time = maneuver->FindMember("time");
		if (maneuver_time != maneuver->MemberEnd()) {
			MAP_DEBUG(">>>>> __parse_maneuvers: PROCESS maneuver_time: %d", maneuver_time->value.GetInt());
			maneuver_resp->time = maneuver_time->value.GetInt();
		}

		// Set maneuver type
		type = maneuver->FindMember("type");
		if (type != maneuver->MemberEnd()) {
			maneuver_resp->type = convert_int_to_mapzen_maneuver_type(type->value.GetInt());
			MAP_DEBUG(">>>>> PROCESS __parse_maneuvers: type=%d", maneuver_resp->type);
		}

		// Set maneuver instruction
		__get_string((*maneuver), "instruction", &maneuver_resp->instruction);

		// Set maneuver street names
		maneuver_resp->street_name = NULL;
		street_names_resp = maneuver->FindMember("street_names");
		if (street_names_resp != maneuver->MemberEnd() && street_names_resp->value.IsArray()) {
			std::string street_name;

			for (rapidjson::Value::ConstValueIterator street_name_resp = street_names_resp->value.Begin();
					street_name_resp != street_names_resp->value.End(); ++street_name_resp) {
				street_name += std::string(street_name.size()? "/": "");
				street_name += street_name_resp->GetString();
			}

			if (street_name.size() > 0) {
				MAP_DEBUG(">>>>> PROCESS __parse_maneuvers: street_name=%s", street_name.c_str());
				maneuver_resp->street_name = (gchar *)g_malloc0(street_name.size() + 1);
				if (maneuver_resp->street_name) {
					strncpy (maneuver_resp->street_name, street_name.c_str(), sizeof(gchar) * street_name.size());
					MAP_DEBUG(">>>>> PROCESS __parse_maneuvers: maneuver_resp->street_name=%s", maneuver_resp->street_name);
				}
			}
		}

		// Append maneuver to maneuver list
		if (segment_resp->maneuvers == NULL) {
			MAP_DEBUG(">>>>> __parse_route_response: PROCESS segment_resp->maneuvers == NULL) - append");
			segment_resp->maneuvers = g_list_append(segment_resp->maneuvers, (gpointer)maneuver_resp);
		} else {
			MAP_DEBUG(">>>>> __parse_route_response: PROCESS segment_resp->maneuvers != NULL) - insert before");
			segment_resp->maneuvers = g_list_insert_before(segment_resp->maneuvers, NULL, (gpointer)maneuver_resp);
		}
	} // end process each maneuver

	MAP_DEBUG(">>>>> END __parse_maneuvers");
}

static bool __determine_segment_origin_destination(const rapidjson::Value& locations,
		rapidjson::SizeType& location_index, coords_s& origin,
		coords_s& destination) {
	bool origin_set = false;
	bool destination_set = false;

	// Set origin
	for (rapidjson::SizeType i = location_index; i < locations.Size(); i++) {
		auto& location = locations[i];
		if (location["type"] == "break") {
			origin.latitude = location["lat"].GetDouble();
			origin.longitude = location["lon"].GetDouble();
			origin_set = true;
			// Increment location index to find segment destination
			location_index = ++i;
			break;
		}
	}

	// Set destination
	for (rapidjson::SizeType i = location_index; i < locations.Size(); i++) {
		auto& location = locations[i];
		if (location["type"] == "break") {
			destination.latitude = location["lat"].GetDouble();
			destination.longitude = location["lon"].GetDouble();
			destination_set = true;
			// Assign location index for next segment origin
			location_index = i;
			break;
		}
	}

	// Validate that both origin and destination are set
	if (origin_set and destination_set) {
		return true;
	}

	return false;
}

static void __parse_route_response(char *response, int size, int *status, mapzen_route_resp_s **routeResp) {
	MAP_DEBUG(">>>>> START __parse_route_response");

	if (!response || !status || !routeResp)
		return;
	*routeResp = NULL;
	MAP_DEBUG(">>>>> PROCESS __parse_route_response: ROUTE RESPONSE returned=%s", response);

	rapidjson::Document document;
	document.Parse(std::string(response, size).c_str());

	if (document.HasParseError()) {
		MAP_DEBUG("Error Parsing JSON Response: %s (%u)", GetParseError_En(document.GetParseError()), document.GetErrorOffset());
		return;
	}

	rapidjson::Value::MemberIterator trip = document.FindMember("trip");
	if (trip == document.MemberEnd()) {
		MAP_DEBUG(">>>>> PROCESS __parse_route_response: trip object not found in response");
		return;
	}

	// Process status
	rapidjson::Value::ConstMemberIterator trip_status = trip->value.FindMember("status");
	if (trip_status != trip->value.MemberEnd()) {
		*status = trip_status->value.GetInt();
		MAP_DEBUG(">>>>> PROCESS __parse_route_response: trip_status=%d", *status);
	}
	if (*status != 0) {
		MAP_DEBUG(">>>>> PROCESS __parse_route_response: invalid status in response");
		return;
	}

	// Create route response object and initialize lists to NULL
	*routeResp = (mapzen_route_resp_s *) g_malloc(sizeof(mapzen_route_resp_s));
	if (!(*routeResp))
		return;
	(*routeResp)->segments = NULL;
	(*routeResp)->shapePoints = NULL;

	// Get trip locations
	rapidjson::SizeType location_index = 0;
	rapidjson::Value& locations = trip->value["locations"];
	MAP_DEBUG(">>>>> PROCESS __parse_route_response: locations.Size(): %d", locations.Size());

	// Set trip origin
	rapidjson::Value& mapzen_origin = locations[0];
	(*routeResp)->origin.latitude = mapzen_origin["lat"].GetDouble();
	(*routeResp)->origin.longitude = mapzen_origin["lon"].GetDouble();
	MAP_DEBUG(">>>>> PROCESS __parse_route_response: trip origin: %f,%f", (*routeResp)->origin.latitude, (*routeResp)->origin.longitude);

	// Set trip destination
	rapidjson::Value& mapzen_destination = locations[locations.Size()-1];
	(*routeResp)->destination.latitude = mapzen_destination["lat"].GetDouble();
	(*routeResp)->destination.longitude = mapzen_destination["lon"].GetDouble();
	MAP_DEBUG(">>>>> PROCESS __parse_route_response: trip destination: %f,%f", (*routeResp)->destination.latitude, (*routeResp)->destination.longitude);

	// Set language
    __get_string(trip->value, "language", &(*routeResp)->language);

	// Set distance units
    // TODO maybe just read units from response and set value
	(*routeResp)->distance_unit = __route_unit;
	MAP_DEBUG(">>>>> PROCESS __parse_route_response: distance_unit=%d", (*routeResp)->distance_unit);

	// Process trip summary
	rapidjson::Value::ConstMemberIterator trip_summary = trip->value.FindMember("summary");
	if (trip_summary != trip->value.MemberEnd()) {
		// Set trip distance
		rapidjson::Value::ConstMemberIterator trip_distance = trip_summary->value.FindMember("length");
		if (trip_distance != trip_summary->value.MemberEnd()) {
			(*routeResp)->distance = trip_distance->value.GetDouble();
			MAP_DEBUG(">>>>> PROCESS __parse_route_response: trip length : %f", (*routeResp)->distance);
		}

		// Set trip time
		rapidjson::Value::ConstMemberIterator trip_time = trip_summary->value.FindMember("time");
		if (trip_time != trip_summary->value.MemberEnd()) {
			(*routeResp)->time = trip_time->value.GetInt();
			MAP_DEBUG(">>>>> PROCESS __parse_route_response: trip time: %d", trip_time->value.GetInt());
		}

		// Set trip bounding box
		__set_bounding_box(trip_summary, (*routeResp)->bounding_box);
		MAP_DEBUG(">>>>> PROCESS __parse_route_response: trip_bounding_box.top_left =%f,%f", (*routeResp)->bounding_box.top_left.latitude, (*routeResp)->bounding_box.top_left.longitude);
		MAP_DEBUG(">>>>> PROCESS __parse_route_response: trip_bounding_box.bottom_right =%f,%f", (*routeResp)->bounding_box.bottom_right.latitude, (*routeResp)->bounding_box.bottom_right.longitude);
	}

	// Type
	(*routeResp)->type = __route_type;
	MAP_DEBUG(">>>>> PROCESS __parse_route_response: route_type=%d", (*routeResp)->type);

	///////////////////////////////////////////////////////////////////////////
	// Process segments (legs)
	rapidjson::Value::MemberIterator legs = trip->value.FindMember("legs");
	if (legs == trip->value.MemberEnd() || !legs->value.IsArray()) {
		MAP_DEBUG(">>>>> PROCESS __parse_route_response: invalid legs in response");
		return;
	}
	for (rapidjson::Value::ConstValueIterator leg = legs->value.Begin(); leg != legs->value.End(); ++leg) {
		// Determine segment origin and destination
		coords_s segment_origin;
		coords_s segment_destination;
		if (!__determine_segment_origin_destination(locations, location_index, segment_origin, segment_destination)) {
			MAP_DEBUG(">>>>> PROCESS __parse_route_response: invalid segment origin and destination");
			return;
		}

		// Create segment response
		mapzen_route_segment *segment_resp = (mapzen_route_segment *)g_malloc(sizeof(mapzen_route_segment));
		segment_resp->maneuvers = NULL;
		segment_resp->shapePoints = NULL;

		// Set segment origin
		segment_resp->origin = segment_origin;
		MAP_DEBUG(">>>>> PROCESS __parse_route_response: segment_origin: %f,%f", segment_origin.latitude, segment_origin.longitude);

		// Set segment destination
		segment_resp->destination = segment_destination;
		MAP_DEBUG(">>>>> PROCESS __parse_route_response: segment_destination: %f,%f", segment_destination.latitude, segment_destination.longitude);

		// Process segment summary
		rapidjson::Value::ConstMemberIterator leg_summary = leg->FindMember("summary");
		if (leg_summary != leg->MemberEnd()) {
			// Set segment distance
			rapidjson::Value::ConstMemberIterator leg_distance = leg_summary->value.FindMember("length");
			if (leg_distance != leg_summary->value.MemberEnd()) {
				segment_resp->distance = leg_distance->value.GetDouble();
				MAP_DEBUG(">>>>> PROCESS __parse_route_response: leg length=%f", segment_resp->distance);
			}

			// Set segment time
			rapidjson::Value::ConstMemberIterator leg_time = leg_summary->value.FindMember("time");
			if (leg_time != leg_summary->value.MemberEnd()) {
				segment_resp->time = leg_time->value.GetInt();
				MAP_DEBUG(">>>>> PROCESS __parse_route_response: leg time=%d", segment_resp->time);
			}

			// Set segment bounding box
			__set_bounding_box(leg_summary, segment_resp->bounding_box);
			MAP_DEBUG(">>>>> PROCESS __parse_route_response: segment_bounding_box.top_left =%f,%f", segment_resp->bounding_box.top_left.latitude, segment_resp->bounding_box.top_left.longitude);
			MAP_DEBUG(">>>>> PROCESS __parse_route_response: segment_bounding_box.bottom_right =%f,%f", segment_resp->bounding_box.bottom_right.latitude, segment_resp->bounding_box.bottom_right.longitude);
		}

		// Get segment encoded shape and decode
		auto encoded_shape = leg->FindMember("shape");
		MAP_DEBUG(">>>>> PROCESS __parse_route_response: ENCODED SHAPE=%s", encoded_shape->value.GetString());
		std::vector<coords_s> decoded_shape;
		decode_shape(encoded_shape->value.GetString(), decoded_shape);
		MAP_DEBUG(">>>>> PROCESS __parse_route_response: Segment shape point count=%d", decoded_shape.size());

		// Append shape for route trip and segment
		// TODO - optimize/improve
		// TODO - duplicated points between route trip and segments
		// TODO - currently no generalization - test application has 500 point limit
		MAP_DEBUG(">>>>> PROCESS __parse_route_response: PRE __process_shape");
		__process_shape(decoded_shape, &((*routeResp)->shapePoints));
		__process_shape(decoded_shape, &(segment_resp->shapePoints));
		MAP_DEBUG(">>>>> PROCESS __parse_route_response: POST __process_shape");

		rapidjson::Value::ConstMemberIterator maneuvers = leg->FindMember("maneuvers");
		if (maneuvers == leg->MemberEnd() || !maneuvers->value.IsArray()) {
			g_free(segment_resp);
			return;
		}

		///////////////////////////////////////////////////////////////////////
		// Process each maneuver
		MAP_DEBUG(">>>>> PROCESS __parse_route_response: PRE __parse_maneuvers");
		__parse_maneuvers(maneuvers, segment_resp, decoded_shape);
		MAP_DEBUG(">>>>> PROCESS __parse_route_response: POST __parse_maneuvers");
		///////////////////////////////////////////////////////////////////////

		// Add segment to segment list
		if ((*routeResp)->segments == NULL) {
			MAP_DEBUG(">>>>> PROCESS __parse_route_response: (*routeResp)->maneuvers == NULL) - append");
			(*routeResp)->segments = g_list_append((*routeResp)->segments, (gpointer)segment_resp);
		} else {
			MAP_DEBUG(">>>>> PROCESS __parse_route_response: (*routeResp)->maneuvers != NULL) - insert before");
			(*routeResp)->segments = g_list_insert_before((*routeResp)->segments, NULL, (gpointer)segment_resp);
		}
	} // end process each leg
	///////////////////////////////////////////////////////////////////////////

	MAP_DEBUG(">>>>> END __parse_route_response");
}

void post_curl_response(char *response, int size, mapzen_resp_type type, void *user_data)
{
	MAP_DEBUG(">>>>> START post_curl_response");
	if (!response) return;

	if (!user_data) {
		MAP_DEBUG("Response data is NULL");
		return;
	}

	MAP_DEBUG("Response received from Curl. [Size=%d]", size);
	switch (type) {
	case RESP_TYPE_GEOCODE:
		{
			MAP_DEBUG("Inside Geocode JSON Parsing..");
			int status = -1;
			MapzenGeocodeQueryData *queryData = (MapzenGeocodeQueryData *)user_data;
			MapzenGeocodeResponseData *responseData = (MapzenGeocodeResponseData *)g_malloc(sizeof(MapzenGeocodeResponseData));

			if (responseData) {
				responseData->requestId = queryData->requestId;
				responseData->geocode_cb = queryData->geocode_cb;
				responseData->user_data = queryData->user_data;

				if (response && (size > 0)) {
					GList *coordsList = NULL;
					__parse_geocode_response(response, size, &status, &coordsList);

					if (coordsList != NULL) {
						/* Put the response in queue */
						responseData->error = __convert_status(status);
						responseData->coords = coordsList;
					} else {
						/* Response parsing failure */
						responseData->error = __convert_status(status);
						responseData->coords = NULL;
					}
				} else {
					responseData->error = __convert_status(status);
					responseData->coords = NULL;
				}

				mapzen_push_to_queue(type, (gpointer)responseData);
			}

			if (queryData) {
				g_free(queryData);
				queryData = NULL;
			}
			break;
		}
	case RESP_TYPE_REVGEOCODE:
		{
			MAP_DEBUG("Inside Rev Geocode JSON Parsing..");
			int status = -1;
			MapzenRevGeocodeQueryData *queryData = (MapzenRevGeocodeQueryData *)user_data;
			MapzenRevGeocodeResponseData *responseData = (MapzenRevGeocodeResponseData *)g_malloc(sizeof(MapzenRevGeocodeResponseData));

			if (responseData) {
				responseData->requestId = queryData->requestId;
				responseData->reverse_geocode_cb = queryData->reverse_geocode_cb;
				responseData->user_data = queryData->user_data;

				if (response && (size > 0)) {
					/* Coords Result GList */
					mapzen_address_resp_s *addrResponse = NULL;

					MAP_DEBUG("Rev Geocode :- Parsing json Response");
					__parse_revgeocode_response(response, size, &status, &addrResponse);

					if (addrResponse != NULL) {
						/* Put the response in queue */
						responseData->error = __convert_status(status);
						responseData->addressDetails = addrResponse;
					} else {
						/* REPSONSE PARSING FAILURE */
						MAP_DEBUG("addr Response is NULL");
						responseData->error = __convert_status(status);
						responseData->addressDetails = NULL;
					}
				} else {
					MAP_DEBUG("JSON Response is NULL..");
					responseData->error = __convert_status(status);
					responseData->addressDetails = NULL;
				}
				mapzen_push_to_queue(type, (gpointer)responseData);
			}

			if (queryData) {
				g_free(queryData);
				queryData = NULL;
			}
			break;
		}
	case RESP_TYPE_PLACES:
		{
			MAP_DEBUG("Inside Places JSON Parsing..");
			MapzenPlaceQueryData *queryData = (MapzenPlaceQueryData *)user_data;
			MapzenPlaceResponseData *responseData = (MapzenPlaceResponseData *)g_malloc(sizeof(MapzenPlaceResponseData));

			if (responseData) {
				responseData->requestId = queryData->requestId;
				responseData->place_search_cb = queryData->place_search_cb;
				responseData->user_data = queryData->user_data;

				if (response && (size > 0)) {
					/* Coords Result GList */
					GList *placeList = NULL;

					MAP_DEBUG("Search Places :- Parsing Json Response");
					__parse_place_response(response, size, &placeList);

					if (placeList != NULL) {
						/* Put the response in queue */
						responseData->error = MAPZEN_ERROR_NONE;
						responseData->places = placeList;
					} else {
						/* REPSONSE PARSING FAILURE */
						MAP_DEBUG("addr Response is NULL");
						responseData->error = MAPZEN_ERROR_UNKNOWN;
						responseData->places = NULL;
					}
				} else {
					responseData->error = MAPZEN_ERROR_UNKNOWN;
					responseData->places = NULL;
				}

				mapzen_push_to_queue(type, (gpointer)responseData);
			}

			if (queryData) {
				g_free(queryData);
				queryData = NULL;
			}

			break;
		}
	case RESP_TYPE_PLACES_DETAILS:
		{
			MAP_DEBUG("Inside Places Details JSON Parsing..");
			MapzenPlaceDetailsQueryData *queryData = (MapzenPlaceDetailsQueryData *)user_data;
			MapzenPlaceDetailsResponseData *responseData = (MapzenPlaceDetailsResponseData *)g_malloc(sizeof(MapzenPlaceDetailsResponseData));

			if (responseData) {
				responseData->requestId = queryData->requestId;
				responseData->get_place_details_cb = queryData->get_place_details_cb;
				responseData->user_data = queryData->user_data;

				if (response && (size > 0)) {
					/* Coords Result GList */
					GList *placeList = NULL;

					MAP_DEBUG("Search Places Details :- Parsing Json Response");
					__parse_place_response(response, size, &placeList);

					if (placeList != NULL) {
						/* Put the response in queue */
						responseData->error = MAPZEN_ERROR_NONE;
						responseData->places = placeList;
					} else {
						/* REPSONSE PARSING FAILURE */
						MAP_DEBUG("addr Response is NULL");
						responseData->error = MAPZEN_ERROR_UNKNOWN;
						responseData->places = NULL;
					}
				} else {
					responseData->error = MAPZEN_ERROR_UNKNOWN;
					responseData->places = NULL;
				}

				mapzen_push_to_queue(type, (gpointer)responseData);
			}

			if (queryData) {
				g_free(queryData);
				queryData = NULL;
			}

			break;
		}
	case RESP_TYPE_PLACES_LIST:
		{
			MAP_DEBUG("Inside Places List JSON Parsing..");
			MapzenPlaceListQueryData *queryData = (MapzenPlaceListQueryData *)user_data;
			MapzenPlaceListResponseData *responseData = (MapzenPlaceListResponseData *)g_malloc(sizeof(MapzenPlaceListResponseData));

			if (responseData) {
				responseData->requestId = queryData->requestId;
				responseData->place_list_search_cb = queryData->place_list_search_cb;
				responseData->user_data = queryData->user_data;

				if (response && (size > 0)) {
					/* Coords Result GList */
					GList *placeList = NULL;

					MAP_DEBUG("Search Places :- Parsing Json Response");
					__parse_place_response(response, size, &placeList);

					if (placeList != NULL) {
						/* Put the response in queue */
						responseData->error = MAPZEN_ERROR_NONE;
						responseData->places = placeList;
					} else {
						/* REPSONSE PARSING FAILURE */
						MAP_DEBUG("addr Response is NULL");
						responseData->error = MAPZEN_ERROR_UNKNOWN;
						responseData->places = NULL;
					}
				} else {
					responseData->error = MAPZEN_ERROR_UNKNOWN;
					responseData->places = NULL;
				}

				mapzen_push_to_queue(type, (gpointer)responseData);
			}

			if (queryData) {
				g_free(queryData);
				queryData = NULL;
			}

			break;
		}
	case RESP_TYPE_ROUTE:
	{
		MAP_DEBUG(">>>>> RESP_TYPE_ROUTE");
		int status = -1;
		MapzenRouteQueryData *queryData = (MapzenRouteQueryData *)user_data;
		MAP_DEBUG(">>>>> PROCESS post_curl_response set the following: __route_unit and __route_type");
		__route_unit = queryData->unit;
		__route_type = queryData->type;
		MAP_DEBUG(">>>>> PROCESS post_curl_response __route_unit=%d", __route_unit);
		MAP_DEBUG(">>>>> PROCESS post_curl_response __route_type=%d", __route_type);

		MAP_DEBUG(">>>>> PROCESS post_curl_response: prepare MapzenRouteResponseData");
		MapzenRouteResponseData *responseData = (MapzenRouteResponseData *)g_malloc(sizeof(MapzenRouteResponseData));

		if (responseData) {
			responseData->requestId = queryData->requestId;
			responseData->route_cb = queryData->route_cb;
			responseData->user_data = queryData->user_data;

			if (response && (size > 0)) {
				/* Coords Result GList */
				mapzen_route_resp_s *routeResponse = NULL;

				MAP_DEBUG(">>>>> PROCESS post_curl_response: PRE __parse_route_response");
				__parse_route_response(response, size, &status, &routeResponse);
				MAP_DEBUG(">>>>> PROCESS post_curl_response: POST __parse_route_response");

				if (routeResponse != NULL) {
					MAP_DEBUG(">>>>> PROCESS post_curl_response: (routeResponse != NULL)");
					MAP_DEBUG(">>>>> PROCESS post_curl_response: status=%d", status);
					MAP_DEBUG(">>>>> PROCESS post_curl_response: __convert_status(status)=%d", __convert_status(status));
					/* Put the response in queue */
					responseData->error = __convert_status(status);
					responseData->routeResponse = routeResponse;
				} else {
					/* REPSONSE PARSING FAILURE */
					MAP_DEBUG(">>>>> PROCESS post_curl_response: routeResponse is NULL");
					responseData->error = __convert_status(status);
					responseData->routeResponse = NULL;
				}
			} else {
				MAP_DEBUG(">>>>> PROCESS post_curl_response: JSON response is NULL");
				responseData->error = __convert_status(status);
				responseData->routeResponse = NULL;
			}

			MAP_DEBUG(">>>>> PROCESS post_curl_response: PRE mapzen_push_to_queue");
			mapzen_push_to_queue(type, (gpointer)responseData);
			MAP_DEBUG(">>>>> PROCESS post_curl_response: POST mapzen_push_to_queue");
		}

		if (queryData) {
			g_free(queryData);
			queryData = NULL;
		}

		break;
	}
	default:
		{
			MAP_DEBUG("Inside default JSON Parsing..");
			break;
		}
	}
	MAP_DEBUG(">>>>> END post_curl_response");
}

#ifdef __cplusplus
}
#endif
