/*
 * Copyright (c) 2014 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stdlib.h>
#include "mapzen_queue.h"
#include "mapzen_geocode.h"
#include "mapzen_revgeocode.h"
#include "mapzen_route.hpp"
#include "mapzen_place.h"
#include "mapzen_restcurl.h"
#include "mapzen_util.h"
#include "mapzen_debug.h"

pthread_mutex_t __requestLock;
static GAsyncQueue *responseQueue = NULL;
int __ResponseQueueLen = 0;
bool __response_timer_running = false;
guint __timeout_id = 0;

/* Request List */
GList *requestList = NULL;

/* Request Cancel List */
GList *reqCancelList = NULL;

static void __free_geocode_response(void *ptr);
static void __free_revgeocode_response(void *ptr);
static void __free_place_response(void *ptr);
static void __free_place_details_response(void *ptr);
static void __free_place_list_response(void *ptr);
static void __free_route_response(void *ptr);

int add_to_geocode_list(mapzen_geocode_req_s *req_details, mapzen_geocode_cb callback, int request_id, void *user_data)
{
	mapzen_request_s *req = (mapzen_request_s *)g_malloc0(sizeof(mapzen_request_s));

	if (req != NULL) {

		req->type = REQ_TYPE_GEOCODE;

		mapzen_geocode_req *data = (mapzen_geocode_req *)g_malloc0(sizeof(mapzen_geocode_req));

		if (data != NULL) {
			data->req_details = req_details;
			data->requestId = request_id;
			data->geocode_cb = callback;
			data->user_data = user_data;

			req->request = data;

			MAP_DEBUG("Added GEOCODE REQUEST to geocode List");

			pthread_mutex_lock(&__requestLock);
			if (requestList == NULL)
				requestList = g_list_append(requestList, (gpointer)req);
			else
				requestList = g_list_insert_before(requestList, NULL, (gpointer)req);
			pthread_mutex_unlock(&__requestLock);
		} else {
			g_free(req);
		}
	}

	return MAPZEN_ERROR_NONE;
}

int add_to_revgeocode_list(mapzen_revgeocode_req_s *req_details, mapzen_reverse_geocode_cb callback, int request_id, void *user_data)
{
	mapzen_request_s *req = (mapzen_request_s *)g_malloc0(sizeof(mapzen_request_s));

	if (req != NULL) {

		req->type = REQ_TYPE_REVGEOCODE;

		mapzen_revgeocode_req *data = (mapzen_revgeocode_req *)g_malloc0(sizeof(mapzen_revgeocode_req));

		if (data != NULL) {
			data->req_details = req_details;
			data->requestId = request_id;
			data->revgeocode_cb = callback;
			data->user_data = user_data;

			req->request = data;

			MAP_DEBUG("Added REVERSE GEOCODE request to geocode List");

			pthread_mutex_lock(&__requestLock);
			if (requestList == NULL)
				requestList = g_list_append(requestList, (gpointer)req);
			else
				requestList = g_list_insert_before(requestList, NULL, (gpointer)req);
			pthread_mutex_unlock(&__requestLock);
		} else {
			g_free(req);
		}
	}

	return MAPZEN_ERROR_NONE;
}

int add_to_place_list(mapzen_search_req_s *req_details, mapzen_place_search_cb callback, int request_id, void *user_data)
{
	mapzen_request_s *req = (mapzen_request_s *)g_malloc0(sizeof(mapzen_request_s));

	if (req != NULL) {

		req->type = REQ_TYPE_PLACES;

		mapzen_search_place_req *data = (mapzen_search_place_req *)g_malloc0(sizeof(mapzen_search_place_req));

		if (data != NULL) {
			data->req_details = req_details;
			data->requestId = request_id;
			data->search_place_cb = callback;
			data->user_data = user_data;

			req->request = data;

			MAP_DEBUG("Added PLACE request to places List");

			pthread_mutex_lock(&__requestLock);
			if (requestList == NULL)
				requestList = g_list_append(requestList, (gpointer)req);
			else
				requestList = g_list_insert_before(requestList, NULL, (gpointer)req);
			pthread_mutex_unlock(&__requestLock);
		} else {
			g_free(req);
		}
	}

	return MAPZEN_ERROR_NONE;
}

int add_to_place_details_list(mapzen_get_details_req_s *req_details, mapzen_get_place_details_cb callback, int request_id, void *user_data)
{
	mapzen_request_s *req = (mapzen_request_s *)g_malloc0(sizeof(mapzen_request_s));

	if (req != NULL) {

		req->type = REQ_TYPE_PLACES_DETAILS;

		mapzen_get_place_details_req *data = (mapzen_get_place_details_req *)g_malloc0(sizeof(mapzen_get_place_details_req));

		if (data != NULL) {
			data->req_details = req_details;
			data->requestId = request_id;
			data->get_place_details_cb = callback;
			data->user_data = user_data;

			req->request = data;

			MAP_DEBUG("Added PLACE Details request to places List");

			pthread_mutex_lock(&__requestLock);
			if (requestList == NULL)
				requestList = g_list_append(requestList, (gpointer)req);
			else
				requestList = g_list_insert_before(requestList, NULL, (gpointer)req);
			pthread_mutex_unlock(&__requestLock);
		} else {
			g_free(req);
		}
	}

	return MAPZEN_ERROR_NONE;
}

int add_to_places_list(mapzen_search_req_s *req_details, mapzen_place_list_search_cb callback, int request_id, void *user_data)
{

	MAP_DEBUG("add_to_place_list");


	mapzen_request_s *req = (mapzen_request_s *)g_malloc0(sizeof(mapzen_request_s));

	if (req != NULL) {

		req->type = REQ_TYPE_PLACES_LIST;

		mapzen_search_place_list_req *data = (mapzen_search_place_list_req *)g_malloc0(sizeof(mapzen_search_place_list_req));

		if (data != NULL) {
			data->req_details = req_details;
			data->requestId = request_id;
			data->search_place_list_cb = callback;
			data->user_data = user_data;

			req->request = data;

			MAP_DEBUG("Added PLACE request to places List");

			pthread_mutex_lock(&__requestLock);
			if (requestList == NULL)
				requestList = g_list_append(requestList, (gpointer)req);
			else
				requestList = g_list_insert_before(requestList, NULL, (gpointer)req);
			pthread_mutex_unlock(&__requestLock);
		} else {
			g_free(req);
		}
	}

	return MAPZEN_ERROR_NONE;
}

int add_to_route_list(mapzen_route_req_s *req_details, mapzen_route_cb callback, int request_id, void *user_data)
{

	MAP_DEBUG("add_to_place_list");


	mapzen_request_s *req = (mapzen_request_s *)g_malloc0(sizeof(mapzen_request_s));

	if (req != NULL) {

		req->type = REQ_TYPE_ROUTE;

		mapzen_route_req *data = (mapzen_route_req *)g_malloc0(sizeof(mapzen_route_req));

		if (data != NULL) {
			data->req_details = req_details;
			data->requestId = request_id;
			data->route_cb = callback;
			data->user_data = user_data;

			req->request = data;

			MAP_DEBUG("Added request to route List");

			pthread_mutex_lock(&__requestLock);
			if (requestList == NULL)
				requestList = g_list_append(requestList, (gpointer)req);
			else
				requestList = g_list_insert_before(requestList, NULL, (gpointer)req);
			pthread_mutex_unlock(&__requestLock);
		} else {
			g_free(req);
		}
	}

	return MAPZEN_ERROR_NONE;
}

static int __add_to_cancel_list(int request_id)
{
	if (reqCancelList == NULL)
		reqCancelList = g_list_append(reqCancelList, (gpointer) GINT_TO_POINTER(request_id));
	else
		reqCancelList = g_list_insert_before(reqCancelList, NULL, (gpointer) GINT_TO_POINTER(request_id));

	return MAPZEN_ERROR_NONE;
}

static int __remove_from_cancel_list(int request_id)
{
	if (g_list_length(reqCancelList) != 0)
		reqCancelList = g_list_remove(reqCancelList, (gpointer) GINT_TO_POINTER(request_id));

	return MAPZEN_ERROR_NONE;
}

int remove_from_request_list(int request_id)
{
	bool is_request_cancelled = false;
	GList *list = NULL;
	pthread_mutex_lock(&__requestLock);
	list = g_list_first(requestList);
	pthread_mutex_unlock(&__requestLock);
	mapzen_request_s *req = NULL;

	while (list != NULL) {
		req = (mapzen_request_s *) list->data;

		if (req != NULL) {
			if (req->type == REQ_TYPE_GEOCODE) {
				mapzen_geocode_req *geocode_req = (mapzen_geocode_req *) req->request;

				if (geocode_req->requestId == request_id) {
					/* Means the request is pending in queue. Not started yet */
					/* Deleting the request from the request queue */
					pthread_mutex_lock(&__requestLock);
					requestList = g_list_remove(requestList, (gpointer)req);
					pthread_mutex_unlock(&__requestLock);
					mapzen_geocode_req_s *geocode_req_details = (mapzen_geocode_req_s *)geocode_req->req_details;

					if (geocode_req_details)
						g_free(geocode_req_details);
					geocode_req_details = NULL;

					g_free(geocode_req);
					geocode_req = NULL;

					g_free(req);
					req = NULL;

					__remove_from_cancel_list(request_id);

					is_request_cancelled = true;
					break;
				}
			} else if (req->type == REQ_TYPE_REVGEOCODE) {
				mapzen_revgeocode_req *revgeocode_req = (mapzen_revgeocode_req *)req->request;

				if (revgeocode_req->requestId == request_id) {
					/* Means the request is pending in queue. Not started yet */
					/* Deleting the request from the request queue */
					pthread_mutex_lock(&__requestLock);
					requestList = g_list_remove(requestList, (gpointer)req);
					pthread_mutex_unlock(&__requestLock);
					mapzen_revgeocode_req_s *reverse_geocode_req_details = (mapzen_revgeocode_req_s *)revgeocode_req->req_details;

					if (reverse_geocode_req_details)
						g_free(reverse_geocode_req_details);
					reverse_geocode_req_details = NULL;

					g_free(revgeocode_req);
					revgeocode_req = NULL;

					g_free(req);
					req = NULL;

					__remove_from_cancel_list(request_id);

					is_request_cancelled = true;
					break;
				}
			} else if (req->type == REQ_TYPE_ROUTE) {
				mapzen_route_req *route_req = (mapzen_route_req *) req->request;

				if (route_req->requestId == request_id) {
					/* Means the request is pending in queue. Not started yet */
					/* Deleting the request from the request queue */
					pthread_mutex_lock(&__requestLock);
					requestList = g_list_remove(requestList, (gpointer)req);
					pthread_mutex_unlock(&__requestLock);
					mapzen_route_req_s *route_req_details = (mapzen_route_req_s *)route_req->req_details;

					if (route_req_details)
						g_free(route_req_details);
					route_req_details = NULL;

					g_free(route_req);
					route_req = NULL;

					g_free(req);
					req = NULL;

					__remove_from_cancel_list(request_id);

					is_request_cancelled = true;
					break;
				}
			} else if (req->type == REQ_TYPE_PLACES) {
				mapzen_search_place_req *places_req = (mapzen_search_place_req *) req->request;

				if (places_req->requestId == request_id) {
					/* Means the request is pending in queue. Not started yet */
					/* Deleting the request from the request queue */
					pthread_mutex_lock(&__requestLock);
					requestList = g_list_remove(requestList, (gpointer)req);
					pthread_mutex_unlock(&__requestLock);
					mapzen_search_req_s *place_req_details = (mapzen_search_req_s *)places_req->req_details;

					if (place_req_details)
						g_free(place_req_details);
					place_req_details = NULL;

					g_free(places_req);
					places_req = NULL;

					g_free(req);
					req = NULL;

					__remove_from_cancel_list(request_id);

					is_request_cancelled = true;
					break;
				}
			} else if (req->type == REQ_TYPE_PLACES_DETAILS) {
				mapzen_get_place_details_req *places_req = (mapzen_get_place_details_req *) req->request;

				if (places_req->requestId == request_id) {
					/* Means the request is pending in queue. Not started yet */
					/* Deleting the request from the request queue */
					pthread_mutex_lock(&__requestLock);
					requestList = g_list_remove(requestList, (gpointer)req);
					pthread_mutex_unlock(&__requestLock);
					mapzen_get_details_req_s *place_req_details = (mapzen_get_details_req_s *)places_req->req_details;

					if (place_req_details)
						g_free(place_req_details);
					place_req_details = NULL;

					g_free(places_req);
					places_req = NULL;

					g_free(req);
					req = NULL;

					__remove_from_cancel_list(request_id);

					is_request_cancelled = true;
					break;
				}
			} else if (req->type == REQ_TYPE_PLACES_LIST) {
				mapzen_search_place_list_req *places_req = (mapzen_search_place_list_req *) req->request;

				if (places_req->requestId == request_id) {
					/* Means the request is pending in queue. Not started yet */
					/* Deleting the request from the request queue */
					pthread_mutex_lock(&__requestLock);
					requestList = g_list_remove(requestList, (gpointer)req);
					pthread_mutex_unlock(&__requestLock);
					mapzen_search_req_s *place_req_details = (mapzen_search_req_s *)places_req->req_details;

					if (place_req_details)
						g_free(place_req_details);
					place_req_details = NULL;

					g_free(places_req);
					places_req = NULL;

					g_free(req);
					req = NULL;

					__remove_from_cancel_list(request_id);

					is_request_cancelled = true;
					break;
				}
			}
		}

		list = g_list_next(list);
	}

	if (!is_request_cancelled) {
		/* Means it is currently running. */
		/* Cancel the request */
		cancel_request(request_id);
		/* add to cancel list */
		__add_to_cancel_list(request_id);
	}

	return MAPZEN_ERROR_NONE;
}

static bool __is_request_in_cancel_list(int requestId)
{
	bool match_id = false;

	GList *list = NULL;
	list = g_list_first(reqCancelList);

	while (list != NULL) {

		int id = (int) GPOINTER_TO_INT(list->data);

		if (id == requestId) {
			match_id = true;
			break;
		}

		list = g_list_next(list);
	}

	return match_id;
}

mapzen_request_s *get_next_request()
{
	if (g_list_length(requestList) > 0) {
		pthread_mutex_lock(&__requestLock);
		GList *list = g_list_first(requestList);
		pthread_mutex_unlock(&__requestLock);

		if (list) {
			mapzen_request_s *req = (mapzen_request_s *) list->data;
			return req;
		} else {
			return NULL;
		}
	}

	return NULL;
}

static gboolean __timeout_cb(gpointer user_data)
{
	MAP_DEBUG("timeout_cb");
	__response_timer_running = true;
	g_async_queue_ref(responseQueue);

	__ResponseQueueLen = g_async_queue_length(responseQueue);
	MAP_DEBUG("Queue Len :: [%d]", __ResponseQueueLen);
	while (__ResponseQueueLen > 0) {
		gpointer data = g_async_queue_try_pop(responseQueue);
		if (!data)
			continue;

		MapzenQueueData *response = (MapzenQueueData *) data;

		switch (response->type) {
		case RESP_TYPE_GEOCODE:
			{
				MAP_DEBUG("Got geocode response..");
				MapzenGeocodeResponseData *geocodeData = (MapzenGeocodeResponseData *) (response->data);
				if (geocodeData) {
					if (__is_request_in_cancel_list(geocodeData->requestId) == false) {
						/* Deliver results to consumer */
						geocodeData->geocode_cb(geocodeData->error, geocodeData->requestId, geocodeData->coords, geocodeData->user_data);
					} else {
						/* If present in cancel list, dont trigger the response. */
						/* Remove the requestId from cancel list. */
						__remove_from_cancel_list(geocodeData->requestId);
					}
					__free_geocode_response(geocodeData);

					g_free(geocodeData);
					geocodeData = NULL;
				}

				/* Start the Next valid request */
				if (get_num_running_requests() < CURL_MAX_CONNECTS) {
					MAP_DEBUG("Fetching the next request from the request list....");
					mapzen_request_s *req = get_next_request();
					if (req) {
						pthread_mutex_lock(&__requestLock);
						requestList = g_list_remove(requestList, (gpointer)req);
						pthread_mutex_unlock(&__requestLock);

						if (req->type == REQ_TYPE_GEOCODE) {
							mapzen_geocode_req *geocode_req = (mapzen_geocode_req *) req->request;
							if (geocode_req != NULL) {
								start_geocode_service(geocode_req->req_details, geocode_req->geocode_cb, geocode_req->requestId, geocode_req->user_data);

								g_free(geocode_req);
								geocode_req = NULL;
							}
						} else if (req->type == REQ_TYPE_REVGEOCODE) {
							mapzen_revgeocode_req *revgeocode_req = (mapzen_revgeocode_req *) req->request;
							if (revgeocode_req != NULL) {
								start_reversegeocode_service(revgeocode_req->req_details, revgeocode_req->revgeocode_cb, revgeocode_req->requestId, revgeocode_req->user_data);

								g_free(revgeocode_req);
								revgeocode_req = NULL;
							}
						} else if (req->type == REQ_TYPE_PLACES) {
							mapzen_search_place_req *places_req = (mapzen_search_place_req *) req->request;
							if (places_req) {
								start_place_service(places_req->req_details, places_req->search_place_cb, places_req->requestId, places_req->user_data);

								g_free(places_req);
								places_req = NULL;
							}
						} else if (req->type == REQ_TYPE_PLACES_DETAILS) {
							mapzen_get_place_details_req *places_req = (mapzen_get_place_details_req *) req->request;
							if (places_req) {
								start_place_details_service(places_req->req_details, places_req->get_place_details_cb, places_req->requestId, places_req->user_data);

								g_free(places_req);
								places_req = NULL;
							}
						} else if (req->type == REQ_TYPE_PLACES_LIST) {
							mapzen_search_place_list_req *places_req = (mapzen_search_place_list_req *) req->request;
							if (places_req) {
								start_places_list_service(places_req->req_details, places_req->search_place_list_cb, places_req->requestId, places_req->user_data);

								g_free(places_req);
								places_req = NULL;
							}
						} else if (req->type == REQ_TYPE_ROUTE) {
							mapzen_route_req *route_req = (mapzen_route_req *) req->request;
							if (route_req) {
								start_route_service(route_req->req_details, route_req->route_cb, route_req->requestId, route_req->user_data);

								g_free(route_req);
								route_req = NULL;
							}
						}

						g_free(req);
						req = NULL;
					} else {
						MAP_DEBUG("No request in queue or request type is wrong");
					}
				} else {
					MAP_DEBUG("Libcurl request queue is FULL..");
				}
			}
			break;

		case RESP_TYPE_REVGEOCODE:
			{
				MapzenRevGeocodeResponseData *revGeocodeData = (MapzenRevGeocodeResponseData *) (response->data);
				if (revGeocodeData) {
					if (__is_request_in_cancel_list(revGeocodeData->requestId) == false) {
						/* Deliver results to consumer */
						revGeocodeData->reverse_geocode_cb(revGeocodeData->error, revGeocodeData->requestId, revGeocodeData->addressDetails, revGeocodeData->user_data);
					} else {
						/* If present in cancel list, dont trigger the response. */
						/* Remove the requestId from cancel list. */
						__remove_from_cancel_list(revGeocodeData->requestId);
					}
					__free_revgeocode_response(revGeocodeData);

					g_free(revGeocodeData);
					revGeocodeData = NULL;
				}

				/* Start the Next valid request */
				MAP_DEBUG(">> Condition Check <<");
				if (get_num_running_requests() < CURL_MAX_CONNECTS) {
					MAP_DEBUG("Fetching the next request from the request list....");
					mapzen_request_s *req = get_next_request();
					if (req) {
						pthread_mutex_lock(&__requestLock);
						requestList = g_list_remove(requestList, (gpointer)req);
						pthread_mutex_unlock(&__requestLock);

						if (req->type == REQ_TYPE_GEOCODE) {
							mapzen_geocode_req *geocode_req = (mapzen_geocode_req *) req->request;
							if (geocode_req != NULL) {
								start_geocode_service(geocode_req->req_details, geocode_req->geocode_cb, geocode_req->requestId, geocode_req->user_data);

								g_free(geocode_req);
								geocode_req = NULL;
							}
						} else if (req->type == REQ_TYPE_REVGEOCODE) {
							mapzen_revgeocode_req *revgeocode_req = (mapzen_revgeocode_req *) req->request;
							if (revgeocode_req != NULL) {
								start_reversegeocode_service(revgeocode_req->req_details, revgeocode_req->revgeocode_cb, revgeocode_req->requestId, revgeocode_req->user_data);

								g_free(revgeocode_req);
								revgeocode_req = NULL;
							}
						} else if (req->type == REQ_TYPE_PLACES) {
							mapzen_search_place_req *places_req = (mapzen_search_place_req *) req->request;
							if (places_req) {
								start_place_service(places_req->req_details, places_req->search_place_cb, places_req->requestId, places_req->user_data);

								g_free(places_req);
								places_req = NULL;
							}
						} else if (req->type == REQ_TYPE_PLACES_DETAILS) {
							mapzen_get_place_details_req *places_req = (mapzen_get_place_details_req *) req->request;
							if (places_req) {
								start_place_details_service(places_req->req_details, places_req->get_place_details_cb, places_req->requestId, places_req->user_data);

								g_free(places_req);
								places_req = NULL;
							}
						} else if (req->type == REQ_TYPE_PLACES_LIST) {
							mapzen_search_place_list_req *places_req = (mapzen_search_place_list_req *) req->request;
							if (places_req) {
								start_places_list_service(places_req->req_details, places_req->search_place_list_cb, places_req->requestId, places_req->user_data);

								g_free(places_req);
								places_req = NULL;
							}
						} else if (req->type == REQ_TYPE_ROUTE) {
							mapzen_route_req *route_req = (mapzen_route_req *) req->request;
							if (route_req) {
								start_route_service(route_req->req_details, route_req->route_cb, route_req->requestId, route_req->user_data);

								g_free(route_req);
								route_req = NULL;
							}
						}

						g_free(req);
						req = NULL;
					} else {
						MAP_DEBUG("No request in queue or request type is wrong");
					}
				} else {
					MAP_DEBUG("Libcurl request queue is FULL..");
				}
			}
			break;

		case RESP_TYPE_ROUTE:
			{
				MapzenRouteResponseData *routeData = (MapzenRouteResponseData *) (response->data);
				if (routeData != NULL) {
					if (__is_request_in_cancel_list(routeData->requestId) == false) {
						/* Deliver results to consumer */
						routeData->route_cb(routeData->error, routeData->requestId, routeData->routeResponse, routeData->user_data);
					} else {
						/* If present in cancel list, dont trigger the response. */
						/* Remove the requestId from cancel list. */
						__remove_from_cancel_list(routeData->requestId);
					}
					__free_route_response(routeData);

					g_free(routeData);
					routeData = NULL;
				}

				/* Start the Next valid request */
				if (get_num_running_requests() < CURL_MAX_CONNECTS) {
					MAP_DEBUG("Fetching the next request from the request list....");
					mapzen_request_s *req = get_next_request();
					if (req) {
						pthread_mutex_lock(&__requestLock);
						requestList = g_list_remove(requestList, (gpointer)req);
						pthread_mutex_unlock(&__requestLock);

						if (req->type == REQ_TYPE_GEOCODE) {
							mapzen_geocode_req *geocode_req = (mapzen_geocode_req *) req->request;
							if (geocode_req != NULL) {
								start_geocode_service(geocode_req->req_details, geocode_req->geocode_cb, geocode_req->requestId, geocode_req->user_data);

								g_free(geocode_req);
								geocode_req = NULL;
							}
						} else if (req->type == REQ_TYPE_REVGEOCODE) {
							mapzen_revgeocode_req *revgeocode_req = (mapzen_revgeocode_req *) req->request;
							if (revgeocode_req != NULL) {
								start_reversegeocode_service(revgeocode_req->req_details, revgeocode_req->revgeocode_cb, revgeocode_req->requestId, revgeocode_req->user_data);

								g_free(revgeocode_req);
								revgeocode_req = NULL;
							}
						} else if (req->type == REQ_TYPE_PLACES) {
							mapzen_search_place_req *places_req = (mapzen_search_place_req *) req->request;
							if (places_req) {
								start_place_service(places_req->req_details, places_req->search_place_cb, places_req->requestId, places_req->user_data);

								g_free(places_req);
								places_req = NULL;
							}
						} else if (req->type == REQ_TYPE_PLACES_DETAILS) {
							mapzen_get_place_details_req *places_req = (mapzen_get_place_details_req *) req->request;
							if (places_req) {
								start_place_details_service(places_req->req_details, places_req->get_place_details_cb, places_req->requestId, places_req->user_data);

								g_free(places_req);
								places_req = NULL;
							}
						} else if (req->type == REQ_TYPE_PLACES_LIST) {
							mapzen_search_place_list_req *places_req = (mapzen_search_place_list_req *) req->request;
							if (places_req) {
								start_places_list_service(places_req->req_details, places_req->search_place_list_cb, places_req->requestId, places_req->user_data);

								g_free(places_req);
								places_req = NULL;
							}
						} else if (req->type == REQ_TYPE_ROUTE) {
							mapzen_route_req *route_req = (mapzen_route_req *) req->request;
							if (route_req) {
								start_route_service(route_req->req_details, route_req->route_cb, route_req->requestId, route_req->user_data);

								g_free(route_req);
								route_req = NULL;
							}
						}

						g_free(req);
						req = NULL;
					} else {
						MAP_DEBUG("No request in queue or request type is wrong");
					}
				} else {
					MAP_DEBUG("Libcurl request queue is FULL..");
				}
			}
			break;

		case RESP_TYPE_PLACES:
			{
				MapzenPlaceResponseData *placesData = (MapzenPlaceResponseData *) (response->data);
				if (placesData != NULL) {
					MAP_DEBUG("Sending places result.");
					if (__is_request_in_cancel_list(placesData->requestId) == false) {
						/* Deliver results to consumer */
						placesData->place_search_cb(placesData->error, placesData->requestId, placesData->places, placesData->user_data);
					} else {
						/* If present in cancel list, dont trigger the response. */
						/* Remove the requestId from cancel list. */
						__remove_from_cancel_list(placesData->requestId);
					}
					__free_place_response(placesData);

					g_free(placesData);
					placesData = NULL;
				}

				/* Start the Next valid request */
				if (get_num_running_requests() < CURL_MAX_CONNECTS) {
					MAP_DEBUG("Fetching the next request from the request list....");
					mapzen_request_s *req = get_next_request();
					if (req) {
						pthread_mutex_lock(&__requestLock);
						requestList = g_list_remove(requestList, (gpointer)req);
						pthread_mutex_unlock(&__requestLock);

						if (req->type == REQ_TYPE_GEOCODE) {
							mapzen_geocode_req *geocode_req = (mapzen_geocode_req *) req->request;
							if (geocode_req != NULL) {
								start_geocode_service(geocode_req->req_details, geocode_req->geocode_cb, geocode_req->requestId, geocode_req->user_data);

								g_free(geocode_req);
								geocode_req = NULL;
							}
						} else if (req->type == REQ_TYPE_REVGEOCODE) {
							mapzen_revgeocode_req *revgeocode_req = (mapzen_revgeocode_req *) req->request;
							if (revgeocode_req != NULL) {
								start_reversegeocode_service(revgeocode_req->req_details, revgeocode_req->revgeocode_cb, revgeocode_req->requestId, revgeocode_req->user_data);

								g_free(revgeocode_req);
								revgeocode_req = NULL;
							}
						} else if (req->type == REQ_TYPE_PLACES) {
							mapzen_search_place_req *places_req = (mapzen_search_place_req *) req->request;
							if (places_req) {
								start_place_service(places_req->req_details, places_req->search_place_cb, places_req->requestId, places_req->user_data);

								g_free(places_req);
								places_req = NULL;
							}
						} else if (req->type == REQ_TYPE_PLACES_DETAILS) {
							mapzen_get_place_details_req *places_req = (mapzen_get_place_details_req *) req->request;
							if (places_req) {
								start_place_details_service(places_req->req_details, places_req->get_place_details_cb, places_req->requestId, places_req->user_data);

								g_free(places_req);
								places_req = NULL;
							}
						} else if (req->type == REQ_TYPE_PLACES_LIST) {
							mapzen_search_place_list_req *places_req = (mapzen_search_place_list_req *) req->request;
							if (places_req) {
								start_places_list_service(places_req->req_details, places_req->search_place_list_cb, places_req->requestId, places_req->user_data);

								g_free(places_req);
								places_req = NULL;
							}
						} else if (req->type == REQ_TYPE_ROUTE) {
							mapzen_route_req *route_req = (mapzen_route_req *) req->request;
							if (route_req) {
								start_route_service(route_req->req_details, route_req->route_cb, route_req->requestId, route_req->user_data);

								g_free(route_req);
								route_req = NULL;
							}
						}

						g_free(req);
						req = NULL;
					} else {
						MAP_DEBUG("No request in queue or request type is wrong");
					}
				} else {
					MAP_DEBUG("Libcurl request queue is FULL..");
				}
			}
			break;
		case RESP_TYPE_PLACES_DETAILS:
			{
				MapzenPlaceDetailsResponseData *placesData = (MapzenPlaceDetailsResponseData *) (response->data);
				if (placesData != NULL) {
					MAP_DEBUG("Sending places details result.");
					if (__is_request_in_cancel_list(placesData->requestId) == false) {
						/* Deliver results to consumer */
						placesData->get_place_details_cb(placesData->error, placesData->requestId, placesData->places, placesData->user_data);
					} else {
						/* If present in cancel list, dont trigger the response. */
						/* Remove the requestId from cancel list. */
						__remove_from_cancel_list(placesData->requestId);
					}
					__free_place_details_response(placesData);

					g_free(placesData);
					placesData = NULL;
				}

				/* Start the Next valid request */
				if (get_num_running_requests() < CURL_MAX_CONNECTS) {
					MAP_DEBUG("Fetching the next request from the request list....");
					mapzen_request_s *req = get_next_request();
					if (req) {
						pthread_mutex_lock(&__requestLock);
						requestList = g_list_remove(requestList, (gpointer)req);
						pthread_mutex_unlock(&__requestLock);

						if (req->type == REQ_TYPE_GEOCODE) {
							mapzen_geocode_req *geocode_req = (mapzen_geocode_req *) req->request;
							if (geocode_req != NULL) {
								start_geocode_service(geocode_req->req_details, geocode_req->geocode_cb, geocode_req->requestId, geocode_req->user_data);

								g_free(geocode_req);
								geocode_req = NULL;
							}
						} else if (req->type == REQ_TYPE_REVGEOCODE) {
							mapzen_revgeocode_req *revgeocode_req = (mapzen_revgeocode_req *) req->request;
							if (revgeocode_req != NULL) {
								start_reversegeocode_service(revgeocode_req->req_details, revgeocode_req->revgeocode_cb, revgeocode_req->requestId, revgeocode_req->user_data);

								g_free(revgeocode_req);
								revgeocode_req = NULL;
							}
						} else if (req->type == REQ_TYPE_PLACES) {
							mapzen_search_place_req *places_req = (mapzen_search_place_req *) req->request;
							if (places_req) {
								start_place_service(places_req->req_details, places_req->search_place_cb, places_req->requestId, places_req->user_data);

								g_free(places_req);
								places_req = NULL;
							}
						} else if (req->type == REQ_TYPE_PLACES_DETAILS) {
							mapzen_get_place_details_req *places_req = (mapzen_get_place_details_req *) req->request;
							if (places_req) {
								start_place_details_service(places_req->req_details, places_req->get_place_details_cb, places_req->requestId, places_req->user_data);

								g_free(places_req);
								places_req = NULL;
							}
						} else if (req->type == REQ_TYPE_PLACES_LIST) {
							mapzen_search_place_list_req *places_req = (mapzen_search_place_list_req *) req->request;
							if (places_req) {
								start_places_list_service(places_req->req_details, places_req->search_place_list_cb, places_req->requestId, places_req->user_data);

								g_free(places_req);
								places_req = NULL;
							}
						} else if (req->type == REQ_TYPE_ROUTE) {
							mapzen_route_req *route_req = (mapzen_route_req *) req->request;
							if (route_req) {
								start_route_service(route_req->req_details, route_req->route_cb, route_req->requestId, route_req->user_data);

								g_free(route_req);
								route_req = NULL;
							}
						}

						g_free(req);
						req = NULL;
					} else {
						MAP_DEBUG("No request in queue or request type is wrong");
					}
				} else {
					MAP_DEBUG("Libcurl request queue is FULL..");
				}
			}
			break;
		case RESP_TYPE_PLACES_LIST:
			{
				MapzenPlaceListResponseData *placesData = (MapzenPlaceListResponseData *) (response->data);
				if (placesData != NULL) {
					MAP_DEBUG("Sending places list result.");
					if (__is_request_in_cancel_list(placesData->requestId) == false) {
						/* Deliver results to consumer */
						placesData->place_list_search_cb(placesData->error, placesData->requestId, placesData->places, placesData->user_data);
					} else {
						/* If present in cancel list, dont trigger the response. */
						/* Remove the requestId from cancel list. */
						__remove_from_cancel_list(placesData->requestId);
					}
					__free_place_list_response(placesData);

					g_free(placesData);
					placesData = NULL;
				}

				/* Start the Next valid request */
				if (get_num_running_requests() < CURL_MAX_CONNECTS) {
					MAP_DEBUG("Fetching the next request from the request list....");
					mapzen_request_s *req = get_next_request();
					if (req) {
						pthread_mutex_lock(&__requestLock);
						requestList = g_list_remove(requestList, (gpointer)req);
						pthread_mutex_unlock(&__requestLock);

						if (req->type == REQ_TYPE_GEOCODE) {
							mapzen_geocode_req *geocode_req = (mapzen_geocode_req *) req->request;
							if (geocode_req != NULL) {
								start_geocode_service(geocode_req->req_details, geocode_req->geocode_cb, geocode_req->requestId, geocode_req->user_data);

								g_free(geocode_req);
								geocode_req = NULL;
							}
						} else if (req->type == REQ_TYPE_REVGEOCODE) {
							mapzen_revgeocode_req *revgeocode_req = (mapzen_revgeocode_req *) req->request;
							if (revgeocode_req != NULL) {
								start_reversegeocode_service(revgeocode_req->req_details, revgeocode_req->revgeocode_cb, revgeocode_req->requestId, revgeocode_req->user_data);

								g_free(revgeocode_req);
								revgeocode_req = NULL;
							}
						} else if (req->type == REQ_TYPE_PLACES) {
							mapzen_search_place_req *places_req = (mapzen_search_place_req *) req->request;
							if (places_req) {
								start_place_service(places_req->req_details, places_req->search_place_cb, places_req->requestId, places_req->user_data);

								g_free(places_req);
								places_req = NULL;
							}
						} else if (req->type == REQ_TYPE_PLACES_DETAILS) {
							mapzen_get_place_details_req *places_req = (mapzen_get_place_details_req *) req->request;
							if (places_req) {
								start_place_details_service(places_req->req_details, places_req->get_place_details_cb, places_req->requestId, places_req->user_data);

								g_free(places_req);
								places_req = NULL;
							}
						} else if (req->type == REQ_TYPE_PLACES_LIST) {
							mapzen_search_place_list_req *places_req = (mapzen_search_place_list_req *) req->request;
							if (places_req) {
								start_places_list_service(places_req->req_details, places_req->search_place_list_cb, places_req->requestId, places_req->user_data);

								g_free(places_req);
								places_req = NULL;
							}
						} else if (req->type == REQ_TYPE_ROUTE) {
							mapzen_route_req *route_req = (mapzen_route_req *) req->request;
							if (route_req) {
								start_route_service(route_req->req_details, route_req->route_cb, route_req->requestId, route_req->user_data);

								g_free(route_req);
								route_req = NULL;
							}
						}

						g_free(req);
						req = NULL;
					} else {
						MAP_DEBUG("No request in queue or request type is wrong");
					}
				} else {
					MAP_DEBUG("Libcurl request queue is FULL..");
				}
			}
			break;
		default:
			MAP_DEBUG("UNKNOWN RESPONSE TYPE");
			break;
		}

		__ResponseQueueLen = g_async_queue_length(responseQueue);

		if (response) {
			g_free(response);
			response = NULL;
		}
	};

	g_async_queue_unref(responseQueue);

	__response_timer_running = false;
	__timeout_id = 0;

	return false;
}

static void __free_geocode_response(void *ptr)
{
	MapzenGeocodeResponseData *geocodeData = (MapzenGeocodeResponseData *) (ptr);

	if (geocodeData) {
		GList *coords = NULL;
		coords = g_list_first(geocodeData->coords);

		while (coords) {
			coords_s *data = (coords_s *) coords->data;

			if (data) {
				geocodeData->coords = g_list_remove(geocodeData->coords, (gpointer)data);

				g_free(data);
				data = NULL;
			}
			coords = g_list_first(geocodeData->coords);
		}

		g_list_free(geocodeData->coords);
		geocodeData->coords = NULL;

		if (geocodeData->user_data) {
			g_free(geocodeData->user_data);
			geocodeData->user_data = NULL;
		}
	}
}

static void __free_revgeocode_response(void *ptr)
{
	MapzenRevGeocodeResponseData *revGeocodeData = (MapzenRevGeocodeResponseData *) (ptr);
	if (revGeocodeData) {
		if (revGeocodeData->addressDetails) {
			if (revGeocodeData->addressDetails->street) {
				g_free(revGeocodeData->addressDetails->street);
				revGeocodeData->addressDetails->street = NULL;
			}
			if (revGeocodeData->addressDetails->neighbourhood) {
				g_free(revGeocodeData->addressDetails->neighbourhood);
				revGeocodeData->addressDetails->neighbourhood = NULL;
			}
			if (revGeocodeData->addressDetails->localadmin) {
				g_free(revGeocodeData->addressDetails->localadmin);
				revGeocodeData->addressDetails->localadmin = NULL;
			}
			if (revGeocodeData->addressDetails->county) {
				g_free(revGeocodeData->addressDetails->county);
				revGeocodeData->addressDetails->county = NULL;
			}
			if (revGeocodeData->addressDetails->region) {
				g_free(revGeocodeData->addressDetails->region);
				revGeocodeData->addressDetails->region = NULL;
			}
			if (revGeocodeData->addressDetails->country) {
				g_free(revGeocodeData->addressDetails->country);
				revGeocodeData->addressDetails->country = NULL;
			}
			if (revGeocodeData->addressDetails->country_a) {
				g_free(revGeocodeData->addressDetails->country_a);
				revGeocodeData->addressDetails->country_a = NULL;
			}
			if (revGeocodeData->addressDetails->postalcode) {
				g_free(revGeocodeData->addressDetails->postalcode);
				revGeocodeData->addressDetails->postalcode = NULL;
			}

			g_free(revGeocodeData->addressDetails);
			revGeocodeData->addressDetails = NULL;
		}

		if (revGeocodeData->user_data) {
			g_free(revGeocodeData->user_data);
			revGeocodeData->user_data = NULL;
		}
	}
}

static void __free_place_response(void *ptr)
{
	MapzenPlaceResponseData *placeData = (MapzenPlaceResponseData *) (ptr);
	if (placeData) {
		GList *place = NULL;
		place = g_list_first(placeData->places);

		while (place) {
			mapzen_place_resp_s *mapzen_place = (mapzen_place_resp_s *) place->data;

			if (mapzen_place->address) {
					if (mapzen_place->address->street) {
						g_free(mapzen_place->address->street);
						mapzen_place->address->street = NULL;
					}
					if (mapzen_place->address->neighbourhood) {
						g_free(mapzen_place->address->neighbourhood);
						mapzen_place->address->neighbourhood = NULL;
					}
					if (mapzen_place->address->housenumber) {
						g_free(mapzen_place->address->housenumber);
						mapzen_place->address->housenumber = NULL;
					}
					if (mapzen_place->address->localadmin) {
						g_free(mapzen_place->address->localadmin);
						mapzen_place->address->localadmin = NULL;
					}
					if (mapzen_place->address->county) {
						g_free(mapzen_place->address->county);
						mapzen_place->address->county = NULL;
					}
					if (mapzen_place->address->region) {
						g_free(mapzen_place->address->region);
						mapzen_place->address->region = NULL;
					}
					if (mapzen_place->address->country) {
						g_free(mapzen_place->address->country);
						mapzen_place->address->country = NULL;
					}
					if (mapzen_place->address->country_a) {
						g_free(mapzen_place->address->country_a);
						mapzen_place->address->country_a = NULL;
					}
					if (mapzen_place->address->postalcode) {
						g_free(mapzen_place->address->postalcode);
						mapzen_place->address->postalcode = NULL;
					}
				g_free(mapzen_place->address);
				mapzen_place->address = NULL;
			}

			if (mapzen_place->categories) {
				if (g_list_length(mapzen_place->categories) > 0) {

					g_list_foreach(mapzen_place->categories, (GFunc)g_free, NULL);
					g_list_free(mapzen_place->categories);
					mapzen_place->categories = NULL;
				}
			}

			if (mapzen_place->place_id) {
				g_free(mapzen_place->place_id);
				mapzen_place->place_id = NULL;
			}

			if (mapzen_place->display_name) {
				g_free(mapzen_place->display_name);
				mapzen_place->display_name = NULL;
			}

			placeData->places = g_list_remove(placeData->places, (gpointer)mapzen_place);

			g_free(mapzen_place);
			mapzen_place = NULL;

			place = g_list_first(placeData->places);
		}
		g_list_free(placeData->places);
		placeData->places = NULL;

		if (placeData->user_data) {
			g_free(placeData->user_data);
			placeData->user_data = NULL;
		}
	}
}

static void __free_place_details_response(void *ptr)
{
	MapzenPlaceDetailsResponseData *placeData = (MapzenPlaceDetailsResponseData *) (ptr);
	if (placeData) {
		GList *place = NULL;
		place = g_list_first(placeData->places);

		while (place) {
			mapzen_place_resp_s *mapzen_place = (mapzen_place_resp_s *) place->data;

			if (mapzen_place->address) {
					if (mapzen_place->address->street) {
						g_free(mapzen_place->address->street);
						mapzen_place->address->street = NULL;
					}
					if (mapzen_place->address->neighbourhood) {
						g_free(mapzen_place->address->neighbourhood);
						mapzen_place->address->neighbourhood = NULL;
					}
					if (mapzen_place->address->housenumber) {
						g_free(mapzen_place->address->housenumber);
						mapzen_place->address->housenumber = NULL;
					}
					if (mapzen_place->address->localadmin) {
						g_free(mapzen_place->address->localadmin);
						mapzen_place->address->localadmin = NULL;
					}
					if (mapzen_place->address->county) {
						g_free(mapzen_place->address->county);
						mapzen_place->address->county = NULL;
					}
					if (mapzen_place->address->region) {
						g_free(mapzen_place->address->region);
						mapzen_place->address->region = NULL;
					}
					if (mapzen_place->address->country) {
						g_free(mapzen_place->address->country);
						mapzen_place->address->country = NULL;
					}
					if (mapzen_place->address->country_a) {
						g_free(mapzen_place->address->country_a);
						mapzen_place->address->country_a = NULL;
					}
					if (mapzen_place->address->postalcode) {
						g_free(mapzen_place->address->postalcode);
						mapzen_place->address->postalcode = NULL;
					}
				g_free(mapzen_place->address);
				mapzen_place->address = NULL;
			}

			if (mapzen_place->categories) {
				if (g_list_length(mapzen_place->categories) > 0) {

					g_list_foreach(mapzen_place->categories, (GFunc)g_free, NULL);
					g_list_free(mapzen_place->categories);
					mapzen_place->categories = NULL;
				}
			}

			if (mapzen_place->place_id) {
				g_free(mapzen_place->place_id);
				mapzen_place->place_id = NULL;
			}

			if (mapzen_place->display_name) {
				g_free(mapzen_place->display_name);
				mapzen_place->display_name = NULL;
			}

			placeData->places = g_list_remove(placeData->places, (gpointer)mapzen_place);

			g_free(mapzen_place);
			mapzen_place = NULL;

			place = g_list_first(placeData->places);
		}
		g_list_free(placeData->places);
		placeData->places = NULL;

		if (placeData->user_data) {
			g_free(placeData->user_data);
			placeData->user_data = NULL;
		}
	}
}

static void __free_place_list_response(void *ptr)
{
	MapzenPlaceListResponseData *placeData = (MapzenPlaceListResponseData *) (ptr);
	if (placeData) {
		GList *place = NULL;
		place = g_list_first(placeData->places);

		while (place) {
			mapzen_place_resp_s *mapzen_place = (mapzen_place_resp_s *) place->data;

			if (mapzen_place->address) {
					if (mapzen_place->address->street) {
						g_free(mapzen_place->address->street);
						mapzen_place->address->street = NULL;
					}
					if (mapzen_place->address->neighbourhood) {
						g_free(mapzen_place->address->neighbourhood);
						mapzen_place->address->neighbourhood = NULL;
					}
					if (mapzen_place->address->housenumber) {
						g_free(mapzen_place->address->housenumber);
						mapzen_place->address->housenumber = NULL;
					}
					if (mapzen_place->address->localadmin) {
						g_free(mapzen_place->address->localadmin);
						mapzen_place->address->localadmin = NULL;
					}
					if (mapzen_place->address->county) {
						g_free(mapzen_place->address->county);
						mapzen_place->address->county = NULL;
					}
					if (mapzen_place->address->region) {
						g_free(mapzen_place->address->region);
						mapzen_place->address->region = NULL;
					}
					if (mapzen_place->address->country) {
						g_free(mapzen_place->address->country);
						mapzen_place->address->country = NULL;
					}
					if (mapzen_place->address->country_a) {
						g_free(mapzen_place->address->country_a);
						mapzen_place->address->country_a = NULL;
					}
					if (mapzen_place->address->postalcode) {
						g_free(mapzen_place->address->postalcode);
						mapzen_place->address->postalcode = NULL;
					}
				g_free(mapzen_place->address);
				mapzen_place->address = NULL;
			}

			if (mapzen_place->categories) {
				if (g_list_length(mapzen_place->categories) > 0) {

					g_list_foreach(mapzen_place->categories, (GFunc)g_free, NULL);
					g_list_free(mapzen_place->categories);
					mapzen_place->categories = NULL;
				}
			}

			if (mapzen_place->place_id) {
				g_free(mapzen_place->place_id);
				mapzen_place->place_id = NULL;
			}

			if (mapzen_place->display_name) {
				g_free(mapzen_place->display_name);
				mapzen_place->display_name = NULL;
			}

			placeData->places = g_list_remove(placeData->places, (gpointer)mapzen_place);

			g_free(mapzen_place);
			mapzen_place = NULL;

			place = g_list_first(placeData->places);
		}
		g_list_free(placeData->places);
		placeData->places = NULL;

		if (placeData->user_data) {
			g_free(placeData->user_data);
			placeData->user_data = NULL;
		}
	}
}

static void __free_route_response(void *ptr)
{
	MAP_DEBUG(">>>>> START __free_route_response");

	MapzenRouteResponseData *routeData = (MapzenRouteResponseData *) (ptr);
	if (routeData) {
		 mapzen_route_resp_s *route_info = routeData->routeResponse;
		if (route_info) {
			GList *segment_data = NULL;
			segment_data = g_list_first(route_info->segments);
			while (segment_data) {
				mapzen_route_segment *segment = (mapzen_route_segment *) segment_data->data;

				GList *maneuver_data = NULL;
				maneuver_data = g_list_first(segment->maneuvers);
				while (maneuver_data) {
					mapzen_route_maneuver *maneuver = (mapzen_route_maneuver *) maneuver_data->data;
					if (maneuver->instruction) {
						g_free(maneuver->instruction);
						maneuver->instruction = NULL;
					}
					if (maneuver->street_name) {
						g_free(maneuver->street_name);
						maneuver->street_name = NULL;
					}
					segment->maneuvers = g_list_remove(segment->maneuvers, (gpointer)maneuver);
					g_free(maneuver);
					maneuver = NULL;
					/* Fetching the next item from Maneuver/Segment list */
					maneuver_data = g_list_first(segment->maneuvers);
				} /* Loop over maneuvers */

				g_list_free(segment->maneuvers);
				segment->maneuvers = NULL;

				GList *segmentShapePoints = NULL;
				segmentShapePoints = g_list_first(segment->shapePoints);
				while (segmentShapePoints) {
					coords_s *data = (coords_s *) segmentShapePoints->data;
					segment->shapePoints = g_list_remove(segment->shapePoints, (gpointer)data);
					g_free(data);
					data = NULL;
					segmentShapePoints = g_list_first(segment->shapePoints);
				}
				g_list_free(segment->shapePoints);
				segment->shapePoints = NULL;
				route_info->segments = g_list_remove(route_info->segments, (gpointer)segment);
				g_free(segment);
				segment = NULL;
				/* Fetching the next item from Maneuver/Segment list */
				segment_data = g_list_first(route_info->segments);

			} /* Loop over segments */

			g_list_free(route_info->segments);
			route_info->segments = NULL;

			GList *shapePoints = NULL;
			shapePoints = g_list_first(route_info->shapePoints);
			while (shapePoints) {
				coords_s *data = (coords_s *) shapePoints->data;
				route_info->shapePoints = g_list_remove(route_info->shapePoints, (gpointer)data);
				g_free(data);
				data = NULL;
				shapePoints = g_list_first(route_info->shapePoints);
			}
			g_list_free(route_info->shapePoints);
			route_info->shapePoints = NULL;
			if (route_info->language) {
				g_free(route_info->language);
				route_info->language = NULL;
			}
			g_free(route_info);
			route_info = NULL;
		}
		if (routeData->user_data) {
			g_free(routeData->user_data);
			routeData->user_data = NULL;
		}
	}
	MAP_DEBUG(">>>>> END __free_route_response");
}

int mapzen_init_queue()
{
	pthread_mutex_init(&__requestLock, NULL);

	/* Queue initialize */
	responseQueue = g_async_queue_new();

	/* Initialise Curl Service */
	int ret = init_curl();

	if (ret != MAPZEN_ERROR_NONE)
		return MAPZEN_ERROR_UNKNOWN;

	pthread_mutex_lock(&__requestLock);
	requestList = NULL;
	pthread_mutex_unlock(&__requestLock);
	reqCancelList = NULL;
	__ResponseQueueLen = 0;
	__response_timer_running = false;

	return MAPZEN_ERROR_NONE;
}

int mapzen_deinit_queue()
{
	deinit_curl();
	if (__timeout_id != 0) {
		g_source_remove (__timeout_id);
		__timeout_id = 0;
	}

	g_list_free(reqCancelList);
	reqCancelList = NULL;

	pthread_mutex_destroy(&__requestLock);

	/* Free the queue */
	if (responseQueue) {
		g_async_queue_ref(responseQueue);
		__ResponseQueueLen = g_async_queue_length(responseQueue);
		MAP_DEBUG("Queue len : %d", __ResponseQueueLen);

		while (__ResponseQueueLen > 0) {
			gpointer data = g_async_queue_try_pop(responseQueue);
			if (!data) {
				__ResponseQueueLen = g_async_queue_length(responseQueue);
				continue;
			}

			MapzenQueueData *response = (MapzenQueueData *) data;

			switch (response->type) {
			case RESP_TYPE_GEOCODE:
				{
					MapzenGeocodeResponseData *geocodeData = (MapzenGeocodeResponseData *) (response->data);
					if (geocodeData) {
						__free_geocode_response(geocodeData);

						g_free(geocodeData);
						geocodeData = NULL;
					}
				}
				break;
			case RESP_TYPE_REVGEOCODE:
				{
					MapzenRevGeocodeResponseData *revGeocodeData = (MapzenRevGeocodeResponseData *) (response->data);
					if (revGeocodeData) {
						__free_revgeocode_response(revGeocodeData);

						g_free(revGeocodeData);
						revGeocodeData = NULL;
					}
				}
				break;
			case RESP_TYPE_PLACES:
				{
					MapzenPlaceResponseData *placeData = (MapzenPlaceResponseData *) (response->data);
					if (placeData != NULL) {
						__free_place_response(placeData);

						g_free(placeData);
						placeData = NULL;
					}
				}
				break;
			case RESP_TYPE_PLACES_DETAILS:
				{
					MapzenPlaceDetailsResponseData *placeData = (MapzenPlaceDetailsResponseData *) (response->data);
					if (placeData != NULL) {
						__free_place_details_response(placeData);

						g_free(placeData);
						placeData = NULL;
					}
				}
				break;
			case RESP_TYPE_PLACES_LIST:
				{
					MapzenPlaceListResponseData *placeData = (MapzenPlaceListResponseData *) (response->data);
					if (placeData != NULL) {
						__free_place_list_response(placeData);

						g_free(placeData);
						placeData = NULL;
					}
				}
				break;
			case RESP_TYPE_ROUTE:
				{
					MapzenRouteResponseData *routeData = (MapzenRouteResponseData *) (response->data);
					if (routeData != NULL) {
						__free_route_response(routeData);

						g_free(routeData);
						routeData = NULL;
					}
				}
				break;
			default:
				break;
			}

			__ResponseQueueLen = g_async_queue_length(responseQueue);

			if (response) {
				g_free(response);
				response = NULL;
			}
		}
		g_async_queue_unref(responseQueue);
		responseQueue = NULL;
	}

	return MAPZEN_ERROR_NONE;
}

int mapzen_push_to_queue(mapzen_resp_type type, gpointer data)
{
	MAP_DEBUG("Pushing to Queue....");

	MapzenQueueData *queueData = (MapzenQueueData *)g_malloc0(sizeof(MapzenQueueData));
	if (!queueData) return MAPZEN_ERROR_OUT_OF_MEMORY;

	queueData->type = type;
	queueData->data = data;

	g_async_queue_ref(responseQueue);
	g_async_queue_push(responseQueue, (gpointer)queueData);
	g_async_queue_unref(responseQueue);

	if (!__response_timer_running) {
		MAP_DEBUG(">>>>>>>>>>>>>>>>>>>>>>> TRIGGER RESPONSE TIMER <<<<<<<<<<<<<<<<<<<<<<<<<");
		__timeout_id = g_timeout_add_seconds(1, __timeout_cb, NULL);	/* timeout : 1 sec */
	} else {
		__ResponseQueueLen = g_async_queue_length(responseQueue);
		MAP_DEBUG(">>>>>>>>>>>>>>>>>> Timer already running. Response Queue Len [%d] <<<<<<<<<<<<<<<<<", __ResponseQueueLen);
	}

	return MAPZEN_ERROR_NONE;
}

int start_geocode_service(mapzen_geocode_req_s *req_details, mapzen_geocode_cb callback, int request_id, void *user_data)
{
	int num_running_req = get_num_running_requests();

	if (num_running_req >= CURL_MAX_CONNECTS) {
		/* Add the request to queue */
		add_to_geocode_list(req_details, callback, request_id, user_data);
	} else {
		if (!req_details)
			return MAPZEN_ERROR_INVALID_PARAMETER;

		MapzenGeocodeQueryData *queryData = (MapzenGeocodeQueryData *)g_malloc0(sizeof(MapzenGeocodeQueryData));

		if (queryData != NULL) {
			queryData->requestId = request_id;
			queryData->geocode_cb = callback;
			queryData->user_data = user_data;

			if (req_details->boundary != NULL) {
				if (req_details->boundary->type == MAPZEN_BOUNDARY_RECT) {
					MAP_DEBUG("TOP LEFT >>><<< LATITUDE : %f, LONGITUDE : %f", req_details->boundary->rect.top_left.latitude, req_details->boundary->rect.top_left.longitude);
					MAP_DEBUG("BOTTOM RIGHT >>><<< LATITUDE : %f, LONGITUDE : %f", req_details->boundary->rect.bottom_right.latitude, req_details->boundary->rect.bottom_right.longitude);

					query_geocode_within_bounding_box(req_details->maps_key, req_details->address, req_details->boundary->rect.top_left, req_details->boundary->rect.bottom_right, req_details->num_res, queryData);

				} else if (req_details->boundary->type == MAPZEN_BOUNDARY_CIRCLE) {

					coords_s circle = req_details->boundary->circle.center;
					MAP_DEBUG("User input LATITUDE : %f, LONGITUDE : %f", circle.latitude, circle.longitude);
					query_geocode_within_circle(req_details->maps_key, req_details->address, circle, req_details->boundary->circle.radius, req_details->num_res, queryData);

				} else {
					query_geocode(req_details->maps_key, req_details->address, req_details->num_res, queryData);
				}

			} else {
				MAP_DEBUG("BOUNDARY is NULL.. NORMAL GEOCODE QUERY WITHOUT BOUNDS");

				query_geocode(req_details->maps_key, req_details->address, req_details->num_res, queryData);
			}
		}

		if (req_details->address) {
			g_free(req_details->address);
			req_details->address = NULL;
		}
		if (req_details->maps_key) {
			g_free(req_details->maps_key);
			req_details->maps_key = NULL;
		}
		if (req_details->boundary) {
			g_free(req_details->boundary);
			req_details->boundary = NULL;
		}
		g_free(req_details);
		req_details = NULL;
	}

	return MAPZEN_ERROR_NONE;
}

int start_reversegeocode_service(mapzen_revgeocode_req_s *req_details, mapzen_reverse_geocode_cb callback, int request_id, void *user_data)
{
	int num_running_req = get_num_running_requests();

	if (num_running_req >= CURL_MAX_CONNECTS) {
		/* Add the request to queue */
		add_to_revgeocode_list(req_details, callback, request_id, user_data);
	} else {
		if (!req_details)
			return MAPZEN_ERROR_INVALID_PARAMETER;

		MapzenRevGeocodeQueryData *queryData = (MapzenRevGeocodeQueryData *)g_malloc0(sizeof(MapzenRevGeocodeQueryData));

		if (queryData != NULL) {
			queryData->requestId = request_id;
			queryData->reverse_geocode_cb = callback;
			queryData->user_data = user_data;

			coords_s coords = req_details->coordinates;

			query_revgeocode(req_details->maps_key, coords.latitude, coords.longitude, queryData);
		}

		if (req_details->maps_key) {
			g_free(req_details->maps_key);
			req_details->maps_key = NULL;
		}
		g_free(req_details);
		req_details = NULL;
	}

	return MAPZEN_ERROR_NONE;
}

int start_route_service(mapzen_route_req_s *req_details, mapzen_route_cb callback, int request_id, void *user_data)
{
	int num_running_req = get_num_running_requests();

	if (num_running_req >= CURL_MAX_CONNECTS) {
		/* Add the request to queue */
		add_to_route_list(req_details, callback, request_id, user_data);
	} else {
		if (!req_details)
			return MAPZEN_ERROR_INVALID_PARAMETER;

		MapzenRouteQueryData *queryData = (MapzenRouteQueryData *)g_malloc0(sizeof(MapzenRouteQueryData));

		if (queryData != NULL) {
			queryData->requestId = request_id;
			queryData->route_cb = callback;
			queryData->origin = req_details->from;
			queryData->destination = req_details->to;
			queryData->unit = req_details->unit;
			queryData->type = req_details->type;
			queryData->user_data = user_data;

			query_route(req_details->maps_key, req_details->from, req_details->to, req_details->type, req_details->avoids, req_details->driving_style, req_details->way_points, queryData);
		}

		if (req_details->maps_key) {
			g_free(req_details->maps_key);
			req_details->maps_key = NULL;
		}
		if (req_details->way_points) {
			if (g_list_length(req_details->way_points) > 0) {
				GList *list = NULL;
				list = g_list_first(req_details->way_points);
				while (list) {
					coords_s *data = (coords_s *)list->data;
					if (data) {
						req_details->way_points = g_list_remove(req_details->way_points, (gpointer)data);

						g_free(data);
						data = NULL;
					}
					list = g_list_first(req_details->way_points);
				}
			}
			g_list_free(req_details->way_points);
			req_details->way_points = NULL;
		}
		g_free(req_details);
		req_details = NULL;
	}

	return MAPZEN_ERROR_NONE;
}

int start_place_service(mapzen_search_req_s *req_details, mapzen_place_search_cb callback, int request_id, void *user_data)
{
	MAP_DEBUG("Inside start_place_service.");

	int num_running_req = get_num_running_requests();

	if (num_running_req >= CURL_MAX_CONNECTS) {
		/* Add the request to queue */
		add_to_place_list(req_details, callback, request_id, user_data);
	} else {
		if (!req_details)
			return MAPZEN_ERROR_INVALID_PARAMETER;

		MapzenPlaceQueryData *queryData = (MapzenPlaceQueryData *)g_malloc0(sizeof(MapzenPlaceQueryData));

		if (queryData != NULL) {
			queryData->requestId = request_id;
			queryData->place_search_cb = callback;
			queryData->user_data = user_data;

			query_places(req_details->maps_key, req_details->search_string, req_details->categories, req_details->boundary, req_details->num_res, queryData, REQ_TYPE_PLACES);
		}

		if (req_details->maps_key) {
			g_free(req_details->maps_key);
			req_details->maps_key = NULL;
		}
		if (req_details->boundary) {
			g_free(req_details->boundary);
			req_details->boundary = NULL;
		}
		if (req_details->search_string) {
			g_free(req_details->search_string);
			req_details->search_string = NULL;
		}
		if (req_details->categories) {
			g_free(req_details->categories);
			req_details->categories = NULL;
		}
		g_free(req_details);
		req_details = NULL;
	}

	return MAPZEN_ERROR_NONE;
}

int start_place_details_service(mapzen_get_details_req_s *req_details, mapzen_get_place_details_cb callback, int request_id, void *user_data)
{
	MAP_DEBUG("Inside start_place_service.");

	int num_running_req = get_num_running_requests();

	if (num_running_req >= CURL_MAX_CONNECTS) {
		/* Add the request to queue */
		add_to_place_details_list(req_details, callback, request_id, user_data);
	} else {
		if (!req_details)
			return MAPZEN_ERROR_INVALID_PARAMETER;

		MapzenPlaceDetailsQueryData *queryData = (MapzenPlaceDetailsQueryData *)g_malloc0(sizeof(MapzenPlaceDetailsQueryData));

		if (queryData != NULL) {
			queryData->requestId = request_id;
			queryData->get_place_details_cb = callback;
			queryData->user_data = user_data;

			query_place_details(req_details->maps_key, req_details->uri, queryData);
		}

		if (req_details->maps_key) {
			g_free(req_details->maps_key);
			req_details->maps_key = NULL;
		}
		if (req_details->uri) {
			g_free(req_details->uri);
			req_details->uri = NULL;
		}
		g_free(req_details);
		req_details = NULL;
	}

	return MAPZEN_ERROR_NONE;
}

int start_places_list_service(mapzen_search_req_s *req_details, mapzen_place_list_search_cb callback, int request_id, void *user_data)
{
	MAP_DEBUG("Inside start_places_list_service.");

	int num_running_req = get_num_running_requests();

	if (num_running_req >= CURL_MAX_CONNECTS) {
		/* Add the request to queue */
		add_to_places_list(req_details, callback, request_id, user_data);
	} else {
		if (!req_details)
			return MAPZEN_ERROR_INVALID_PARAMETER;

		MapzenPlaceListQueryData *queryData = (MapzenPlaceListQueryData *)g_malloc0(sizeof(MapzenPlaceListQueryData));

		if (queryData != NULL) {
			queryData->requestId = request_id;
			queryData->place_list_search_cb = callback;
			queryData->user_data = user_data;

			query_places(req_details->maps_key, req_details->search_string, req_details->categories, req_details->boundary, req_details->num_res, queryData, REQ_TYPE_PLACES_LIST);
		}

		if (req_details->maps_key) {
			g_free(req_details->maps_key);
			req_details->maps_key = NULL;
		}
		if (req_details->boundary) {
			g_free(req_details->boundary);
			req_details->boundary = NULL;
		}
		if (req_details->search_string) {
			g_free(req_details->search_string);
			req_details->search_string = NULL;
		}
		if (req_details->categories) {
			g_free(req_details->categories);
			req_details->categories = NULL;
		}
		g_free(req_details);
		req_details = NULL;
	}

	return MAPZEN_ERROR_NONE;
}
