/*
 * Copyright (c) 2014 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdlib.h>
#include <curl/curl.h>
#include <curl/easy.h>
#include <pthread.h>
#include <network/net_connection.h>
#include "mapzen_restcurl.h"
#include "mapzen_types.h"
#include "mapzen_jsonparser.hpp"
#include "mapzen_debug.h"
#include "mapzen_util.h"

#define RESTCURL_USE_MULTI_CURL
#define TIZEN_CFG_CURL_TIMEOUT 120

CURLM *__cm_handle;
pthread_t __CurlThread;
pthread_mutex_t __MultiCurlArgsLock;
pthread_mutex_t __CmHandleLock;
static bool __thread_running_for_curlmulti = false;
static int __MultiCurlArgsRemain = 0;
pthread_cond_t __curl_delivered_cond;
pthread_mutex_t __curl_deliver_mutex;
static char *__proxy_address = NULL;
GList *__cancel_req_list = NULL;

typedef struct {
	char *memory;
	size_t size;
} MemoryStruct_s;

struct MultiCurlArg {
	MemoryStruct_s m_chunk;
	CURL *curl_handle;
	mapzen_req_type req_type;
	int request_id;
	void *user_data;
	char url[1024];
};

/****************** LIST for maintaining easy curl related data *******************/

struct MultiCurlArg_List {
	struct MultiCurlArg *mcArg;
	struct MultiCurlArg_List *next;
};

struct MultiCurlArg_List *head = NULL;
struct MultiCurlArg_List *curr = NULL;

static void __show_url_log(const char *url)
{
	char urlLog[1024] = {}, *tmp = NULL;
	int url_index = 0, urlLog_index = 0;
	const int bufferSize = sizeof(urlLog);

	if (!url) return;
	tmp = strstr(url, "api_key=");
	if (!tmp) {
		MAP_DEBUG("HTTP URL [%s]", url);
		return;
	}
	for (; url_index < strlen(url) && url_index < 1024; url_index++) {
		if (&url[url_index] == tmp)
			break;
		urlLog[url_index] = url[url_index];
	}
	STRCAT(urlLog, "api_key=");
	url_index = url_index + 8;
	urlLog_index = url_index;
	if ((url_index + 3 < strlen(url) && url[url_index] != '&') && (url[url_index] != 'n' || url[url_index+1] != 'u' || url[url_index+2] != 'l' || url[url_index+3] != 'l')) {
		STRCAT(urlLog, "xxxx-xxxx");
		for(;url_index < strlen(url); url_index++) {
			if (url[url_index] == '&')
				break;
		}
		urlLog_index = urlLog_index + 9;
	}
	for (; url_index < strlen(url) && url_index < bufferSize && urlLog_index < bufferSize; url_index++, urlLog_index++)
		urlLog[urlLog_index] = url[url_index];
	MAP_DEBUG("HTTP URL [%s]", urlLog);
}

static struct MultiCurlArg_List *__create_list(struct MultiCurlArg *mcArg)
{
	MAP_DEBUG("\n creating list with headnode as [%p]\n", mcArg);
	struct MultiCurlArg_List *ptr = (struct MultiCurlArg_List *)g_malloc0(sizeof(struct MultiCurlArg_List));
	if (NULL == ptr) {
		MAP_DEBUG("\n Node creation failed \n");
		return NULL;
	}
	ptr->mcArg = mcArg;
	ptr->next = NULL;

	head = curr = ptr;
	return ptr;
}

static struct MultiCurlArg_List *__add_to_list(struct MultiCurlArg *mcArg, bool add_to_end)
{
	if (NULL == head)
		return (__create_list(mcArg));

	struct MultiCurlArg_List *ptr = (struct MultiCurlArg_List *)g_malloc0(sizeof(struct MultiCurlArg_List));
	if (NULL == ptr) {
		MAP_DEBUG("\n Node creation failed \n");
		return NULL;
	}
	ptr->mcArg = mcArg;
	ptr->next = NULL;

	if (add_to_end) {
		curr->next = ptr;
		curr = ptr;
	} else {
		ptr->next = head;
		head = ptr;
	}
	return ptr;
}

static struct MultiCurlArg_List *__search_in_list(CURL * handle, struct MultiCurlArg_List **prev)
{
	struct MultiCurlArg_List *ptr = head;
	struct MultiCurlArg_List *tmp = NULL;
	bool found = false;

	MAP_DEBUG("\n Searching the list for value\n");

	while (ptr != NULL) {
		if (ptr->mcArg->curl_handle == handle) {
			found = true;
			break;
		} else {
			tmp = ptr;
			ptr = ptr->next;
		}
	}

	if (true == found) {
		if (prev)
			*prev = tmp;
		return ptr;
	} else {
		return NULL;
	}
}

static struct MultiCurlArg_List *__delete_from_list(CURL *handle)
{
	struct MultiCurlArg_List *prev = NULL;
	struct MultiCurlArg_List *del = NULL;

	MAP_DEBUG("\n Deleting value handle from list\n");

	del = __search_in_list(handle, &prev);
	if (del == NULL) {
		return NULL;
	} else {
		if (prev != NULL)
			prev->next = del->next;

		if (del == curr) {
			curr = prev;
			if (prev == NULL)
				head = NULL;
		} else if (del == head) {
			head = del->next;
		}
	}

	del->next = NULL;

	return del;
}

/*****************************************/

mapzen_error_e get_proxy_address(char **proxyAddress)
{
	*proxyAddress = NULL;

	mapzen_error_e err = MAPZEN_ERROR_NONE;

	connection_h con = NULL;
	int errorCode = CONNECTION_ERROR_NOT_SUPPORTED;

	errorCode = connection_create(&con);

	if (errorCode == CONNECTION_ERROR_NONE)
		errorCode = connection_get_proxy(con, CONNECTION_ADDRESS_FAMILY_IPV4, proxyAddress);

	if (errorCode != CONNECTION_ERROR_NONE) {
		err = MAPZEN_ERROR_UNKNOWN;
		switch (errorCode) {
		case CONNECTION_ERROR_INVALID_PARAMETER:
			MAP_DEBUG("Invalid parameter");
			err = MAPZEN_ERROR_INVALID_PARAMETER;
			break;
		case CONNECTION_ERROR_OUT_OF_MEMORY:
			MAP_DEBUG("Out of memory error");
			err = MAPZEN_ERROR_OUT_OF_MEMORY;
			break;
		case CONNECTION_ERROR_INVALID_OPERATION:
			MAP_DEBUG("Invalid Operation");
			err = MAPZEN_ERROR_INVALID_OPERATION;
			break;
		case CONNECTION_ERROR_ADDRESS_FAMILY_NOT_SUPPORTED:
			MAP_DEBUG("Address family not supported");
			err = MAPZEN_ERROR_NOT_SUPPORTED;
			break;
		/*
		case CONNECTION_ERROR_PERMISSION_DENIED:
			MAP_DEBUG("Permission denied");
			err = MAPZEN_ERROR_PERMISSION_DENIED;
			break;
		*/
		case CONNECTION_ERROR_OPERATION_FAILED:
			MAP_DEBUG("Operation failed");
			err = MAPZEN_ERROR_INVALID_OPERATION;
			break;
		case CONNECTION_ERROR_ITERATOR_END:
			MAP_DEBUG("End of iteration");
			break;
		case CONNECTION_ERROR_NO_CONNECTION:
			MAP_DEBUG("There is no connection");
			err = MAPZEN_ERROR_NETWORK_UNREACHABLE;
			break;
		case CONNECTION_ERROR_NOW_IN_PROGRESS:
			MAP_DEBUG("Now in progress");
			err = MAPZEN_ERROR_RESOURCE_BUSY;
			break;
		case CONNECTION_ERROR_ALREADY_EXISTS:
			MAP_DEBUG("Already exists");
			break;
		case CONNECTION_ERROR_OPERATION_ABORTED:
			MAP_DEBUG("Operation is aborted");
			err = MAPZEN_ERROR_CANCELED;
			break;
		case CONNECTION_ERROR_DHCP_FAILED:
			MAP_DEBUG("DHCP failed");
			break;
		case CONNECTION_ERROR_INVALID_KEY:
			MAP_DEBUG("Invalid key");
			err = MAPZEN_ERROR_KEY_NOT_AVAILABLE;
			break;
		case CONNECTION_ERROR_NO_REPLY:
			MAP_DEBUG("No reply");
			err = MAPZEN_ERROR_RESOURCE_BUSY;
			break;
		case CONNECTION_ERROR_NOT_SUPPORTED:
			MAP_DEBUG("Not Supported");
			err = MAPZEN_ERROR_NOT_SUPPORTED;
			break;
		default:
			MAP_DEBUG("Unknown");
			break;
		}
		MAP_DEBUG("errorCode = %ld", (long)errorCode);
	}

	if (con)
		connection_destroy(con);

	if (err != MAPZEN_ERROR_NONE)
		return err;

	MAP_DEBUG("Proxy = %s", (*proxyAddress ? *proxyAddress : "(null)"));

	return MAPZEN_ERROR_NONE;
}

static size_t __write_memory_callback(void *contents, size_t size, size_t nmemb, void *userp)
{
	size_t realsize = size * nmemb;
#ifdef RESTCURL_USE_MULTI_CURL
	MemoryStruct_s *mem = &(((struct MultiCurlArg *) userp)->m_chunk);
#else
	MemoryStruct_s *mem = (MemoryStruct_s *) userp;
#endif

	if (mem->memory == NULL)
		mem->memory = g_malloc(realsize);
	else
		mem->memory = g_realloc(mem->memory, mem->size + realsize);

	if (mem->memory == NULL)
		return 0;

	memcpy(mem->memory + mem->size, contents, realsize);

	mem->size += realsize;

	MAP_DEBUG("Real size : %d", realsize);
	MAP_DEBUG("Response size :: %d", strlen(mem->memory));

	return realsize;
}

static int __xferinfo(void *p, curl_off_t dltotal, curl_off_t dlnow, curl_off_t ultotal, curl_off_t ulnow)
{
	struct MultiCurlArg *pMultiCurlArg = (struct MultiCurlArg *) p;

	GList *list = NULL;
	list = g_list_first(__cancel_req_list);
	int req_id_to_be_cancelled = -1;
	bool cancel_request = false;

	while (list) {
		req_id_to_be_cancelled = (int) GPOINTER_TO_INT(list->data);
		MAP_DEBUG("Req id in the list waiting to get cancelled [%d]", req_id_to_be_cancelled);
		if (pMultiCurlArg->request_id == req_id_to_be_cancelled) {
			MAP_DEBUG(">>>>>>>>>>>> Req id matched :: To be cancelled [%d]", req_id_to_be_cancelled);
			cancel_request = true;
			break;
		}
		list = list->next;
	}

	if (cancel_request) {
		MAP_DEBUG(">>>>>>>> Req Id to be aborted :: [%d] <<<<<<<<<<<<<", req_id_to_be_cancelled);
		/* Remove the request id from cancel list */
		__cancel_req_list = g_list_remove(__cancel_req_list, (gpointer) GINT_TO_POINTER(req_id_to_be_cancelled));
		MAP_DEBUG(">>>>>>>>>> Request removed from cancel list <<<<<<<<");
		return 1;		/* Returning Non-Zero value will abort the transfer */
	}

	return 0;
}

static int __progress_info(void *p, double dltotal, double dlnow, double ultotal, double ulnow)
{
	return __xferinfo(p, (curl_off_t)dltotal, (curl_off_t)dlnow, (curl_off_t)ultotal, (curl_off_t)ulnow);
}

int add_handle(const char *url, mapzen_req_type type, void *user_data)
{

	if (!url || !user_data) return -1;

	int req_id = -1;

	switch (type) {
	case REQ_TYPE_GEOCODE:
		req_id = ((MapzenGeocodeQueryData *) user_data)->requestId;
		break;
	case REQ_TYPE_REVGEOCODE:
		req_id = ((MapzenRevGeocodeQueryData *) user_data)->requestId;
		break;
	case REQ_TYPE_PLACES:
		req_id = ((MapzenPlaceQueryData *) user_data)->requestId;
		break;
	case REQ_TYPE_PLACES_LIST:
		req_id = ((MapzenPlaceListQueryData *) user_data)->requestId;
		break;
	case REQ_TYPE_PLACES_DETAILS:
		req_id = ((MapzenPlaceDetailsQueryData *) user_data)->requestId;
		break;
	case REQ_TYPE_ROUTE:
		req_id = ((MapzenRouteQueryData *) user_data)->requestId;
		break;
	default:
		break;
	}

#ifdef RESTCURL_USE_MULTI_CURL
	// block to hide key in log
	//MAP_DEBUG("HTTP Req URL [%s]", url);
	__show_url_log(url);
	CURL *curl_handle;

	struct MultiCurlArg *pMultiCurlArg = (struct MultiCurlArg *)g_malloc0(sizeof(struct MultiCurlArg));

	if (!pMultiCurlArg) return -1;

	(pMultiCurlArg->m_chunk).memory = (char *) g_malloc0(1);	/* will be grown as needed by the realloc above */
	(pMultiCurlArg->m_chunk).size = 0;	/* no data at this point */

	/* init the curl session */
	curl_handle = curl_easy_init();

	/* set URL to get */
	curl_easy_setopt(curl_handle, CURLOPT_URL, url);

	curl_easy_setopt(curl_handle, CURLOPT_PROGRESSFUNCTION, __progress_info);
	curl_easy_setopt(curl_handle, CURLOPT_PROGRESSDATA, NULL);

#if LIBCURL_VERSION_NUM >= 0x72000
	curl_easy_setopt(curl_handle, CURLOPT_XFERINFOFUNCTION, __xferinfo);
	curl_easy_setopt(curl_handle, CURLOPT_XFERINFODATA, (void *)pMultiCurlArg);
#endif

	curl_easy_setopt(curl_handle, CURLOPT_NOPROGRESS, 0L);

	curl_easy_setopt(curl_handle, CURLOPT_SSL_VERIFYPEER, 1L);

	curl_easy_setopt(curl_handle, CURLOPT_PROXY, __proxy_address);

	/* send all data to this function  */
	curl_easy_setopt(curl_handle, CURLOPT_WRITEFUNCTION, __write_memory_callback);

	/* we want the headers to this file handle */
	curl_easy_setopt(curl_handle, CURLOPT_WRITEDATA, (void *)pMultiCurlArg);

	/* some servers don't like requests that are made without a user-agent field, so we provide one */
	curl_easy_setopt(curl_handle, CURLOPT_USERAGENT, "libcurl-agent/1.0");

	/* Timeout for the entire request */
	curl_easy_setopt(curl_handle, CURLOPT_TIMEOUT, TIZEN_CFG_CURL_TIMEOUT);

	pMultiCurlArg->curl_handle = curl_handle;
	pMultiCurlArg->req_type = type;
	pMultiCurlArg->request_id = req_id;
	pMultiCurlArg->user_data = user_data;
	strncpy(pMultiCurlArg->url, url, sizeof(pMultiCurlArg->url)-1);

	pthread_mutex_lock(&__MultiCurlArgsLock);
	__MultiCurlArgsRemain++;
	__add_to_list(pMultiCurlArg, true);
	pthread_mutex_unlock(&__MultiCurlArgsLock);

	pthread_mutex_lock(&__CmHandleLock);
	curl_multi_add_handle(__cm_handle, curl_handle);
	pthread_mutex_unlock(&__CmHandleLock);

	pthread_mutex_lock(&__curl_deliver_mutex);
	pthread_cond_signal(&__curl_delivered_cond);
	pthread_mutex_unlock(&__curl_deliver_mutex);
#endif
	return 0;
}

static void __post_curl(mapzen_req_type type, MemoryStruct_s *m_chunk, void *user_data)
{
	mapzen_resp_type resp_type = RESP_TYPE_NONE;
	switch (type) {
	case REQ_TYPE_GEOCODE:
		resp_type = RESP_TYPE_GEOCODE;
		break;
	case REQ_TYPE_REVGEOCODE:
		resp_type = RESP_TYPE_REVGEOCODE;
		break;
	case REQ_TYPE_PLACES:
		resp_type = RESP_TYPE_PLACES;
		break;
	case REQ_TYPE_PLACES_LIST:
		resp_type = RESP_TYPE_PLACES_LIST;
		break;
	case REQ_TYPE_PLACES_DETAILS:
		resp_type = RESP_TYPE_PLACES_DETAILS;
		break;
	case REQ_TYPE_ROUTE:
		resp_type = RESP_TYPE_ROUTE;
		break;
	default:
		resp_type = RESP_TYPE_NONE;
		break;
	}

	post_curl_response(m_chunk->memory, m_chunk->size, resp_type, user_data);
}

static void *__curl_thread_handler(void *user_data)
{
	//__thread_running_for_curlmulti = true;
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);

#ifdef RESTCURL_USE_MULTI_CURL
	CURLMsg *msg;
	int Q;
	int still_running = -1;

	while (__thread_running_for_curlmulti || __MultiCurlArgsRemain) {
		while (__MultiCurlArgsRemain) {
			pthread_mutex_lock(&__CmHandleLock);
			curl_multi_perform(__cm_handle, &still_running);
			pthread_mutex_unlock(&__CmHandleLock);

			if (still_running != __MultiCurlArgsRemain) {
				pthread_mutex_lock(&__CmHandleLock);
				msg = curl_multi_info_read(__cm_handle, &Q);
				pthread_mutex_unlock(&__CmHandleLock);

				if (msg == 0)
					break;

				if (msg->msg == CURLMSG_DONE) {
					MAP_DEBUG("Transfer completed for one handle");
					CURL *e = msg->easy_handle;

					pthread_mutex_lock(&__MultiCurlArgsLock);
					struct MultiCurlArg_List *del = NULL;
					if ((del = __delete_from_list(e)) != NULL) {
							MAP_DEBUG("Decrementing the handle counter in multiCurl..");
							__MultiCurlArgsRemain--;
					}
					pthread_mutex_unlock(&__MultiCurlArgsLock);

					if (del && del->mcArg) {
						// block to hide key in log
						//MAP_DEBUG("HTTP Url [%s]", del->mcArg->url);
						__show_url_log(del->mcArg->url);
						MAP_DEBUG("Posting the result..");
						__post_curl(del->mcArg->req_type, &(del->mcArg->m_chunk), del->mcArg->user_data);

						if ((del->mcArg->m_chunk).memory) {
							g_free((del->mcArg->m_chunk).memory);
							(del->mcArg->m_chunk).memory = NULL;
						}

						pthread_mutex_lock(&__CmHandleLock);
						curl_multi_remove_handle(__cm_handle, del->mcArg->curl_handle);
						pthread_mutex_unlock(&__CmHandleLock);

						curl_easy_cleanup(del->mcArg->curl_handle);

						g_free(del->mcArg);
						del->mcArg = NULL;

						g_free(del);
						del = NULL;
					}
				} else {
					MAP_DEBUG("E: CURLMsg (%d)\n", msg->msg);
				}
			}
		}

		pthread_mutex_lock(&__curl_deliver_mutex);
		if (__thread_running_for_curlmulti && !__MultiCurlArgsRemain) {
			#if 1
			struct timeval  now;
			struct timespec timeout;
			gettimeofday(&now, NULL);
			timeout.tv_sec = now.tv_sec + 3;
			timeout.tv_nsec = now.tv_usec * 1000;

			pthread_cond_timedwait(&__curl_delivered_cond, &__curl_deliver_mutex, &timeout);
			#else
			pthread_cond_wait(&__curl_delivered_cond, &__curl_deliver_mutex);
			#endif
		}
		pthread_mutex_unlock(&__curl_deliver_mutex);
	}
#else
	/* curl_easy_perform */
#endif
	pthread_exit(NULL);

	return 0;
}

void cancel_request(int request_id)
{
	if (__cancel_req_list == NULL)
		__cancel_req_list = g_list_append(__cancel_req_list, (gpointer) GINT_TO_POINTER(request_id));
	else
		__cancel_req_list = g_list_insert_before(__cancel_req_list, NULL, (gpointer) GINT_TO_POINTER(request_id));
}

int get_num_running_requests()
{
	MAP_DEBUG("Num of running requests :: >>>> [%d]", __MultiCurlArgsRemain);
	return __MultiCurlArgsRemain;
}

int init_curl()
{
	__cancel_req_list = NULL;

	head = NULL;
	curr = NULL;

	get_proxy_address(&__proxy_address);

#ifdef RESTCURL_USE_MULTI_CURL
	pthread_mutex_init(&__MultiCurlArgsLock, NULL);
	pthread_mutex_init(&__CmHandleLock, NULL);
	pthread_mutex_init(&__curl_deliver_mutex, NULL);
	pthread_cond_init(&__curl_delivered_cond, NULL);

	curl_global_init(CURL_GLOBAL_ALL);

	__cm_handle = curl_multi_init();
	curl_multi_setopt(__cm_handle, CURLMOPT_MAXCONNECTS, (long)CURL_MAX_CONNECTS);

	pthread_mutex_lock(&__curl_deliver_mutex);
	__MultiCurlArgsRemain = 0;
	__thread_running_for_curlmulti = true;
	pthread_mutex_unlock(&__curl_deliver_mutex);

	int ret = pthread_create(&__CurlThread, NULL, &__curl_thread_handler, NULL);
#endif

	return ret;
}

int deinit_curl()
{
#ifdef RESTCURL_USE_MULTI_CURL
	pthread_mutex_lock(&__curl_deliver_mutex);
	__thread_running_for_curlmulti = false;
	__MultiCurlArgsRemain = 0;
	pthread_cond_signal(&__curl_delivered_cond);
	pthread_mutex_unlock(&__curl_deliver_mutex);

	//pthread_cancel(__CurlThread);

	pthread_join(__CurlThread, NULL);

	curl_multi_cleanup(__cm_handle);
	curl_global_cleanup();

	pthread_mutex_destroy(&__MultiCurlArgsLock);
	pthread_mutex_destroy(&__CmHandleLock);
	pthread_cond_destroy(&__curl_delivered_cond);
	pthread_mutex_destroy(&__curl_deliver_mutex);
#endif

	if (__proxy_address) {
		free(__proxy_address);
		__proxy_address = NULL;
	}

	struct MultiCurlArg_List *temp = head;
	struct MultiCurlArg_List *next;

	while (temp) {
		next = temp->next;
		struct MultiCurlArg *curlArg = temp->mcArg;

		if (curlArg) {
			if ((curlArg->m_chunk).memory) {
				g_free((curlArg->m_chunk).memory);
				(curlArg->m_chunk).memory = NULL;
			}

			if (curlArg->user_data) {
				switch (curlArg->req_type) {
				case REQ_TYPE_GEOCODE:
				{
					MapzenGeocodeQueryData *queryData = (MapzenGeocodeQueryData *)curlArg->user_data;
					if (queryData) {
						if (queryData->user_data) {
							g_free(queryData->user_data);
							queryData->user_data = NULL;
						}
						g_free(queryData);
						queryData = NULL;
					}
				}
					break;
				case REQ_TYPE_REVGEOCODE:
				{
					MapzenRevGeocodeQueryData *queryData = (MapzenRevGeocodeQueryData *)curlArg->user_data;
					if (queryData) {
						if (queryData->user_data) {
							g_free(queryData->user_data);
							queryData->user_data = NULL;
						}
						g_free(queryData);
						queryData = NULL;
					}
				}
					break;
				case REQ_TYPE_PLACES:
				{
					MapzenPlaceQueryData *queryData = (MapzenPlaceQueryData *)curlArg->user_data;
					if (queryData) {
						if (queryData->user_data) {
							g_free(queryData->user_data);
							queryData->user_data = NULL;
						}
						g_free(queryData);
						queryData = NULL;
					}
				}
					break;
				case REQ_TYPE_PLACES_DETAILS:
				{
					MapzenPlaceDetailsQueryData *queryData = (MapzenPlaceDetailsQueryData *)curlArg->user_data;
					if (queryData) {
						if (queryData->user_data) {
							g_free(queryData->user_data);
							queryData->user_data = NULL;
						}
						g_free(queryData);
						queryData = NULL;
					}
				}
					break;
				case REQ_TYPE_PLACES_LIST:
				{
					MapzenPlaceListQueryData *queryData = (MapzenPlaceListQueryData *)curlArg->user_data;
					if (queryData) {
						if (queryData->user_data) {
							g_free(queryData->user_data);
							queryData->user_data = NULL;
						}
						g_free(queryData);
						queryData = NULL;
					}
				}
					break;
				case REQ_TYPE_ROUTE:
				{
					MapzenRouteQueryData *queryData = (MapzenRouteQueryData *)curlArg->user_data;
					if (queryData) {
						if (queryData->user_data) {
							g_free(queryData->user_data);
							queryData->user_data = NULL;
						}
						g_free(queryData);
						queryData = NULL;
					}
				}
					break;
				default:
					break;
				}
			}

			curl_easy_cleanup(curlArg->curl_handle);

			g_free(curlArg);
			curlArg  = NULL;
		}
		g_free(temp);

		temp = next;
	}
	head = NULL;

	g_list_free(__cancel_req_list);
	__cancel_req_list = NULL;

	return 0;
}
