/*
 * Copyright (c) 2014 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <glib.h>
#include <pthread.h>
#include <dlog.h>
#include "mapzen_route.hpp"

extern "C" {
#include "mapzen_types.h"
#include "mapzen_server_private.h"
#include "mapzen_debug.h"
#include "mapzen_queue.h"
#include "mapzen_restcurl.h"
#include "mapzen_util.h"
}

#include "rapidjson/rapidjson.h"
#include "rapidjson/allocators.h"
#include "rapidjson/document.h"
#include "rapidjson/stringbuffer.h"
#include "rapidjson/writer.h"
#include <curl/curl.h>

#ifdef __cplusplus
extern "C" {
#endif

#define ROUTE_URL       "https://valhalla.mapzen.com/route?api_key=%s&json=%s"


std::string url_encode(const std::string& unencoded) {
   char* encoded = curl_escape(unencoded.c_str(), static_cast<int>(unencoded.size()));
   MAP_DEBUG("route json after url encoding: %s", (encoded ? encoded : "(null)"));
   std::string encoded_str(encoded);
   curl_free(encoded);
   return encoded_str;
 }

int query_route(gchar *maps_key, coords_s startPoint, coords_s endPoint, route_type type, route_feature_avoids avoids, route_driving_style style, GList *waypoints, gpointer user_data)
{
	MAP_DEBUG(">>>>> START query_route");

	rapidjson::StringBuffer buffer;
	rapidjson::Writer<rapidjson::StringBuffer> writer(buffer);

	// start json request object
    writer.StartObject();

	//{"locations":[{"lat":42.358528,"lon":-83.271400,"street":"Appleton"},{"lat":42.996613,"lon":-78.749855,"street":"Ranch Trail"}]
	int length = g_list_length(waypoints);
	if (length != 0) {
		MAP_DEBUG(">>>>> query_route: PROCESS list of route locations");
		GList *waypoints_list = NULL;
		waypoints_list = g_list_first(waypoints);
	    writer.Key("locations");
	    writer.StartArray();

		while (waypoints_list) {
			coords_s *data = (coords_s *) waypoints_list->data;
			writer.StartObject();
			writer.Key("lat");
			writer.Double(data->latitude);
			writer.Key("lon");
			writer.Double(data->longitude);
			writer.EndObject();
			waypoints_list = g_list_next(waypoints_list);
		}
	    writer.EndArray();
	} else {
		MAP_DEBUG(">>>>> query_route: PROCESS two route locations");
	    writer.Key("locations");
	    writer.StartArray();

	    writer.StartObject();
	    writer.Key("lat");
	    writer.Double(startPoint.latitude);
	    writer.Key("lon");
	    writer.Double(startPoint.longitude);
	    writer.EndObject();

	    writer.StartObject();
	    writer.Key("lat");
	    writer.Double(endPoint.latitude);
	    writer.Key("lon");
	    writer.Double(endPoint.longitude);
	    writer.EndObject();

	    writer.EndArray();
	}

	MAP_DEBUG(">>>>> query_route: PROCESS type");
	std::string costing_str;
	switch (type) {
	case COSTING_AUTO:
		costing_str = "auto";
		break;
	case COSTING_AUTO_SHORTER:
		costing_str = "auto_shorter";
		break;
	case COSTING_PEDESTRIAN:
		costing_str = "pedestrian";
		break;
	case COSTING_MULTIMODAL:
		costing_str = "multimodal";
		break;
	case COSTING_BICYCLE:
		costing_str = "bicycle";
		break;
	default:
		costing_str = "auto";
	}

	MAP_DEBUG(">>>>> query_route: PROCESS costing");
    writer.Key("costing");
    writer.String(costing_str.c_str());

	MAP_DEBUG(">>>>> query_route: PROCESS directions_options");
    writer.Key("directions_options");
    writer.StartObject();
    writer.Key("units");
    writer.String("km");  // TODO determine if we should set from userdata
    writer.EndObject();

	if (avoids) {
		MAP_DEBUG(">>>>> query_route: PROCESS avoids");
		writer.Key("costing_options");
		writer.StartObject();
		switch (avoids) {
		case PENALTY_NONE:
			//TODO
			break;
		case PENALTY_LIMITED_ACCESS:
			//TODO
			break;
		case PENALTY_TOLL_ROADS:
			writer.Key("toll_booth_penalty");
			writer.String("1");
			break;
		case PENALTY_USE_FERRY:
			writer.Key("use_ferry");
			writer.String("0");
			break;
		case PENALTY_USE_UNPAVED:
			//TODO
			break;
		case PENALTY_SEASONAL_CLOSURE:
			//TODO
			break;
		case PENALTY_COUNTRY_CROSSING:
			writer.Key("country_crossing_penalty");
			writer.String("1");
			break;
		}
		writer.EndObject();
	}

	// end json request object
    writer.EndObject();

	MAP_DEBUG(">>>>> query_route: JSON request=%s", buffer.GetString());

	const char* jsonString = buffer.GetString();

	//we think that this is already being encoded somewhere else
	//std::string encoded_json = url_encode(json);

	MAP_DEBUG(">>>>> query_route: PROCESS routeRequest");
	char routeRequest[1024];
	if (maps_key != NULL)
		snprintf(routeRequest, sizeof(routeRequest), ROUTE_URL, maps_key, jsonString);
	else
		snprintf(routeRequest, sizeof(routeRequest), ROUTE_URL, "null", jsonString);

	add_handle(routeRequest, REQ_TYPE_ROUTE, user_data);

	MAP_DEBUG(">>>>> END query_route");
	return 0;
}

#ifdef __cplusplus
}
#endif
