/*
 * Copyright (c) 2014 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _MAPZEN_TYPES_H_
#define _MAPZEN_TYPES_H_

#include <glib.h>

typedef enum {
	MAPZEN_ERROR_NONE = 0,                /**< Successful */
        MAPZEN_ERROR_PERMISSION_DENIED,       /**< Permission Denied */
        MAPZEN_ERROR_OUT_OF_MEMORY,           /**< Out of memory */
        MAPZEN_ERROR_INVALID_PARAMETER,       /**< Invalid parameter */
        MAPZEN_ERROR_NOT_SUPPORTED,           /**< Not supported */
        MAPZEN_ERROR_CONNECTION_TIMED_OUT,    /**< Timeout error, no answer */
        MAPZEN_ERROR_NETWORK_UNREACHABLE,     /**< Network unavailable */
        MAPZEN_ERROR_INVALID_OPERATION,       /**< Opeartion is not valid */
        MAPZEN_ERROR_KEY_NOT_AVAILABLE,       /**< Invalid key */
        MAPZEN_ERROR_RESOURCE_BUSY,           /**< Resource busy */
        MAPZEN_ERROR_CANCELED,                /**< Service canceled */
        MAPZEN_ERROR_UNKNOWN,                 /**< Unknown error */
        MAPZEN_ERROR_SERVICE_NOT_AVAILABLE,   /**< Service unavailabe*/
        MAPZEN_ERROR_NOT_FOUND,               /**< Result not found */
} mapzen_error_e;

typedef enum {
	MAPZEN_BOUNDARY_NONE = 0,	/*  Undefined geographical area type. */
	MAPZEN_BOUNDARY_RECT,		/*  Rectangular geographical area type. */
	MAPZEN_BOUNDARY_CIRCLE,	/*  Circle geographical area type. */
} boundary_type;

typedef enum {
	COSTING_AUTO = 0,
	COSTING_AUTO_SHORTER,
	COSTING_BICYCLE,
	COSTING_MULTIMODAL,
	COSTING_PEDESTRIAN
} route_type;

typedef enum {
	UNIT_KM = 0,	/* for kilometers */
	UNIT_MI = 1 	/* for miles */
} route_unit;

typedef enum {
	MAPZEN_MANEUVER_TYPE_NONE = 0,
	MAPZEN_MANEUVER_TYPE_START = 1,
	MAPZEN_MANEUVER_TYPE_START_RIGHT = 2,
	MAPZEN_MANEUVER_TYPE_START_LEFT = 3,
	MAPZEN_MANEUVER_TYPE_DESTINATION = 4,
	MAPZEN_MANEUVER_TYPE_DESTINATION_RIGHT = 5,
	MAPZEN_MANEUVER_TYPE_DESTINATION_LEFT = 6,
	MAPZEN_MANEUVER_TYPE_BECOMES = 7,
	MAPZEN_MANEUVER_TYPE_CONTINUE = 8,
	MAPZEN_MANEUVER_TYPE_SLIGHT_RIGHT = 9,
	MAPZEN_MANEUVER_TYPE_RIGHT = 10,
	MAPZEN_MANEUVER_TYPE_SHARP_RIGHT = 11,
	MAPZEN_MANEUVER_TYPE_UTURN_RIGHT = 12,
	MAPZEN_MANEUVER_TYPE_UTURN_LEFT = 13,
	MAPZEN_MANEUVER_TYPE_SHARP_LEFT = 14,
	MAPZEN_MANEUVER_TYPE_LEFT = 15,
	MAPZEN_MANEUVER_TYPE_SLIGHT_LEFT = 16,
	MAPZEN_MANEUVER_TYPE_RAMP_STRAIGHT = 17,
	MAPZEN_MANEUVER_TYPE_RAMP_RIGHT = 18,
	MAPZEN_MANEUVER_TYPE_RAMP_LEFT = 19,
	MAPZEN_MANEUVER_TYPE_EXIT_RIGHT = 20,
	MAPZEN_MANEUVER_TYPE_EXIT_LEFT = 21,
	MAPZEN_MANEUVER_TYPE_STAY_STRAIGHT = 22,
	MAPZEN_MANEUVER_TYPE_STAY_RIGHT = 23,
	MAPZEN_MANEUVER_TYPE_STAY_LEFT = 24,
	MAPZEN_MANEUVER_TYPE_MERGE = 25,
	MAPZEN_MANEUVER_TYPE_ROUNDABOUT_ENTER = 26,
	MAPZEN_MANEUVER_TYPE_ROUNDABOUT_EXIT = 27,
	MAPZEN_MANEUVER_TYPE_FERRY_ENTER = 28,
	MAPZEN_MANEUVER_TYPE_FERRY_EXIT = 29,
	MAPZEN_MANEUVER_TYPE_TRANSIT = 30,
	MAPZEN_MANEUVER_TYPE_TRANSIT_TRANSFER = 31,
	MAPZEN_MANEUVER_TYPE_TRANSIT_REMAIN_ON = 32,
	MAPZEN_MANEUVER_TYPE_TRANSIT_CONNECTION_START = 33,
	MAPZEN_MANEUVER_TYPE_TRANSIT_CONNECTION_TRANSFER = 34,
	MAPZEN_MANEUVER_TYPE_TRANSIT_CONNECTION_DESTINATION = 35,
	MAPZEN_MANEUVER_TYPE_POST_TRANSIT_CONNECTION_DESTINATION = 36
} mapzen_maneuver_type;

typedef enum {
	PENALTY_NONE = 0, //TODO: Mapzen does not have currently
	PENALTY_LIMITED_ACCESS, //TODO: Mapzen does not have currently
	PENALTY_TOLL_ROADS, //set to > 0 to avoid
	PENALTY_USE_FERRY, //This is range of values between 0 and 1. Values near 0 attempt to avoid ferries and
	                         //values near 1 will favor ferries.
	PENALTY_USE_UNPAVED, //TODO: Mapzen does not have currently
	PENALTY_SEASONAL_CLOSURE, //TODO: Mapzen does not have currently
	PENALTY_COUNTRY_CROSSING // set to > 0 to avoid
} route_feature_avoids;

typedef enum {
	DRIVING_STYLE_NORMAL = 0,
	DRIVING_STYLE_CAUTIOUS,
	DRIVING_STYLE_AGGRESSIVE
} route_driving_style;

typedef enum {
	TANGRAM_VIEW_NORMAL = 0,
	TANGRAM_VIEW_TERRAIN,
	TANGRAM_VIEW_NONE
} tangram_view_type;

typedef struct {
	gdouble latitude;
	gdouble longitude;
} coords_s;

typedef struct {
	coords_s top_left;
	coords_s bottom_right;
} rectangle_s;

typedef struct {
	coords_s center;
	gdouble radius;
} circle_s;

typedef struct {
	boundary_type type;
	union {
		rectangle_s rect;
		circle_s circle;
	};
} mapzen_boundary_s;

typedef struct {
	gchar *maps_key;
	gchar *address;
	mapzen_boundary_s *boundary;
	gint num_res;
} mapzen_geocode_req_s;

typedef struct {
	gchar *maps_key;
	gint num_res;
	coords_s coordinates;
} mapzen_revgeocode_req_s;

typedef struct {
	gchar *maps_key;
	gchar *search_string;
	gchar *categories;
	gint num_res;
	mapzen_boundary_s *boundary;
} mapzen_search_req_s;

typedef struct {
	gchar *maps_key;
	gchar *uri;
} mapzen_get_details_req_s;

typedef struct {
	gchar *maps_key;
	coords_s from;
	coords_s to;
	route_unit unit;
	route_type type;
	route_feature_avoids avoids;		/* List of type strings. (Limited Access, Toll Road, Ferry, Unpaved, Seasonal Closure, Country Crossing) */
	route_driving_style driving_style;	/* (1 - cautious, 2 - normal, 3 - aggressive) */
	GList *way_points;				/* List of type coords_s */
} mapzen_route_req_s;

typedef struct {
	gint zoom_level;
	coords_s center_coords;
} mapzen_tiledata_req_s;

typedef struct {
	gchar *street;
	gchar *neighbourhood;
	gchar *housenumber;
	gchar *localadmin;
	gchar *county;
	gchar *region;
	gchar *country;
	gchar *country_a;
	gchar *postalcode;
} mapzen_address_resp_s;

typedef struct {
	gchar *place_id;
	gchar *display_name;
	mapzen_address_resp_s *address;
	coords_s coordinates;
	GList *categories;
} mapzen_place_resp_s;

typedef struct {
	guint type;
	guint direction;
	guint text;
	gchar *icon_url;
} mapzen_route_maneuver_signs;

typedef struct {
	coords_s start_point;
	coords_s end_point;
	gdouble distance;
	guint time;
	mapzen_maneuver_type type;
	gchar *instruction;
	gchar *street_name;
} mapzen_route_maneuver;

typedef struct {
	coords_s origin;
	coords_s destination;
	rectangle_s bounding_box;
	gdouble distance;
	guint time;
	GList *maneuvers;		/* List of type mapzen_route_maneuver */
	GList *shapePoints;
} mapzen_route_segment;

typedef struct {
	coords_s origin;
	coords_s destination;
	rectangle_s bounding_box;
	gdouble distance;
	route_unit distance_unit;
	guint time;
	route_type type;
	GList *segments;		/* List of type mapzen_route_segment */
	GList *shapePoints;
	gchar *language;
} mapzen_route_resp_s;

#endif /* _MAPZEN_TYPES_H_ */
