/*
 * Copyright (c) 2014 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "mapzen_util.h"
#include "mapzen_debug.h"
#include <maps_error.h>
#include <maps_view.h>
#include <maps_route_maneuver.h>
#include <math.h>

#define PI 3.14159265359
#define EARTH_RADIUS	6371

void calculate_point(gdouble Lat1, gdouble Lon1, int dBearing, gdouble dist, coords_s **coord)
{
	Lat1 = (Lat1 / 180.0) * PI;
	Lon1 = (Lon1 / 180.0) * PI;

	gdouble dLat = asin(sin(Lat1) * cos(dist / EARTH_RADIUS) + (cos(Lat1) * sin(dist / EARTH_RADIUS) * cos(dBearing)));

	dLat = radians_to_degrees(dLat);

	gdouble dLon = Lon1 + atan2(sin(dBearing) * sin(dist / EARTH_RADIUS) * cos(Lat1), cos(dist / EARTH_RADIUS) - sin(Lat1) * sin((dLat / 180.0) * PI));

	dLon = radians_to_degrees(dLon);

	if (*coord == NULL)
		*coord = (coords_s *)g_malloc0(sizeof(coords_s));

	if (*coord) {
		(*coord)->latitude = dLat;
		(*coord)->longitude = dLon;
	}
}

double wrap_longitude(double longitude)
{
	return longitude - round(longitude / 360.0) * 360.0;
}

double wrap_latitude(double latitude)
{
	return latitude - round(latitude / 180.0) * 180.0;
}

double degrees_to_radians(double degrees)
{
	static double deg_to_rad = PI / 180.0;
	return degrees * deg_to_rad;
}

double radians_to_degrees(double radians)
{
	static double rad_to_deg = 180.0 / PI;
	return radians * rad_to_deg;
}

int convert_maps_error_to_mapzen_error(int error)
{
	switch (error) {
	case MAPS_ERROR_NONE:                  return MAPZEN_ERROR_NONE;
	case MAPS_ERROR_PERMISSION_DENIED:     return MAPZEN_ERROR_PERMISSION_DENIED;
	case MAPS_ERROR_OUT_OF_MEMORY:         return MAPZEN_ERROR_OUT_OF_MEMORY;
	case MAPS_ERROR_INVALID_PARAMETER:     return MAPZEN_ERROR_INVALID_PARAMETER;
	case MAPS_ERROR_NOT_SUPPORTED:         return MAPZEN_ERROR_NOT_SUPPORTED;
	case MAPS_ERROR_CONNECTION_TIME_OUT:   return MAPZEN_ERROR_CONNECTION_TIMED_OUT;
	case MAPS_ERROR_NETWORK_UNREACHABLE:   return MAPZEN_ERROR_NETWORK_UNREACHABLE;
	case MAPS_ERROR_INVALID_OPERATION:     return MAPZEN_ERROR_INVALID_OPERATION;
	case MAPS_ERROR_KEY_NOT_AVAILABLE:     return MAPZEN_ERROR_KEY_NOT_AVAILABLE;
	case MAPS_ERROR_RESOURCE_BUSY:         return MAPZEN_ERROR_RESOURCE_BUSY;
	case MAPS_ERROR_CANCELED:              return MAPZEN_ERROR_CANCELED;
	case MAPS_ERROR_UNKNOWN:               return MAPZEN_ERROR_UNKNOWN;
	case MAPS_ERROR_SERVICE_NOT_AVAILABLE: return MAPZEN_ERROR_SERVICE_NOT_AVAILABLE;
	case MAPS_ERROR_NOT_FOUND:             return MAPZEN_ERROR_NOT_FOUND;
	default: return MAPZEN_ERROR_UNKNOWN;
	}
}

int convert_mapzen_error_to_maps_error(int error)
{
	switch (error) {
	case MAPZEN_ERROR_NONE:                  return MAPS_ERROR_NONE;
	case MAPZEN_ERROR_PERMISSION_DENIED:     return MAPS_ERROR_PERMISSION_DENIED;
	case MAPZEN_ERROR_OUT_OF_MEMORY:         return MAPS_ERROR_OUT_OF_MEMORY;
	case MAPZEN_ERROR_INVALID_PARAMETER:     return MAPS_ERROR_INVALID_PARAMETER;
	case MAPZEN_ERROR_NOT_SUPPORTED:         return MAPS_ERROR_NOT_SUPPORTED;
	case MAPZEN_ERROR_CONNECTION_TIMED_OUT:  return MAPS_ERROR_CONNECTION_TIME_OUT;
	case MAPZEN_ERROR_NETWORK_UNREACHABLE:   return MAPS_ERROR_NETWORK_UNREACHABLE;
	case MAPZEN_ERROR_INVALID_OPERATION:     return MAPS_ERROR_INVALID_OPERATION;
	case MAPZEN_ERROR_KEY_NOT_AVAILABLE:     return MAPS_ERROR_KEY_NOT_AVAILABLE;
	case MAPZEN_ERROR_RESOURCE_BUSY:         return MAPS_ERROR_RESOURCE_BUSY;
	case MAPZEN_ERROR_CANCELED:              return MAPS_ERROR_CANCELED;
	case MAPZEN_ERROR_UNKNOWN:               return MAPS_ERROR_UNKNOWN;
	case MAPZEN_ERROR_SERVICE_NOT_AVAILABLE: return MAPS_ERROR_SERVICE_NOT_AVAILABLE;
	case MAPZEN_ERROR_NOT_FOUND:             return MAPS_ERROR_NOT_FOUND;
	default: return MAPS_ERROR_UNKNOWN;
	}
}

mapzen_maneuver_type convert_int_to_mapzen_maneuver_type(int maneuver_type_id)
{
	switch (maneuver_type_id) {
	case 0:  return MAPZEN_MANEUVER_TYPE_NONE;
	case 1:  return MAPZEN_MANEUVER_TYPE_START;
	case 2:  return MAPZEN_MANEUVER_TYPE_START_RIGHT;
	case 3:  return MAPZEN_MANEUVER_TYPE_START_LEFT;
	case 4:  return MAPZEN_MANEUVER_TYPE_DESTINATION;
	case 5:  return MAPZEN_MANEUVER_TYPE_DESTINATION_RIGHT;
	case 6:  return MAPZEN_MANEUVER_TYPE_DESTINATION_LEFT;
	case 7:  return MAPZEN_MANEUVER_TYPE_BECOMES;
	case 8:  return MAPZEN_MANEUVER_TYPE_CONTINUE;
	case 9:  return MAPZEN_MANEUVER_TYPE_SLIGHT_RIGHT;
	case 10: return MAPZEN_MANEUVER_TYPE_RIGHT;
	case 11: return MAPZEN_MANEUVER_TYPE_SHARP_RIGHT;
	case 12: return MAPZEN_MANEUVER_TYPE_UTURN_RIGHT;
	case 13: return MAPZEN_MANEUVER_TYPE_UTURN_LEFT;
	case 14: return MAPZEN_MANEUVER_TYPE_SHARP_LEFT;
	case 15: return MAPZEN_MANEUVER_TYPE_LEFT;
	case 16: return MAPZEN_MANEUVER_TYPE_SLIGHT_LEFT;
	case 17: return MAPZEN_MANEUVER_TYPE_RAMP_STRAIGHT;
	case 18: return MAPZEN_MANEUVER_TYPE_RAMP_RIGHT;
	case 19: return MAPZEN_MANEUVER_TYPE_RAMP_LEFT;
	case 20: return MAPZEN_MANEUVER_TYPE_EXIT_RIGHT;
	case 21: return MAPZEN_MANEUVER_TYPE_EXIT_LEFT;
	case 22: return MAPZEN_MANEUVER_TYPE_STAY_STRAIGHT;
	case 23: return MAPZEN_MANEUVER_TYPE_STAY_RIGHT;
	case 24: return MAPZEN_MANEUVER_TYPE_STAY_LEFT;
	case 25: return MAPZEN_MANEUVER_TYPE_MERGE;
	case 26: return MAPZEN_MANEUVER_TYPE_ROUNDABOUT_ENTER;
	case 27: return MAPZEN_MANEUVER_TYPE_ROUNDABOUT_EXIT;
	case 28: return MAPZEN_MANEUVER_TYPE_FERRY_ENTER;
	case 29: return MAPZEN_MANEUVER_TYPE_FERRY_EXIT;
	case 30: return MAPZEN_MANEUVER_TYPE_TRANSIT;
	case 31: return MAPZEN_MANEUVER_TYPE_TRANSIT_TRANSFER;
	case 32: return MAPZEN_MANEUVER_TYPE_TRANSIT_REMAIN_ON;
	case 33: return MAPZEN_MANEUVER_TYPE_TRANSIT_CONNECTION_START;
	case 34: return MAPZEN_MANEUVER_TYPE_TRANSIT_CONNECTION_TRANSFER;
	case 35: return MAPZEN_MANEUVER_TYPE_TRANSIT_CONNECTION_DESTINATION;
	case 36: return MAPZEN_MANEUVER_TYPE_POST_TRANSIT_CONNECTION_DESTINATION;
	default: return MAPZEN_MANEUVER_TYPE_NONE;
	}
}

int convert_mapzen_maneuver_type_to_maps_turn_type(mapzen_maneuver_type maneuver_type)
{
	switch (maneuver_type) {
	case MAPZEN_MANEUVER_TYPE_NONE:  return MAPS_ROUTE_TURN_TYPE_NONE;
	case MAPZEN_MANEUVER_TYPE_START:  return MAPS_ROUTE_TURN_TYPE_STRAIGHT;
	case MAPZEN_MANEUVER_TYPE_START_RIGHT:  return MAPS_ROUTE_TURN_TYPE_STRAIGHT;
	case MAPZEN_MANEUVER_TYPE_START_LEFT:  return MAPS_ROUTE_TURN_TYPE_STRAIGHT;
	case MAPZEN_MANEUVER_TYPE_DESTINATION:  return MAPS_ROUTE_TURN_TYPE_NONE;
	case MAPZEN_MANEUVER_TYPE_DESTINATION_RIGHT:  return MAPS_ROUTE_TURN_TYPE_NONE;
	case MAPZEN_MANEUVER_TYPE_DESTINATION_LEFT:  return MAPS_ROUTE_TURN_TYPE_NONE;
	case MAPZEN_MANEUVER_TYPE_BECOMES:  return MAPS_ROUTE_TURN_TYPE_STRAIGHT;
	case MAPZEN_MANEUVER_TYPE_CONTINUE:  return MAPS_ROUTE_TURN_TYPE_STRAIGHT;
	case MAPZEN_MANEUVER_TYPE_SLIGHT_RIGHT:  return MAPS_ROUTE_TURN_TYPE_LIGHT_RIGHT;
	case MAPZEN_MANEUVER_TYPE_RIGHT:  return MAPS_ROUTE_TURN_TYPE_RIGHT;
	case MAPZEN_MANEUVER_TYPE_SHARP_RIGHT:  return MAPS_ROUTE_TURN_TYPE_HARD_RIGHT;
	case MAPZEN_MANEUVER_TYPE_UTURN_RIGHT:  return MAPS_ROUTE_TURN_TYPE_UTURN_RIGHT;
	case MAPZEN_MANEUVER_TYPE_UTURN_LEFT:  return MAPS_ROUTE_TURN_TYPE_UTURN_LEFT;
	case MAPZEN_MANEUVER_TYPE_SHARP_LEFT:  return MAPS_ROUTE_TURN_TYPE_HARD_LEFT;
	case MAPZEN_MANEUVER_TYPE_LEFT:  return MAPS_ROUTE_TURN_TYPE_LEFT;
	case MAPZEN_MANEUVER_TYPE_SLIGHT_LEFT:  return MAPS_ROUTE_TURN_TYPE_LIGHT_LEFT;
	case MAPZEN_MANEUVER_TYPE_RAMP_STRAIGHT:  return MAPS_ROUTE_TURN_TYPE_STRAIGHT;
	case MAPZEN_MANEUVER_TYPE_RAMP_RIGHT:  return MAPS_ROUTE_TURN_TYPE_BEAR_RIGHT;
	case MAPZEN_MANEUVER_TYPE_RAMP_LEFT:  return MAPS_ROUTE_TURN_TYPE_BEAR_LEFT;
	case MAPZEN_MANEUVER_TYPE_EXIT_RIGHT:  return MAPS_ROUTE_TURN_TYPE_BEAR_RIGHT;
	case MAPZEN_MANEUVER_TYPE_EXIT_LEFT:  return MAPS_ROUTE_TURN_TYPE_BEAR_LEFT;
	case MAPZEN_MANEUVER_TYPE_STAY_STRAIGHT:  return MAPS_ROUTE_TURN_TYPE_STRAIGHT;
	case MAPZEN_MANEUVER_TYPE_STAY_RIGHT:  return MAPS_ROUTE_TURN_TYPE_BEAR_RIGHT;
	case MAPZEN_MANEUVER_TYPE_STAY_LEFT:  return MAPS_ROUTE_TURN_TYPE_BEAR_LEFT;
	case MAPZEN_MANEUVER_TYPE_MERGE:  return MAPS_ROUTE_TURN_TYPE_STRAIGHT;
	case MAPZEN_MANEUVER_TYPE_ROUNDABOUT_ENTER:  return MAPS_ROUTE_TURN_TYPE_STRAIGHT;
	case MAPZEN_MANEUVER_TYPE_ROUNDABOUT_EXIT:  return MAPS_ROUTE_TURN_TYPE_STRAIGHT;
	case MAPZEN_MANEUVER_TYPE_FERRY_ENTER:  return MAPS_ROUTE_TURN_TYPE_STRAIGHT;
	case MAPZEN_MANEUVER_TYPE_FERRY_EXIT:  return MAPS_ROUTE_TURN_TYPE_STRAIGHT;
	case MAPZEN_MANEUVER_TYPE_TRANSIT:  return MAPS_ROUTE_TURN_TYPE_NONE;
	case MAPZEN_MANEUVER_TYPE_TRANSIT_TRANSFER:  return MAPS_ROUTE_TURN_TYPE_NONE;
	case MAPZEN_MANEUVER_TYPE_TRANSIT_REMAIN_ON:  return MAPS_ROUTE_TURN_TYPE_NONE;
	case MAPZEN_MANEUVER_TYPE_TRANSIT_CONNECTION_START:  return MAPS_ROUTE_TURN_TYPE_NONE;
	case MAPZEN_MANEUVER_TYPE_TRANSIT_CONNECTION_TRANSFER:  return MAPS_ROUTE_TURN_TYPE_NONE;
	case MAPZEN_MANEUVER_TYPE_TRANSIT_CONNECTION_DESTINATION:  return MAPS_ROUTE_TURN_TYPE_NONE;
	case MAPZEN_MANEUVER_TYPE_POST_TRANSIT_CONNECTION_DESTINATION: return MAPS_ROUTE_TURN_TYPE_NONE;
	default: return MAPS_ROUTE_TURN_TYPE_NONE;
	}
}

int convert_tangram_view_type_to_maps_view_type(int maps_view_type)
{
	switch (maps_view_type) {
	case TANGRAM_VIEW_NORMAL:	return MAPS_VIEW_TYPE_NORMAL;
	case TANGRAM_VIEW_TERRAIN:	return MAPS_VIEW_TYPE_TERRAIN;
	default: return MAPS_VIEW_TYPE_NORMAL;
	}
}

int convert_maps_view_type_to_tangram_view_type(int maps_view_type)
{
	switch (maps_view_type) {
	case MAPS_VIEW_TYPE_NORMAL:		return TANGRAM_VIEW_NORMAL;
	case MAPS_VIEW_TYPE_SATELLITE:	return TANGRAM_VIEW_NONE;
	case MAPS_VIEW_TYPE_TERRAIN:	return TANGRAM_VIEW_TERRAIN;
	case MAPS_VIEW_TYPE_HYBRID:		return TANGRAM_VIEW_NONE;
	default: return TANGRAM_VIEW_NONE;
	}
}
