/*
 * Copyright (c) 2014 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

extern "C" {
#include "mapzen_queue.h"
#include "mapzen_debug.h"
#include "mapzen_util.h"
#include "mapzen_restcurl.h"
}

#include "mapzen_plugin_internal.h"
#include "tangram_view.hpp"
#include <tangram.h>
#include <src/platform_tizen.h>
#include <Elementary.h>

#include <string>
#include <cstdio>

#include <sys/syscall.h> // SYS_gettid

#define NORMAL_SCENE_FILE_PATH "/usr/share/maps/mapzen/scenes/bubble-wrap/bubble-wrap.yaml"
#define TERRAIN_SCENE_FILE_PATH "/usr/share/maps/mapzen/scenes/walkabout-style/walkabout-style.yaml"


TangramView::TangramView()
{
	// Nothing to do.
}

TangramView::~TangramView()
{
	// Nothing to do.
}

mapzen_error_e TangramView::create(maps_view_h view, maps_plugin_map_view_ready_cb callback, const char* providerKey)
{
	if (!view) {
		return MAPZEN_ERROR_INVALID_PARAMETER;
	}

	int maps_error = MAPS_ERROR_NONE;
	maps_error = maps_view_get_viewport(view, &m_image);
	if (maps_error != MAPS_ERROR_NONE) {
		return (mapzen_error_e)convert_maps_error_to_mapzen_error(maps_error);
	}

	maps_error = maps_view_get_screen_location(view, &m_x, &m_y, &m_w, &m_h);
	if (maps_error != MAPS_ERROR_NONE) {
		return (mapzen_error_e)convert_maps_error_to_mapzen_error(maps_error);
	}
	m_w = MAX(m_w, 1);
	m_h = MAX(m_h, 1);

	// Create an OpenGL context.
	{
		m_config = evas_gl_config_new();
		if (!m_config) {
			MAPS_LOGE("evas_gl_config_new() failed");
			return MAPZEN_ERROR_SERVICE_NOT_AVAILABLE;
		}

		m_config->color_format = EVAS_GL_RGBA_8888;
		m_config->depth_bits   = EVAS_GL_DEPTH_BIT_24;
		m_config->stencil_bits = EVAS_GL_STENCIL_NONE;
		m_config->options_bits = EVAS_GL_OPTIONS_DIRECT;

		m_gl = evas_gl_new(evas_object_evas_get(m_image));
		if (!m_gl) {
			MAPS_LOGE("evas_gl_new() failed");
			return MAPZEN_ERROR_SERVICE_NOT_AVAILABLE;
		}

		m_api = evas_gl_api_get(m_gl);
		if (!m_api) {
			MAPS_LOGE("evas_gl_api_get() failed");
			return MAPZEN_ERROR_SERVICE_NOT_AVAILABLE;
		}

		m_context = evas_gl_context_create(m_gl, nullptr);
		if (!m_context) {
			MAPS_LOGE("evas_gl_context_create() failed");
			return MAPZEN_ERROR_SERVICE_NOT_AVAILABLE;
		}
	}

	// Set up rendering surface.
	auto error = setupOpenGlSurface(view);
	if (error != MAPZEN_ERROR_NONE) {
		return error;
	}

	m_readyCb = callback;
	// TODO: What to do for multiple instances
	Tangram::setEvasGlAPI(m_api);

	Tangram::UrlClient::Options urlOptions;
	char* proxyAddress = NULL;
	get_proxy_address(&proxyAddress);
	if (proxyAddress) {
		urlOptions.proxyAddress = proxyAddress;
		free(proxyAddress);
	}

	urlOptions.connectionTimeoutMs = 0;
	urlOptions.requestTimeoutMs = 120000;

	m_platform = std::make_shared<Tangram::TizenPlatform>(urlOptions);

	m_uiThreadId = syscall(SYS_gettid);

	m_platform->setRenderCallbackFunction([&]() {

		if (m_uiThreadId == syscall(SYS_gettid)) {
			m_dirty = false;
			evas_object_image_pixels_dirty_set(m_image, EINA_TRUE);
		} else {
			m_dirty = true;
		}
	});

	// NB: Since there is no safe way to use ecore_main_loop_thread_safe_call_async in a
	// module that could be unloaded at runtime we use polling to check on the mainloop
	// whether a render request came in.
	// TODO only run timer when async tasks (Tile- and Scene-Loading) are in progress.
	m_renderRequestTimer = ecore_timer_add(0.05, [](void *data){
		auto tv = static_cast<TangramView*>(data);
		if (tv->m_dirty) {
			tv->m_dirty = false;
			evas_object_image_pixels_dirty_set((Evas_Object*)tv->m_image, EINA_TRUE);
		}
		return EINA_TRUE;
	}, this);

	// Set up the tangram map.
	m_map.reset(new Tangram::Map(m_platform));

	float scaleFactor = elm_config_scale_get();
	MAPS_LOGD("evas_gl_context_create() set PixelScale %f", scaleFactor);
	m_map->setPixelScale(scaleFactor);

	if (providerKey) {
		m_providerKey = providerKey;
	}

	// Start loading the scene by setting the map type.
	setMapType(view);

	// Make the GL context current and perform GL setup.
	evas_gl_make_current(m_gl, m_surface, m_context);
	m_map->setupGL();

	m_map->resize(m_w, m_h);

	if (m_sceneLoaded && m_w > 1 && m_h > 1) {
		readyMapCb((void*)view);
	}

	m_isInitialized = true;

	return MAPZEN_ERROR_NONE;
}

mapzen_error_e TangramView::setupOpenGlSurface(maps_view_h view)
{
	if (!view || !m_gl || !m_config || !m_image || !m_context) {
		return MAPZEN_ERROR_INVALID_PARAMETER;
	}

	// Remove any previous pixel callback on the image.
	evas_object_image_pixels_get_callback_set(m_image, nullptr, nullptr);

	if (m_surface) {
		// Destroy the old surface.
		evas_object_image_native_surface_set(m_image, nullptr);
		evas_gl_surface_destroy(m_gl, m_surface);
	}

	m_w = MAX(m_w, 1);
	m_h = MAX(m_h, 1);

	evas_object_image_size_set(m_image, m_w, m_h);

	Evas_Native_Surface native_surface;
	m_surface = evas_gl_surface_create(m_gl, m_config, m_w, m_h);
	if (!m_surface) {
		MAPS_LOGE("evas_gl_surface_create() failed");
		return MAPZEN_ERROR_SERVICE_NOT_AVAILABLE;
	}

	if (!evas_gl_native_surface_get(m_gl, m_surface, &native_surface)) {
		// Could not get the native surface information, so destroy the surface and exit.
		evas_gl_make_current(m_gl, nullptr, nullptr);
		evas_gl_surface_destroy(m_gl, m_surface);
		m_surface = nullptr;
		MAPS_LOGE("evas_gl_native_surface_get() failed");
		return MAPZEN_ERROR_SERVICE_NOT_AVAILABLE;
	}

	// Set the native surface information and pixel callback on the image.
	evas_object_image_native_surface_set(m_image, &native_surface);
	evas_object_image_pixels_get_callback_set(m_image, pixelGetCb, view);

	return MAPZEN_ERROR_NONE;
}

mapzen_error_e TangramView::destroy(maps_view_h view)
{
	if (!view) {
		return MAPZEN_ERROR_INVALID_PARAMETER;
	}

	m_isInitialized = false;
	m_sceneLoaded = false;

	if (m_renderRequestTimer) {
		ecore_timer_del(m_renderRequestTimer);
		m_renderRequestTimer = nullptr;
	}

	if (m_image) {
		evas_object_image_pixels_get_callback_set(m_image, nullptr, nullptr);
	}

	if (m_gl) {
		if (m_surface && m_context) {
			evas_gl_make_current(m_gl, m_surface, m_context);
		}
	}

	if (m_map) {
		removeAllObjects();

		m_map.reset();
		m_platform.reset();
	}

	if (m_gl) {
		if (m_surface) {
			evas_object_image_native_surface_set(m_image, nullptr);
			evas_gl_surface_destroy(m_gl, m_surface);
		}
		if (m_context) {
			evas_gl_context_destroy(m_gl, m_context);
		}
		evas_gl_free(m_gl);
		m_gl = nullptr;
	}

	if (m_config) {
		evas_gl_config_free(m_config);
	}

	return MAPZEN_ERROR_NONE;
}

mapzen_error_e TangramView::render(maps_view_h view, const maps_coordinates_h coord, double zoom, double angle)
{
	if (!view || !coord) {
		MAPS_LOGE("Render requested with invalid 'view' or 'coord'.");
		return MAPZEN_ERROR_INVALID_PARAMETER;
	}

	if (!m_isInitialized) {
		MAPS_LOGE("Render requested with view that is not initialized.");
		return MAPZEN_ERROR_SERVICE_NOT_AVAILABLE;
	}

	setMapType(view);

	int x = 0, y = 0, w = 0, h = 0;
	maps_view_get_screen_location(view, &x, &y, &w, &h);
	m_w = MAX(m_w, 1);
	m_h = MAX(m_h, 1);

	if (x != m_x || y != m_y || w != m_w || h != m_h) {
		if ( (m_w <= 1 || m_h <= 1) && w > 1 && h > 1 && m_sceneLoaded) {
			// Send a tangram view ready callback when we have a legit canvas size
			readyMapCb((void*)view);
		}
		m_x = x;
		m_y = y;
		m_w = w;
		m_h = h;
		evas_gl_make_current(m_gl, m_surface, m_context);
		m_map->resize(m_w, m_h);
		setupOpenGlSurface(view);
	}

	if (m_zoom != zoom) {
		m_zoom = zoom;
		m_map->setZoom(zoom);
	}

	if (m_angle != angle) {
		m_angle = angle;
		m_map->setRotation(degrees_to_radians(angle));
	}

	double lng = 0, lat = 0;
	maps_coordinates_get_longitude(coord, &lng);
	maps_coordinates_get_latitude(coord, &lat);

	if (m_lng != lng || m_lat != lat) {
		m_lng = lng;
		m_lat = lat;
		m_map->setPosition(m_lng, m_lat);
	}

	m_map->getPlatform()->requestRender();

	return MAPZEN_ERROR_NONE;
}

mapzen_error_e TangramView::moveCenter(maps_view_h view, int delta_x, int delta_y)
{
	if (!view) {
		return MAPZEN_ERROR_INVALID_PARAMETER;
	}

	if (!m_isInitialized) {
		return MAPZEN_ERROR_SERVICE_NOT_AVAILABLE;
	}

	if (delta_x == 0 && delta_y == 0) {
		return MAPZEN_ERROR_NONE;
	}

	// The delta_x and delta_y values are in pixels, so we need to determine the equivalent displacement in
	// longitude and latitude to set the new center position.

	double x = 0.5 * m_w + (double)delta_x;
	double y = 0.5 * m_h + (double)delta_y;
	double lng = 0.0, lat = 0.0;
	if (m_map->screenPositionToLngLat(x, y, &lng, &lat)) {
		lng = wrap_longitude(lng);
		lat = wrap_latitude(lat);
		m_map->setPosition(lng, lat);
	}

	MAPS_LOGD("Moved with delta x: %d, y: %d to coordinates lon: %f, lat: %f", delta_x, delta_y, lng, lat);

	return MAPZEN_ERROR_NONE;
}

mapzen_error_e TangramView::getCenter(maps_view_h view, maps_coordinates_h *center)
{
	if (!view || !center) {
		return MAPZEN_ERROR_INVALID_PARAMETER;
	}

	if (!m_isInitialized) {
		return MAPZEN_ERROR_SERVICE_NOT_AVAILABLE;
	}

	double longitude = 0, latitude = 0;
	m_map->getPosition(longitude, latitude);

	longitude = wrap_longitude(longitude);
	latitude = wrap_latitude(latitude);
	if (*center == nullptr) {
		maps_coordinates_create(latitude, longitude, center);
	} else {
		maps_coordinates_set_latitude(*center, latitude);
		maps_coordinates_set_longitude(*center, longitude);
	}

	return MAPZEN_ERROR_NONE;
}

mapzen_error_e TangramView::setScalebarEnabled(maps_view_h view, bool enable)
{
	return MAPZEN_ERROR_SERVICE_NOT_AVAILABLE;
}

mapzen_error_e TangramView::getScalebarEnabled(maps_view_h view, bool *enabled)
{
	return MAPZEN_ERROR_SERVICE_NOT_AVAILABLE;
}

mapzen_error_e TangramView::convertScreenToGeolocation(maps_view_h view, int x, int y, maps_coordinates_h *coord)
{
	if (!view || !coord) {
		return MAPZEN_ERROR_INVALID_PARAMETER;
	}

	if (!m_isInitialized) {
		return MAPZEN_ERROR_SERVICE_NOT_AVAILABLE;
	}

	double longitude = (double)x, latitude = (double)y;
	m_map->screenPositionToLngLat(x, y, &longitude, &latitude);

	longitude = wrap_longitude(longitude);
	latitude = wrap_latitude(latitude);
	if (*coord == nullptr) {
		maps_coordinates_create(latitude, longitude, coord);
	} else {
		maps_coordinates_set_latitude(*coord, latitude);
		maps_coordinates_set_longitude(*coord, longitude);
	}

	return MAPZEN_ERROR_NONE;
}

mapzen_error_e TangramView::convertGeolocationToScreen(maps_view_h view, const maps_coordinates_h coord, int *x, int *y)
{
	if (!view || !coord) {
		return MAPZEN_ERROR_INVALID_PARAMETER;
	}

	if (!m_isInitialized) {
		return MAPZEN_ERROR_SERVICE_NOT_AVAILABLE;
	}

	double lng, lat;
	maps_coordinates_get_latitude(coord, &lat);
	maps_coordinates_get_longitude(coord, &lng);

	double screenx, screeny;
	m_map->lngLatToScreenPosition(lng, lat, &screenx, &screeny);

	*x = (int)screenx;
	*y = (int)screeny;

	return MAPZEN_ERROR_NONE;
}

mapzen_error_e TangramView::getMinZoomLevel(maps_view_h view, int *zoom)
{
	*zoom = 0;
	return MAPZEN_ERROR_NONE;
}

mapzen_error_e TangramView::getMaxZoomLevel(maps_view_h view, int *zoom)
{
	*zoom = 20;
	return MAPZEN_ERROR_NONE;
}

mapzen_error_e TangramView::onViewObject(maps_view_h view, const maps_view_object_h object, maps_view_object_operation_e operation)
{
	if (!view || !object || operation < MAPS_VIEW_OBJECT_ADD || operation > MAPS_VIEW_OBJECT_REMOVE) {
		return MAPZEN_ERROR_INVALID_PARAMETER;
	}

	if (!m_isInitialized || !m_map) {
		return MAPZEN_ERROR_SERVICE_NOT_AVAILABLE;
	}

	mapzen_error_e error = MAPZEN_ERROR_NONE;

	TangramView *tv = nullptr;
	int maps_error = maps_view_get_maps_plugin_view_handle(view, (void**)&tv);
	if (maps_error != MAPS_ERROR_NONE || !tv) { return MAPZEN_ERROR_INVALID_PARAMETER; }

	maps_view_object_type_e type = MAPS_VIEW_OBJECT_POLYLINE;
	maps_view_object_get_type(object, &type);

	if (type < MAPS_VIEW_OBJECT_POLYLINE || type > MAPS_VIEW_OBJECT_MARKER) { return MAPZEN_ERROR_INVALID_PARAMETER; }

	auto& mapViewObjs = tv->mapViewObjects();
	auto& mutex = tv->viewObjectMutex();
	MapViewObjects::iterator iter;
	MapViewObjects::iterator end;
	{
		std::lock_guard<std::mutex> lock(mutex);
		iter = mapViewObjs.find(object);
		end = mapViewObjs.end();
	}

	switch(operation) {
		case MAPS_VIEW_OBJECT_ADD:
			error = tv->addObject(object);
			break;
		case MAPS_VIEW_OBJECT_SET_VISIBLE:
			if (iter != end) {
				error = tv->setObjectVisible(object, iter->second);
			}
			break;
		case MAPS_VIEW_OBJECT_CHANGE:
			if (iter != end) {
				error = tv->updateObject(object, iter->second);
			}
			break;
		case MAPS_VIEW_OBJECT_REMOVE:
			error = tv->removeObject(object);
			break;
		default: break;
	}

	const char *oper_str[20] = { "ADD", "SET_VISIBLE", "MOVE", "CHANGE", "REMOVE"};
	const char *type_str[20] = { "POLYLINE", "POLYGON", "MARKER", "UNKNOWN"};

	MAPS_LOGD("Done Processing View Object: type=%s, operation=%s, object=%p",
		(type >= MAPS_VIEW_OBJECT_POLYLINE && type <= MAPS_VIEW_OBJECT_MARKER) ? type_str[type] : "?",
		(operation >= MAPS_VIEW_OBJECT_ADD && operation <= MAPS_VIEW_OBJECT_REMOVE) ? oper_str[operation] : "?",
		object);

	if (error != MAPZEN_ERROR_NONE) {
		MAPS_LOGD("Something went wrong in processing this ViewObject operation");
	}

	return error;
}

mapzen_error_e TangramView::addObject(maps_view_object_h object)
{
	if (!object || !m_map) { return MAPZEN_ERROR_INVALID_PARAMETER; }

	mapzen_error_e error = MAPZEN_ERROR_NONE;
	Tangram::MarkerID tvMarker = 0;
	maps_view_object_type_e type = MAPS_VIEW_OBJECT_POLYLINE;

	maps_view_object_get_type(object, &type);

	MapViewObjects::iterator iter;
	MapViewObjects::iterator end;

	{
		std::lock_guard<std::mutex> lock(m_viewObjectMutex);
		iter = m_mapViewObjs.find(object);
		end = m_mapViewObjs.end();
	}
	if (iter == end) {
		tvMarker = m_map->markerAdd();
		if (tvMarker) {
			error = updateObject(object, tvMarker);
			if (error == MAPZEN_ERROR_NONE) {
				std::lock_guard<std::mutex> lock(m_viewObjectMutex);
				m_mapViewObjs.insert(std::make_pair(object, tvMarker));
			} else {
				m_map->markerRemove(tvMarker);
				tvMarker = 0;
			}
		}
	}

	return error;
}

mapzen_error_e TangramView::setObjectVisible(maps_view_object_h object, Tangram::MarkerID tvMarker)
{
	if (!object || !tvMarker || !m_map) { return MAPZEN_ERROR_INVALID_PARAMETER; }

	bool visible = true;
	maps_view_object_get_visible(object, &visible);

	m_map->markerSetVisible(tvMarker, visible);
	return MAPZEN_ERROR_NONE;
}

mapzen_error_e TangramView::updateObject(maps_view_object_h object, Tangram::MarkerID tvMarker)
{
	if (!object || !tvMarker || !m_map) { return MAPZEN_ERROR_INVALID_PARAMETER; }

	mapzen_error_e error = MAPZEN_ERROR_NONE;
	maps_view_object_type_e type = MAPS_VIEW_OBJECT_POLYLINE;

	maps_view_object_get_type(object, &type);

	switch(type) {
		case MAPS_VIEW_OBJECT_MARKER:
			error = updateMarker(object, tvMarker);
			break;
		case MAPS_VIEW_OBJECT_POLYLINE:
			error = updatePolyline(object, tvMarker);
			break;
		case MAPS_VIEW_OBJECT_POLYGON:
			error = updatePolygon(object, tvMarker);
			break;
		default:
			break;
	}

	return error;
}

mapzen_error_e TangramView::getBitmapMarkerImage(maps_view_object_h object, unsigned char *&imgData, int &imgWidth, int &imgHeight) {
	char *imgPath = nullptr;
	Evas_Object *img = nullptr;
	int imgSize = 0;
	int error = MAPS_ERROR_NONE;

	do {
		error = maps_view_object_marker_get_image_file(object, &imgPath);
		if (error != MAPS_ERROR_NONE || !imgPath) { break; }

		img = evas_object_image_add(evas_object_evas_get(m_image));
		evas_object_image_file_set(img, imgPath, nullptr);
		int err = evas_object_image_load_error_get(img);
		if (err != EVAS_LOAD_ERROR_NONE) {
			MAPS_LOGE("Failed to load marker image file: %s", imgPath);
			free(imgPath);
			break;
		}
		free(imgPath);

		evas_object_image_size_get(img, &imgWidth, &imgHeight);
		imgSize = imgWidth * imgHeight * 4;

		// Get Raw data pointer to the image
		const unsigned char *srcData = (unsigned char *)evas_object_image_data_get(img, EINA_FALSE);
		if (!srcData || imgSize <= 0) {
			MAPS_LOGE("Failed to get image data from the evas image for the marker");
			error = MAPS_ERROR_OUT_OF_MEMORY;
			break;
		}

		imgData = (unsigned char *)malloc(imgSize);
		if (!imgData) {
			MAPS_LOGE("Failed to malloc!! for image data");
			error = MAPS_ERROR_OUT_OF_MEMORY;
			break;
		}

		// convert data from rgba to bgra (also flip the image)
		for (int i = 0; i < imgHeight; i++) {
			for (int j = 0; j < imgWidth; j++) {
				int offset = ((i * imgWidth) + j) * 4;
				int offset_flip = (((imgHeight - 1 - i) * imgWidth) + j) * 4;
				*(imgData + offset_flip)	= *(srcData + offset + 2);
				*(imgData + offset_flip + 1)	= *(srcData + offset + 1);
				*(imgData + offset_flip + 2)	= *(srcData + offset);
				*(imgData + offset_flip + 3)	= *(srcData + offset + 3);
			}
		}

	} while (0);

	return (mapzen_error_e)convert_maps_error_to_mapzen_error(error);
}

mapzen_error_e TangramView::updateMarker(maps_view_object_h object, Tangram::MarkerID tvMarker)
{
	maps_coordinates_h mapsCoord = nullptr;
	double lat = 0.0, lng = 0.0;
	int markerWidth = 0, markerHeight = 0;
	int imgWidth = 0, imgHeight = 0;
	int error = MAPS_ERROR_NONE;
	unsigned char *imgData = nullptr;

	const char styleFormat[] = "{ style: 'sdk-point-overlay', color: white, size: [%dpx, %dpx], collide: false, anchor: %s, transition: { [show, hide]: { time: 0s } } }";

	static std::string anchor;

	maps_view_marker_type_e type;
	error = maps_view_object_marker_get_type(object, &type);

	do {
		if (error != MAPS_ERROR_NONE || type < MAPS_VIEW_MARKER_PIN || type > MAPS_VIEW_MARKER_STICKER) { break; }

		error = getBitmapMarkerImage(object, imgData, imgWidth, imgHeight);

		if (error != MAPS_ERROR_NONE || !imgData) {
			MAPS_LOGE("Failed getting image data from marker object, with error code: %d", error);
			break;
		}
		m_map->markerSetBitmap(tvMarker, imgWidth, imgHeight, reinterpret_cast<unsigned int *>(imgData));
		free(imgData);

		anchor = "center";
		if (type == MAPS_VIEW_MARKER_PIN) {
			anchor = "top";
		}

		float scaleFactor = 1.0/elm_config_scale_get();
		error = maps_view_object_marker_get_size(object, &markerWidth, &markerHeight);
		if (error != MAPS_ERROR_NONE) { break; }

		/*
		 * - Sets the marker dimension to be image dimension if both the specified marker dimensions are zero
		 * - Sets the marker dimension such that aspect ratio is maintained when one of the marker dimensions is
		 *   zero
		 */
		if (markerWidth == 0 && markerHeight == 0) {
			markerWidth = imgWidth;
			markerHeight = imgHeight;
		} else if (markerWidth == 0 && markerHeight > 0) {
			markerWidth = (float)markerHeight * (float)imgWidth / (float)imgHeight;
		} else if (markerWidth > 0 && markerHeight == 0) {
			markerHeight = (float)markerWidth * (float)imgHeight / (float)imgWidth;
		}

		float scaledWidth = scaleFactor * markerWidth;
		float scaledHeight = scaleFactor * markerHeight;

		char styleString[256];
		std::snprintf(styleString, sizeof(styleString), styleFormat, int(scaledWidth), int(scaledHeight), anchor.c_str());

		MAPS_LOGD("Marker Style String: %s", styleString);
		m_map->markerSetStylingFromString(tvMarker, styleString);

		error = maps_view_object_marker_get_coordinates(object, &mapsCoord);
		if (error != MAPS_ERROR_NONE) { break; }

		error = maps_coordinates_get_longitude(mapsCoord, &lng);
		if (error != MAPS_ERROR_NONE) {
			maps_coordinates_destroy(mapsCoord);
			break;
		}

		error = maps_coordinates_get_latitude(mapsCoord, &lat);
		if (error != MAPS_ERROR_NONE) {
			maps_coordinates_destroy(mapsCoord);
			break;
		}

		maps_coordinates_destroy(mapsCoord);
		m_map->markerSetPoint(tvMarker, Tangram::LngLat(lng, lat));

		int drawOrder = 0;
		error = maps_view_object_marker_get_z_order(object, &drawOrder);
		if (error != MAPS_ERROR_NONE) { break; }
		m_map->markerSetDrawOrder(tvMarker, drawOrder);
	} while (0);

	return (mapzen_error_e)convert_maps_error_to_mapzen_error(error);
}

mapzen_error_e TangramView::updatePolyline(maps_view_object_h object, Tangram::MarkerID tvMarker)
{
	int error = MAPS_ERROR_NONE;
	Tangram::Coordinates coords = {};
	unsigned char r = 0, g = 0, b = 0, a = 0;
	int width = 0;

	const char styleFormat[] = "{ style: 'ux-route-line-overlay', color: [%d, %d, %d, %d], width: %dpx, order: 500 }";

	do {
		error = maps_view_object_polyline_get_color(object, &r, &g, &b, &a);
		if (error != MAPS_ERROR_NONE) { break; }

		error = maps_view_object_polyline_get_width(object, &width);
		if (error != MAPS_ERROR_NONE) { break; }

		char styleString[256];
		std::snprintf(styleString, sizeof(styleString), styleFormat, r, g, b, a, width);

		MAPS_LOGD("Polyline Style String: %s", styleString);

		m_map->markerSetStylingFromString(tvMarker, styleString);

		error = maps_view_object_polyline_foreach_point(object, emplaceCoord, &coords);
		if (error != MAPS_ERROR_NONE) { break; }

		m_map->markerSetPolyline(tvMarker, coords.data(), static_cast<int>(coords.size()));
	} while (0);

	return (mapzen_error_e)convert_maps_error_to_mapzen_error(error);
}

mapzen_error_e TangramView::updatePolygon(maps_view_object_h object, Tangram::MarkerID tvMarker)
{
	int error = MAPS_ERROR_NONE;
	Tangram::Coordinates coords = {};
	unsigned char r = 0, g = 0, b = 0, a = 0;
	const char styleFormat[] = "{ style: 'polygons', color: [%d, %d, %d, %d] }";

	do {
		error = maps_view_object_polygon_get_fill_color(object, &r, &g, &b, &a);
		if (error != MAPS_ERROR_NONE) { break; }

		char styleString[256];
		std::snprintf(styleString, sizeof(styleString), styleFormat, r, g, b, a);

		MAPS_LOGD("Polygon Style String: %s", styleString);

		m_map->markerSetStylingFromString(tvMarker, styleString);

		error = maps_view_object_polygon_foreach_point(object, emplaceCoord, &coords);
		if (error != MAPS_ERROR_NONE) { break; }

		int count = (int)coords.size();
		m_map->markerSetPolygon(tvMarker, coords.data(), &count, 1);
	} while (0);

	return (mapzen_error_e)convert_maps_error_to_mapzen_error(error);
}

mapzen_error_e TangramView::removeObject(maps_view_object_h object)
{
	if (!object || !m_map) { return MAPZEN_ERROR_INVALID_PARAMETER; }

	mapzen_error_e error = MAPZEN_ERROR_NONE;

	std::lock_guard<std::mutex> lock(m_viewObjectMutex);
	auto iter = m_mapViewObjs.find(object);
	if (iter != m_mapViewObjs.end()) {
		if (iter->second) {
			m_map->markerRemove(iter->second);
		}
		iter = m_mapViewObjs.erase(iter);
		error = MAPZEN_ERROR_NONE;
	}

	return error;
}

void TangramView::removeAllObjects()
{
	std::lock_guard<std::mutex> lock(m_viewObjectMutex);
	m_mapViewObjs.clear();
	m_map->markerRemoveAll();
}

bool TangramView::emplaceCoord(int index, maps_coordinates_h coordinate, void *user_data)
{
	if (!coordinate || !user_data) { return false; }

	int error = 0;
	double lng = 0.0, lat = 0.0;

	error = maps_coordinates_get_longitude(coordinate, &lng);
	if (error != MAPS_ERROR_NONE) {
		MAPS_LOGD("Not able to extract longitude from coordiate");
		return false;
	}
	error = maps_coordinates_get_latitude(coordinate, &lat);
	if (error != MAPS_ERROR_NONE) {
		MAPS_LOGD("Not able to extract longitude from coordiate");
		return false;
	}

	MAPS_LOGD("[%d] %f, %f", index+1, lng, lat);

	Tangram::Coordinates *coords = static_cast<Tangram::Coordinates*>(user_data);
	coords->emplace_back(lng, lat);
	return true;
}

mapzen_error_e TangramView::captureSnapshot(maps_view_h view, void **data, int *width, int *height, maps_view_colorspace_type_e *cs)
{
	if (!view) {
		return MAPZEN_ERROR_INVALID_PARAMETER;
	}

	if (!m_isInitialized) {
		return MAPZEN_ERROR_SERVICE_NOT_AVAILABLE;
	}

	*data = (void*)malloc(m_w * m_h * sizeof(unsigned int));
	if (!*data) {
		return MAPZEN_ERROR_OUT_OF_MEMORY;
	}

	pixelGetCb(view, nullptr);

	unsigned int *pixels = (unsigned int*)malloc(m_w * m_h * sizeof(unsigned int));
	if (!pixels) {
		return MAPZEN_ERROR_OUT_OF_MEMORY;
	}

	m_map->captureSnapshot(pixels);

	for (int i = 0; i < m_h; ++i) {
		memcpy((unsigned int*)*data + m_w * i, pixels + m_w * (m_h - i - 1), m_w * sizeof(unsigned int));
	}

	*cs = MAPS_VIEW_COLORSPACE_RGBA8888;
	*width = m_w;
	*height = m_h;

	free(pixels);
	return MAPZEN_ERROR_NONE;
}


void TangramView::setMapType(maps_view_h view)
{
	if (!view) {
		return;
	}

	std::vector<Tangram::SceneUpdate> sceneUpdates;
	maps_view_type_e map_type;
	maps_view_get_type(view, &map_type);

	tangram_view_type newViewType = (tangram_view_type)convert_maps_view_type_to_tangram_view_type(map_type);

	// If Scene is being reloaded then reset all the scene preferences to their defaults (building/transit/language/providerKey)
	if (newViewType != m_viewType) {
		m_buildingsEnabled = true;
		m_publicTransitEnabled = false;
		m_language = "en";
		m_isProviderKeySet = false;
	}

	bool buildings_enabled = false;
	maps_view_get_buildings_enabled(view, &buildings_enabled);
	if (buildings_enabled != m_buildingsEnabled) {
		m_buildingsEnabled = buildings_enabled;
		if (m_buildingsEnabled) {
			sceneUpdates.push_back( {"global.sdk_building_extrude", "true"} );
		} else {
			sceneUpdates.push_back( {"global.sdk_building_extrude", "false"} );
		}
	}

	bool public_transit_enabled = false;
	maps_view_get_public_transit_enabled(view, &public_transit_enabled);
	if (public_transit_enabled != m_publicTransitEnabled) {
		m_publicTransitEnabled = public_transit_enabled;
		if (m_publicTransitEnabled) {
			sceneUpdates.push_back( {"global.sdk_transit_overlay", "true"} );
		} else {
			sceneUpdates.push_back( {"global.sdk_transit_overlay", "false"} );
		}
	}

	char* language = nullptr;
	maps_view_get_language(view, &language);
	if (m_language != language) {
		m_language = language;
		sceneUpdates.push_back( {"global.ux_language", m_language.substr(0, 2).c_str()} );
	}
	free(language);

	// set provider key for vector tiles
	if (m_providerKey.size() > 0 && !m_isProviderKeySet) {
		// block to hide key in log
		//MAPS_LOGD("Queueing API key update: %s", m_providerKey.c_str());
		sceneUpdates.push_back( {"global.sdk_mapzen_api_key", m_providerKey.c_str()} );
		m_isProviderKeySet = true;
	}

	// Loading a new tangram scene resets the map states/caches.
	if (newViewType != m_viewType) {
		// When the scene is changed, update the 'sdk-point-overlay' style to remove it's texture;
		// this allows us to use this style for markers with custom bitmaps.
		sceneUpdates.push_back( {"styles.sdk-point-overlay.texture", "null"} );

		const char* newSceneFile = NORMAL_SCENE_FILE_PATH;
		switch(newViewType) {
			case TANGRAM_VIEW_TERRAIN:
				newSceneFile = TERRAIN_SCENE_FILE_PATH;
				break;
			case TANGRAM_VIEW_NORMAL:
				newSceneFile = NORMAL_SCENE_FILE_PATH;
				break;
			default:
				return;
		}
		m_viewType = newViewType;
		m_map->loadSceneAsync(newSceneFile, false, sceneLoadedCb, (void*)view, sceneUpdates);
	} else {
		// bool traffic_enabled = false;
		// maps_view_get_traffic_enabled(view, &traffic_enabled);
		if (sceneUpdates.empty()) { return; }
		m_map->queueSceneUpdate(sceneUpdates);
		m_map->applySceneUpdates();
	}
}

void TangramView::sceneLoadedCb(void *data) {
	TangramView *tv = nullptr;
	int maps_error = maps_view_get_maps_plugin_view_handle((maps_view_h)data, (void**)&tv);
	if (maps_error != MAPS_ERROR_NONE || !tv) { return; }

	// only do this for first scene load not for any sceneUpdates
	if (!tv->isSceneLoaded()) {
		if (tv->getWidth() > 1 && tv->getHeight() > 1) {
			readyMapCb(data);
		}
		tv->setSceneLoaded(true);
	}
}

void TangramView::readyMapCb(void *data)
{
	TangramView *tv = nullptr;
	int maps_error = maps_view_get_maps_plugin_view_handle((maps_view_h)data, (void**)&tv);
	if (maps_error != MAPS_ERROR_NONE || !tv) { return; }

	if (tv->m_readyCb) {
		tv->m_readyCb(data);
	}
}

void TangramView::pixelGetCb(void *data, Evas_Object *obj)
{
	TangramView *tv = nullptr;
	int maps_error = maps_view_get_maps_plugin_view_handle(data, (void**)&tv);
	if (maps_error != MAPS_ERROR_NONE || !tv) {
		return;
	}

	tv->m_dirty = false;

	if (!tv->m_gl || !tv->m_surface || !tv->m_context) {
		return;
	}

	evas_gl_make_current(tv->m_gl, tv->m_surface, tv->m_context);
	tv->m_api->glViewport(0, 0, tv->m_w, tv->m_h);
	if (tv->m_map) {
		tv->m_map->update(0);
		tv->m_map->render();
		MAPS_LOGD("pixelGetCb called on TangramView");
	} else {
		MAPS_LOGD("tv->m_map is null, in pixelGelCb");
	}
}
