/*
 * Copyright (c) 2014 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#pragma once

#include <maps_plugin.h>
#include <maps_service.h>
#include <maps_view.h>
#include <maps_view_plugin.h>
#include <Ecore.h>
#include <Ecore_Evas.h>
#include <Evas_GL.h>
#include <map>
#include <memory>
#include <mutex>

#include "mapzen_types.h"
#include <util/types.h>

#include <sys/types.h> // pid_t, gettid()

namespace Tangram {
	class Map;
	class TizenPlatform;
}

typedef std::map<maps_view_object_h, Tangram::MarkerID> MapViewObjects;

class TangramView {

public:
	TangramView();
	~TangramView();

	mapzen_error_e create(maps_view_h view, maps_plugin_map_view_ready_cb callback, const char* providerKey);
	mapzen_error_e destroy(maps_view_h view);
	mapzen_error_e render(maps_view_h view, const maps_coordinates_h coord, double zoom, double angle);
	mapzen_error_e moveCenter(maps_view_h view, int delta_x, int delta_y);
	mapzen_error_e getCenter(maps_view_h view, maps_coordinates_h *center);
	mapzen_error_e setScalebarEnabled(maps_view_h view, bool enable);
	mapzen_error_e getScalebarEnabled(maps_view_h view, bool *enabled);
	mapzen_error_e convertScreenToGeolocation(maps_view_h view, int x, int y, maps_coordinates_h *coord);
	mapzen_error_e convertGeolocationToScreen(maps_view_h view, const maps_coordinates_h coord, int *x, int *y);
	mapzen_error_e getMinZoomLevel(maps_view_h view, int *zoom);
	mapzen_error_e getMaxZoomLevel(maps_view_h view, int *zoom);
	mapzen_error_e onViewObject(maps_view_h view, const maps_view_object_h object, maps_view_object_operation_e operation);
	mapzen_error_e captureSnapshot(maps_view_h view, void **data, int *width, int *height, maps_view_colorspace_type_e *cs);

	int getWidth() { return m_w; }
	int getHeight() { return m_h; }

	void setSceneLoaded(bool sceneLoaded) { m_sceneLoaded = sceneLoaded; }
	bool isSceneLoaded() { return m_sceneLoaded; }

private:
	mapzen_error_e setupOpenGlSurface(maps_view_h view);
	void setMapType(maps_view_h view);

	mapzen_error_e addObject(maps_view_object_h object);
	mapzen_error_e setObjectVisible(maps_view_object_h object, Tangram::MarkerID tvMarker);
	mapzen_error_e getBitmapMarkerImage(maps_view_object_h object, unsigned char *&imgData, int &imgWidth, int &imgHeight);
	mapzen_error_e updateObject(maps_view_object_h object, Tangram::MarkerID tvMarker);
	mapzen_error_e updateMarker(maps_view_object_h object, Tangram::MarkerID tvMarker);
	mapzen_error_e updatePolyline(maps_view_object_h object, Tangram::MarkerID tvMarker);
	mapzen_error_e updatePolygon(maps_view_object_h object, Tangram::MarkerID tvMarker);
	mapzen_error_e removeObject(maps_view_object_h object);
	MapViewObjects& mapViewObjects() { return m_mapViewObjs; }
	std::mutex& viewObjectMutex() { return m_viewObjectMutex; }
	void removeAllObjects();

	static void sceneLoadedCb(void *data);
	static void readyMapCb(void *view);
	static void pixelGetCb(void *data, Evas_Object *obj);
	// Emplaces the lng/lat from `coordinate` to `user_data` (Tangram::Coordinates)
	static bool emplaceCoord(int index, maps_coordinates_h coordinate, void *user_data);

	std::unique_ptr<Tangram::Map> m_map;
	std::shared_ptr<Tangram::TizenPlatform> m_platform;

	pid_t m_uiThreadId = 0;

	Evas_Object *m_image = nullptr;
	Evas_GL_Context *m_context = nullptr;
	Evas_GL_Surface *m_surface = nullptr;
	Evas_GL_Config *m_config = nullptr;
	Evas_GL *m_gl = nullptr;
	Evas_GL_API *m_api = nullptr;
	Ecore_Timer *m_renderRequestTimer = nullptr;

	bool m_isInitialized = false;
	bool m_publicTransitEnabled = false;
	bool m_buildingsEnabled = true;
	bool m_sceneLoaded = false;
	std::string m_language = "en";
	std::string m_providerKey = "";
	bool m_isProviderKeySet = false;
	bool m_dirty = false;

	int m_x = 0;
	int m_y = 0;
	int m_w = 0;
	int m_h = 0;
	double m_lat = 0.;
	double m_lng = 0.;
	double m_zoom = 0.;
	double m_angle = 0.;

	std::mutex m_viewObjectMutex;

	maps_plugin_map_view_ready_cb m_readyCb = nullptr;
	tangram_view_type m_viewType = TANGRAM_VIEW_NONE;

	MapViewObjects m_mapViewObjs; //current set of map objects
};
