/*
* Copyright (c) 2014 Samsung Electronics Co., Ltd All Rights Reserved
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
* http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "mapzen_plugin.h"
#include "mapzen_plugin_internal.h"
#include "mapzen_api.hpp"
#include "mapzen_util.h"
#include <maps_error.h>
#include <maps_route_plugin.h>
#include <maps_route_segment_plugin.h>
#include <maps_route_maneuver_plugin.h>
#include <maps_place_plugin.h>
#include <maps_place_category.h>
#include <maps_place_image_plugin.h>
#include <maps_place_rating_plugin.h>

#define DEFAULT_NUM_RESULTS	10
#define _PROVIDER_KEY_MAX_SIZE 1024
#define R_EARTH 6378137.0

static const double LATITUDE_RANGE = 85.05113;
static const double LONGITUDE_RANGE = 180.0;

static int __request_id = 0;
static maps_plugin_h __plugin = NULL;
static char __provider_key[_PROVIDER_KEY_MAX_SIZE] = { 0 };

static maps_item_hashtable_h preference_plugin = NULL;

int __maps_service_instance_count = 0;

EXPORT_API int maps_plugin_multi_reverse_geocode(const maps_coordinates_list_h geocode_list,
	const maps_preference_h preference, maps_service_multi_reverse_geocode_cb callback,
	void *user_data, int *request_id)
{
	return  MAPS_ERROR_NOT_SUPPORTED ;

}

EXPORT_API int maps_plugin_create_map_view(maps_view_h hView, maps_plugin_map_view_ready_cb pCbFunc)
{
	int ret = mapzen_create_map_view(hView, pCbFunc, __provider_key);
	return convert_mapzen_error_to_maps_error(ret);
}

EXPORT_API int maps_plugin_destroy_map_view(maps_view_h hView)
{
	int ret = mapzen_destroy_map_view(hView);
	return convert_mapzen_error_to_maps_error(ret);
}

EXPORT_API int maps_plugin_render_map(maps_view_h hView,
	const maps_coordinates_h coordinates, double zoom_factor, double rotation_angle)
{
	int ret = mapzen_render_map(hView, coordinates, zoom_factor, rotation_angle);
	return convert_mapzen_error_to_maps_error(ret);
}

EXPORT_API int maps_plugin_move_center(maps_view_h hView, int delta_x, int delta_y)
{
	int ret = mapzen_move_center(hView, delta_x, delta_y);
	return convert_mapzen_error_to_maps_error(ret);
}

EXPORT_API int maps_plugin_set_scalebar(maps_view_h hView, bool enable)
{

	return MAPS_ERROR_NOT_SUPPORTED;
}

EXPORT_API int maps_plugin_get_scalebar(maps_view_h hView, bool *enabled)
{
	return MAPS_ERROR_NOT_SUPPORTED;
}

EXPORT_API int maps_plugin_on_object(maps_view_h hView, const maps_view_object_h object, maps_view_object_operation_e operation)
{
	int ret = mapzen_on_view_object(hView, object, operation);
	return convert_mapzen_error_to_maps_error(ret);
}

EXPORT_API int maps_plugin_screen_to_geography(maps_view_h hView, int x, int y, maps_coordinates_h *mapsCoord)
{
	int ret = mapzen_screen_to_geography(hView, x, y, mapsCoord);
	return convert_mapzen_error_to_maps_error(ret);
}

EXPORT_API int maps_plugin_geography_to_screen(maps_view_h hView, const maps_coordinates_h mapsCoord, int *x, int *y)
{
	int ret = mapzen_geography_to_screen(hView, mapsCoord, x, y);
	return convert_mapzen_error_to_maps_error(ret);
}

EXPORT_API int maps_plugin_get_min_zoom_level(maps_view_h hView, int *min_zoom_level)
{
	int ret = mapzen_get_min_zoom_level(hView, min_zoom_level);
	return convert_mapzen_error_to_maps_error(ret);
}

EXPORT_API int maps_plugin_get_max_zoom_level(maps_view_h hView, int *max_zoom_level)
{
	int ret = mapzen_get_max_zoom_level(hView, max_zoom_level);
	return convert_mapzen_error_to_maps_error(ret);
}

EXPORT_API int maps_plugin_get_center(maps_view_h view, maps_coordinates_h *center)
{
	int ret = mapzen_get_center(view, center);
	return convert_mapzen_error_to_maps_error(ret);
}

EXPORT_API int maps_plugin_capture_snapshot(maps_view_h view, void **data, int *width, int *height, maps_view_colorspace_type_e *cs)
{
	int ret = mapzen_capture_snapshot(view, data, width, height, cs);
	return convert_mapzen_error_to_maps_error(ret);
}

EXPORT_API int maps_plugin_init_module(maps_plugin_h *plugin, const char* module)
{
	MAPS_LOGD("maps_plugin_init_module");
	if (!plugin)
		return MAPS_ERROR_INVALID_PARAMETER;

	int ret = MAPS_ERROR_NONE;
	if (!__plugin)
		ret = mapzen_init();

	if (ret == MAPS_ERROR_NONE) {
		__maps_service_instance_count++;
		__plugin = plugin;
	}

	return convert_mapzen_error_to_maps_error(ret);
}

EXPORT_API int maps_plugin_init(maps_plugin_h *plugin)
{
	MAPS_LOGD("maps_plugin_init");
	if (!plugin)
		return MAPS_ERROR_INVALID_PARAMETER;

	int ret = MAPS_ERROR_NONE;
	if (!__plugin)
		ret = mapzen_init();

	if (ret == MAPS_ERROR_NONE) {
		__maps_service_instance_count++;
		__plugin = plugin;
	}

	return convert_mapzen_error_to_maps_error(ret);
}

EXPORT_API int maps_plugin_shutdown(maps_plugin_h plugin)
{
	MAPS_LOGD("PLUGIN SHUTDOWN");
	if (!plugin)
		return MAPS_ERROR_INVALID_PARAMETER;

	if (__maps_service_instance_count == 0) {
		return MAPS_ERROR_INVALID_PARAMETER;
	}

	__maps_service_instance_count--;

	int ret = MAPS_ERROR_NONE;
	if (__maps_service_instance_count == 0) {
		ret = mapzen_shutdown();
		__plugin = NULL;
	}

	return convert_mapzen_error_to_maps_error(ret);
}

EXPORT_API int maps_plugin_is_service_supported(maps_service_e service, bool *supported)
{

	if (!supported)
		return MAPS_ERROR_INVALID_PARAMETER;

	switch (service) {
	case MAPS_SERVICE_GEOCODE:
	case MAPS_SERVICE_GEOCODE_INSIDE_AREA:
	case MAPS_SERVICE_GEOCODE_BY_STRUCTURED_ADDRESS:
	case MAPS_SERVICE_REVERSE_GEOCODE:
	case MAPS_SERVICE_SEARCH_PLACE:
	case MAPS_SERVICE_SEARCH_PLACE_LIST:
	case MAPS_SERVICE_SEARCH_PLACE_BY_AREA:
	case MAPS_SERVICE_SEARCH_PLACE_BY_ADDRESS:
	case MAPS_SERVICE_SEARCH_ROUTE:
	case MAPS_SERVICE_SEARCH_ROUTE_WAYPOINTS:
	case MAPS_SERVICE_CANCEL_REQUEST:
	case MAPS_SERVICE_VIEW:
	case MAPS_SERVICE_VIEW_SNAPSHOT:
		*supported = true;
		break;
	default:
		*supported = false;
		break;
	}
	return MAPS_ERROR_NONE;
}

EXPORT_API int maps_plugin_is_data_supported(maps_service_data_e data, bool *supported)
{
	if (!supported)
		return MAPS_ERROR_INVALID_PARAMETER;

	switch (data) {
	case MAPS_PLACE_ADDRESS:
	case MAPS_PLACE_CATEGORIES:
	case MAPS_PLACE_IMAGE:
		/* unsupported */
		/* case MAPS_PLACE_RATING: */
		/* case MAPS_PLACE_ATTRIBUTES: */
		/* case MAPS_PLACE_CONTACTS: */
		/* case MAPS_PLACE_EDITORIALS: */
		/* case MAPS_PLACE_REVIEWS: */
		/* case MAPS_PLACE_SUPPLIER: */
		/* case MAPS_PLACE_RELATED: */

	case MAPS_ROUTE_PATH:
	case MAPS_ROUTE_SEGMENTS_PATH:
	case MAPS_ROUTE_SEGMENTS_MANEUVERS:
		*supported = true;
		break;
	default:
		*supported = false;
		break;
	}
	return MAPS_ERROR_NONE;
}

EXPORT_API int maps_plugin_get_info(maps_plugin_info_h *info)
{
	if (!info)
		return MAPS_ERROR_INVALID_PARAMETER;

	maps_plugin_info_create(info);
	maps_plugin_info_set_provider_name(*info, "MAPZEN");

	return MAPS_ERROR_NONE;
}

EXPORT_API int maps_plugin_set_provider_key(const char *provider_key)
{
	if (!provider_key)
		return MAPS_ERROR_INVALID_PARAMETER;

	g_snprintf(__provider_key, _PROVIDER_KEY_MAX_SIZE, "%s", provider_key);

	return MAPS_ERROR_NONE;
}

EXPORT_API int maps_plugin_get_provider_key(char **provider_key)
{
	if (!provider_key)
		return MAPS_ERROR_INVALID_PARAMETER;

	*provider_key = g_strndup(__provider_key, _PROVIDER_KEY_MAX_SIZE);

	return MAPS_ERROR_NONE;
}

EXPORT_API int maps_plugin_set_preference(maps_item_hashtable_h preference)
{
	if (!preference)
		return MAPS_ERROR_INVALID_PARAMETER;

	maps_item_hashtable_clone(preference, &preference_plugin);
	return MAPS_ERROR_NONE;
}

EXPORT_API int maps_plugin_get_preference(maps_item_hashtable_h *preference)
{
	if (!preference)
		return MAPS_ERROR_INVALID_PARAMETER;

	maps_item_hashtable_clone(preference_plugin, preference);
	return MAPS_ERROR_NONE;
}

static void __mapzen_geocode_cb(mapzen_error_e result, int request_id, GList *co_ordinates, void *user_data)
{
	MAPS_LOGD("Got GEOCODE callback from ENGINE");

	callback_info_geocode *calldata_geocode = (callback_info_geocode *) user_data;

	if ((result != MAPZEN_ERROR_NONE) || (co_ordinates == NULL)) {
		MAPS_LOGD(">>>>> Invalid GEOCODE result <<<<<");
		calldata_geocode->callback((maps_error_e)convert_mapzen_error_to_maps_error(result), calldata_geocode->reqID, 0, 0, NULL, calldata_geocode->data);
	} else {
		int total_count = (int) g_list_length(co_ordinates);
		int index = 0;

		GList *coords = NULL;
		coords = g_list_first(co_ordinates);

		while (coords) {
			MAPS_LOGD("coordinate %d", index);
			coords_s *data = (coords_s *) coords->data;

			if (data != NULL) {
				maps_coordinates_h resultCoords;
				maps_coordinates_create(data->latitude, data->longitude, &resultCoords);
				bool b = calldata_geocode->callback(MAPS_ERROR_NONE, calldata_geocode->reqID, index, total_count, resultCoords, calldata_geocode->data);
				if (!b)
					return;
			}
			index++;
			coords = coords->next;
		}
	}
}

EXPORT_API int maps_plugin_geocode(const char *address, const maps_preference_h preference, maps_service_geocode_cb callback, void *user_data, int *request_id)
{
	if (!address || !callback || !request_id)
		return MAPS_ERROR_INVALID_PARAMETER;

	callback_info_geocode *calldata_geocode = (callback_info_geocode *)g_malloc0(sizeof(callback_info_geocode));
	if (calldata_geocode == NULL)
		MAPS_PRINT_ERROR_CODE_RETURN(MAPS_ERROR_OUT_OF_MEMORY);

	calldata_geocode->callback = callback;
	calldata_geocode->data = user_data;

	mapzen_geocode_req_s *geocode_req = (mapzen_geocode_req_s *)g_malloc0(sizeof(mapzen_geocode_req_s));
	if (geocode_req == NULL) {
		g_free(calldata_geocode);
		calldata_geocode = NULL;
		MAPS_PRINT_ERROR_CODE_RETURN(MAPS_ERROR_OUT_OF_MEMORY);
	}

	geocode_req->address = g_strdup((gchar *) address);
	geocode_req->maps_key = g_strdup((gchar *) __provider_key);
	geocode_req->boundary = NULL;

	int max_result = 0;
	maps_preference_get_max_results(preference, &max_result);

	if (max_result <= 0)
		geocode_req->num_res = DEFAULT_NUM_RESULTS;
	else
		geocode_req->num_res = max_result;

	*request_id = ++__request_id;
	calldata_geocode->reqID = __request_id;

	int ret = mapzen_geocode(geocode_req, __mapzen_geocode_cb, __request_id, (void *) calldata_geocode);

	return convert_mapzen_error_to_maps_error(ret);
}

EXPORT_API int maps_plugin_geocode_inside_area(const char *address, const maps_area_h bounds, const maps_preference_h preference, maps_service_geocode_cb callback, void *user_data, int *request_id)
{
	if (!bounds || !address || !callback || !request_id)
		return MAPS_ERROR_INVALID_PARAMETER;

	callback_info_geocode *calldata_geocode = (callback_info_geocode *)g_malloc0(sizeof(callback_info_geocode));
	if (calldata_geocode == NULL)
		MAPS_PRINT_ERROR_CODE_RETURN(MAPS_ERROR_OUT_OF_MEMORY);

	calldata_geocode->callback = callback;
	calldata_geocode->data = user_data;

	mapzen_geocode_req_s *geocode_req = (mapzen_geocode_req_s *)g_malloc0(sizeof(mapzen_geocode_req_s));
	if (geocode_req == NULL) {
		free(calldata_geocode);
		MAPS_PRINT_ERROR_CODE_RETURN(MAPS_ERROR_OUT_OF_MEMORY);
	}

	geocode_req->address = g_strdup((gchar *) address);
	geocode_req->maps_key = g_strdup((gchar *) __provider_key);

	int max_result = 0;
	maps_preference_get_max_results(preference, &max_result);

	if (max_result <= 0)
		geocode_req->num_res = DEFAULT_NUM_RESULTS;
	else
		geocode_req->num_res = max_result;

	geocode_req->boundary = NULL;
	geocode_req->boundary = (mapzen_boundary_s *)g_malloc0(sizeof(mapzen_boundary_s));

	maps_area_s *area = (maps_area_s *) bounds;

	if (area && geocode_req->boundary) {
		if (area->type == MAPS_AREA_RECTANGLE) {
			geocode_req->boundary->type = MAPZEN_BOUNDARY_RECT;
			geocode_req->boundary->rect.top_left.latitude = area->rect.top_left.latitude;
			geocode_req->boundary->rect.top_left.longitude = area->rect.top_left.longitude;
			geocode_req->boundary->rect.bottom_right.latitude = area->rect.bottom_right.latitude;
			geocode_req->boundary->rect.bottom_right.longitude = area->rect.bottom_right.longitude;
		} else if (area->type == MAPS_AREA_CIRCLE) {
			geocode_req->boundary->type = MAPZEN_BOUNDARY_CIRCLE;
			geocode_req->boundary->circle.center.latitude = area->circle.center.latitude;
			geocode_req->boundary->circle.center.longitude = area->circle.center.longitude;
			geocode_req->boundary->circle.radius = area->circle.radius;
		}
	}

	*request_id = ++__request_id;
	calldata_geocode->reqID = __request_id;

	int ret = mapzen_geocode(geocode_req, __mapzen_geocode_cb, __request_id, (void *) calldata_geocode);

	return convert_mapzen_error_to_maps_error(ret);
}

EXPORT_API int maps_plugin_geocode_by_structured_address(const maps_address_h address, const maps_preference_h preference, maps_service_geocode_cb callback, void *user_data, int *request_id)
{
	if (!address || !callback || !request_id)
		return MAPS_ERROR_INVALID_PARAMETER;

	callback_info_geocode *calldata_geocode = (callback_info_geocode *)g_malloc0(sizeof(callback_info_geocode));
	if (calldata_geocode == NULL)
		MAPS_PRINT_ERROR_CODE_RETURN(MAPS_ERROR_OUT_OF_MEMORY);

	calldata_geocode->callback = callback;
	calldata_geocode->data = user_data;

	mapzen_geocode_req_s *geocode_req = (mapzen_geocode_req_s *)g_malloc0(sizeof(mapzen_geocode_req_s));
	if (geocode_req == NULL) {
		free(calldata_geocode);
		MAPS_PRINT_ERROR_CODE_RETURN(MAPS_ERROR_OUT_OF_MEMORY);
	}

	char resultAddressQuery[1024] = "";

	char *house_number = NULL;
	maps_address_get_building_number(address, &house_number);

	char *street = NULL;
	maps_address_get_street(address, &street);
	if (street != NULL) {
		if (house_number != NULL) {
			STRCAT(resultAddressQuery, house_number);
			STRCAT(resultAddressQuery, " ");
		}
		STRCAT(resultAddressQuery, street);
	}

	char *city = NULL;
	maps_address_get_city(address, &city);
	if ((strlen(resultAddressQuery) > 0) && (city != NULL)) {
		STRCAT(resultAddressQuery, ", ");
		STRCAT(resultAddressQuery, city);
	} else if (city != NULL) {
		STRCAT(resultAddressQuery, city);
	}

	char *county = NULL;
	maps_address_get_county(address, &county);
	if ((strlen(resultAddressQuery) > 0) && (county != NULL)) {
		STRCAT(resultAddressQuery, ", ");
		STRCAT(resultAddressQuery, county);
	}  else if (county != NULL) {
		STRCAT(resultAddressQuery, county);
	}

	char *state = NULL;
	maps_address_get_state(address, &state);
	if ((strlen(resultAddressQuery) > 0) && (state != NULL)) {
		STRCAT(resultAddressQuery, ", ");
		STRCAT(resultAddressQuery, state);
	} else if (state != NULL) {
		STRCAT(resultAddressQuery, state);
	}

	char *postal_code = NULL;
	maps_address_get_postal_code(address, &postal_code);
	if ((strlen(resultAddressQuery) > 0) && (postal_code != NULL)) {
		STRCAT(resultAddressQuery, ", ");
		STRCAT(resultAddressQuery, postal_code);
	} else if (postal_code != NULL) {
		STRCAT(resultAddressQuery, postal_code);
	}

	char *country = NULL;
	maps_address_get_country(address, &country);
	if ((strlen(resultAddressQuery) > 0) && (country != NULL)) {
		STRCAT(resultAddressQuery, ", ");
		STRCAT(resultAddressQuery, country);
	} else if (country != NULL) {
		STRCAT(resultAddressQuery, country);
	}

	if (country == NULL) {
		char *country_code = NULL;
		maps_address_get_country_code(address, &country_code);
		if ((strlen(resultAddressQuery) > 0) && (country_code != NULL)) {
			STRCAT(resultAddressQuery, ", ");
			STRCAT(resultAddressQuery, country_code);
		} else if (country_code != NULL) {
			STRCAT(resultAddressQuery, country_code);
		}
	}

	geocode_req->address = g_strdup((gchar *) resultAddressQuery);
	geocode_req->maps_key = g_strdup((gchar *) __provider_key);
	geocode_req->boundary = NULL;

	int max_result = 0;
	maps_preference_get_max_results(preference, &max_result);

	if (max_result <= 0)
		geocode_req->num_res = DEFAULT_NUM_RESULTS;
	else
		geocode_req->num_res = max_result;

	*request_id = ++__request_id;
	calldata_geocode->reqID = __request_id;

	int ret = mapzen_geocode(geocode_req, __mapzen_geocode_cb, __request_id, (void *) calldata_geocode);

	return convert_mapzen_error_to_maps_error(ret);
}

static void __mapzen_reverse_geocode_cb(mapzen_error_e result, int request_id, mapzen_address_resp_s *address, void *user_data)
{
	MAPS_LOGD("Got REV GEOCODE callback from ENGINE");
	callback_info_reverse_geocode *calldata_reverse_geocode = (callback_info_reverse_geocode *) user_data;
	if (result != MAPZEN_ERROR_NONE || address == NULL) {
		calldata_reverse_geocode->callback((maps_error_e) convert_mapzen_error_to_maps_error(result), calldata_reverse_geocode->reqID, 0, 0, NULL, calldata_reverse_geocode->data);
	} else {
		int total_count = 1;
		int index = 0;

		maps_address_h addr = NULL;
		maps_address_create(&addr);

		/*TODO: housenumber?? */
		maps_address_set_street(addr, address->street);
		maps_address_set_city(addr, address->localadmin);
		maps_address_set_county(addr, address->county);
		maps_address_set_state(addr, address->region);
		maps_address_set_country(addr, address->country);
		maps_address_set_country_code(addr, address->country_a);
		maps_address_set_postal_code(addr, address->postalcode);

		calldata_reverse_geocode->callback(MAPS_ERROR_NONE, calldata_reverse_geocode->reqID, index, total_count, addr, calldata_reverse_geocode->data);
	}
}

EXPORT_API int maps_plugin_reverse_geocode(double latitude, double longitude, const maps_preference_h preference, maps_service_reverse_geocode_cb callback, void *user_data, int *request_id)
{
	if (!callback || !request_id)
		return MAPS_ERROR_INVALID_PARAMETER;

	if (latitude > LATITUDE_RANGE || latitude < -LATITUDE_RANGE)
		return MAPS_ERROR_INVALID_PARAMETER;

	if (longitude > LONGITUDE_RANGE || longitude < -LONGITUDE_RANGE)
		return MAPS_ERROR_INVALID_PARAMETER;

	callback_info_reverse_geocode *calldata_reverse_geocode = (callback_info_reverse_geocode *)g_malloc0(sizeof(callback_info_reverse_geocode));
	if (calldata_reverse_geocode == NULL)
		MAPS_PRINT_ERROR_CODE_RETURN(MAPS_ERROR_OUT_OF_MEMORY);

	calldata_reverse_geocode->callback = callback;
	calldata_reverse_geocode->data = user_data;

	mapzen_revgeocode_req_s *reverse_geocode_req = (mapzen_revgeocode_req_s *)g_malloc0(sizeof(mapzen_revgeocode_req_s));
	if (reverse_geocode_req == NULL) {
		g_free(calldata_reverse_geocode);
		calldata_reverse_geocode = NULL;
		MAPS_PRINT_ERROR_CODE_RETURN(MAPS_ERROR_OUT_OF_MEMORY);
	}

	reverse_geocode_req->maps_key = g_strdup((gchar *) __provider_key);
	reverse_geocode_req->coordinates.latitude = latitude;
	reverse_geocode_req->coordinates.longitude = longitude;

	*request_id = ++__request_id;
	calldata_reverse_geocode->reqID = __request_id;

	int ret = mapzen_reverse_geocode(reverse_geocode_req, __mapzen_reverse_geocode_cb, __request_id, (void *) calldata_reverse_geocode);

	return convert_mapzen_error_to_maps_error(ret);
}

static void __mapzen_route_cb(mapzen_error_e result, int request_id, mapzen_route_resp_s *route_info, void *user_data)
{
	MAPS_LOGD(">>>>> START __mapzen_route_cb");

	MAPS_LOGD(">>>>> PRE callback_info_route cast");
	callback_info_route *calldata_route = (callback_info_route *) user_data;
	MAPS_LOGD(">>>>> POST callback_info_route cast");

	maps_error_e error;
	if (route_info) {
		MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS route_info");
		maps_route_h tizen_route;
		int route_create_resp = maps_route_create(&tizen_route);
		if (route_create_resp != MAPS_ERROR_NONE)
			MAPS_LOGD(">>>>> __mapzen_route_cb: ERROR maps_route_create");
		else
			MAPS_LOGD(">>>>> __mapzen_route_cb: GOOD maps_route_create");
		if (!tizen_route)
			MAPS_LOGD(">>>>> __mapzen_route_cb: ERROR tizen_route handle is null START");

		/* Set trip origin */
		MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS assign trip origin");
		maps_coordinates_h origin;
		maps_coordinates_create(route_info->origin.latitude, route_info->origin.longitude, &origin);
		error = maps_route_set_origin(tizen_route, origin);
		if (error != MAPS_ERROR_NONE)
			MAPS_LOGD(">>>>> __mapzen_route_cb: ERROR trip maps_route_set_origin");
		maps_coordinates_destroy(origin);

		/* Set trip destination */
		MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS assign trip destination");
		maps_coordinates_h destination;
		maps_coordinates_create(route_info->destination.latitude, route_info->destination.longitude, &destination);
		error = maps_route_set_destination(tizen_route, destination);
		if (error != MAPS_ERROR_NONE)
			MAPS_LOGD(">>>>> __mapzen_route_cb: ERROR trip maps_route_set_destination");
		maps_coordinates_destroy(destination);

		/* Create trip top left point for bounding box */
		MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS trip top_left");
		maps_coordinates_h top_left;
		maps_coordinates_create(route_info->bounding_box.top_left.latitude, route_info->bounding_box.top_left.longitude, &top_left);

		/* Create trip bottom right point for bounding box */
		MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS trip bottom_right");
		maps_coordinates_h bottom_right;
		maps_coordinates_create(route_info->bounding_box.bottom_right.latitude, route_info->bounding_box.bottom_right.longitude, &bottom_right);

		/* Set trip bounding box */
		MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS trip maps_area_create_rectangle");
		maps_area_h bounds = NULL;
		maps_area_create_rectangle(top_left, bottom_right, &bounds);
		MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS trip maps_route_set_bounding_box");
		maps_route_set_bounding_box(tizen_route, bounds);
		maps_area_destroy(bounds);

		/* Clean up trip top left and bottom right points */
		maps_coordinates_destroy(top_left);
		maps_coordinates_destroy(bottom_right);

		/* Set transport mode */
		switch (route_info->type) {
		case COSTING_BICYCLE:
			maps_route_set_transport_mode(tizen_route, MAPS_ROUTE_TRANSPORT_MODE_BICYCLE);
			MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS route_info->type=COSTING_BICYCLE");
			break;
		case COSTING_MULTIMODAL:
			maps_route_set_transport_mode(tizen_route, MAPS_ROUTE_TRANSPORT_MODE_PUBLICTRANSIT);
			MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS route_info->type=COSTING_MULTIMODAL");
			break;
		case COSTING_PEDESTRIAN:
			maps_route_set_transport_mode(tizen_route, MAPS_ROUTE_TRANSPORT_MODE_PEDESTRIAN);
			MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS route_info->type=COSTING_PEDESTRIAN");
			break;
		case COSTING_AUTO:
		case COSTING_AUTO_SHORTER:
		default:
			maps_route_set_transport_mode(tizen_route, MAPS_ROUTE_TRANSPORT_MODE_CAR);
			MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS route_info->type=COSTING_AUTO");
			break;
		}

		/* Set total distance */
		maps_route_set_total_distance(tizen_route, route_info->distance);

		/* Set total duration */
		maps_route_set_total_duration(tizen_route, (long) route_info->time);

		/* Determine distance units */
		maps_distance_unit_e unit = MAPS_DISTANCE_UNIT_KM;
		MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS distance_unit");
		switch (route_info->distance_unit) {
		default:
			unit = MAPS_DISTANCE_UNIT_KM;
			break;
		}

		/* Set distance units */
		maps_route_set_distance_unit(tizen_route, unit);

		/* Setup for segment list */
		GList *segment_data = NULL;
		segment_data = g_list_first(route_info->segments);
		maps_item_list_h tizen_segment_list = NULL;
		maps_item_list_create(&tizen_segment_list);

		/*********************************************************************/
		/* Loop over segments */
		MAPS_LOGD(">>>>> PROCESS __mapzen_route_cb: Loop over segments");
		while (segment_data) {

			MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS create segment");
			maps_route_segment_h tizen_segment = NULL;
			maps_route_segment_create(&tizen_segment);
			mapzen_route_segment *segment = (mapzen_route_segment *) segment_data->data;

			/* Set segment origin */
			MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS assign segment origin");
			maps_coordinates_h origin;
			maps_coordinates_create(segment->origin.latitude, segment->origin.longitude, &origin);
			error = maps_route_segment_set_origin(tizen_segment, origin);
			if (error != MAPS_ERROR_NONE)
				MAPS_LOGD(">>>>> __mapzen_route_cb: ERROR segment maps_route_set_origin");
			maps_coordinates_destroy(origin);

			/* Set segment destination */
			MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS assign segment destination");
			maps_coordinates_h destination;
			maps_coordinates_create(segment->destination.latitude, segment->destination.longitude, &destination);
			error = maps_route_segment_set_destination(tizen_segment, destination);
			if (error != MAPS_ERROR_NONE)
				MAPS_LOGD(">>>>> __mapzen_route_cb: ERROR segment maps_route_set_destination");
			maps_coordinates_destroy(destination);

			/* Create segment top left point for bounding box */
			MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS segment top_left");
			maps_coordinates_h top_left;
			maps_coordinates_create(segment->bounding_box.top_left.latitude, segment->bounding_box.top_left.longitude, &top_left);

			/* Create segment bottom right point for bounding box */
			MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS segment bottom_right");
			maps_coordinates_h bottom_right;
			maps_coordinates_create(segment->bounding_box.bottom_right.latitude, segment->bounding_box.bottom_right.longitude, &bottom_right);

			/* Set segment bounding box */
			MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS segment maps_area_create_rectangle");
			maps_area_h bounds = NULL;
			maps_area_create_rectangle(top_left, bottom_right, &bounds);
			MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS segment maps_route_set_bounding_box");
			maps_route_segment_set_bounding_box(tizen_segment, bounds);
			maps_area_destroy(bounds);

			/* Clean up segment top left and bottom right points */
			maps_coordinates_destroy(top_left);
			maps_coordinates_destroy(bottom_right);

			/* Set segment distance and time */
			MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS assign segment distance and time");
			maps_route_segment_set_distance(tizen_segment, segment->distance);
			maps_route_segment_set_duration(tizen_segment, segment->time);

			/* Set segment path */
			GList *segment_points = NULL;
			segment_points = g_list_first(segment->shapePoints);
			maps_item_list_h tizen_segment_path_list = NULL;
			maps_item_list_create(&tizen_segment_path_list);
			MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS segment_points");
			while (segment_points) {
				coords_s *data = (coords_s *) segment_points->data;
				maps_coordinates_h shapeCoords;
				maps_coordinates_create(data->latitude, data->longitude, &shapeCoords);
				/* MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS segment shape=%f,%f", data->latitude, data->longitude); */
				maps_item_list_append(tizen_segment_path_list, (gpointer) shapeCoords, maps_coordinates_clone);
				maps_coordinates_destroy(shapeCoords);
				segment_points = g_list_next(segment_points);
			}
			maps_route_segment_set_path(tizen_segment, tizen_segment_path_list);
			maps_item_list_destroy(tizen_segment_path_list);

			/* Set up for maneuver list */
			GList *maneuver_data = NULL;
			maneuver_data = g_list_first(segment->maneuvers);
			maps_item_list_h tizen_maneuver_list = NULL;
			maps_item_list_create(&tizen_maneuver_list);

			/*********************************************************************/
			/* Loop over maneuvers */
			MAPS_LOGD(">>>>> PROCESS __mapzen_route_cb: Loop over maneuvers");
			while (maneuver_data) {

				MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS create maneuver");
				maps_route_maneuver_h tizen_maneuver = NULL;
				maps_route_maneuver_create(&tizen_maneuver);
				mapzen_route_maneuver *maneuver = (mapzen_route_maneuver *) maneuver_data->data;

				/* Set maneuver direction */
				maps_route_maneuver_set_direction_id(tizen_maneuver, MAPS_ROUTE_DIRECTION_NONE);

				/* Set turn type from maneuver type */
				MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS assign turn type");
				maps_route_maneuver_set_turn_type(tizen_maneuver, convert_mapzen_maneuver_type_to_maps_turn_type(maneuver->type));

				/* Set maneuver position */
				MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS assign maneuver position");
				maps_coordinates_h coord;
				maps_coordinates_create(maneuver->start_point.latitude, maneuver->start_point.longitude, &coord);
				maps_route_maneuver_set_position(tizen_maneuver, coord);
				maps_coordinates_destroy(coord);

				/* Set maneuver instruction */
				if (maneuver->instruction) {
					MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS assign instruction: %s", maneuver->instruction);
					maps_route_maneuver_set_instruction_text(tizen_maneuver, (char *) maneuver->instruction);
				}

				/* Set maneuver street name */
				if (maneuver->street_name) {
					MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS assign street names: %s", maneuver->street_name);
					maps_route_maneuver_set_road_name(tizen_maneuver, (char *) maneuver->street_name);
				} else {
					MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS assign street names: NULL");
				}

				/* Set maneuver locale */
				if (route_info->language) {
					MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS assign locale: %s", route_info->language);
					maps_route_maneuver_set_locale(tizen_maneuver, (char *) route_info->language);
				}

				/* Set maneuver distance and time */
				MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS assign maneuver distance and time");
				maps_route_maneuver_set_distance_to_next_instruction(tizen_maneuver, maneuver->distance);
				maps_route_maneuver_set_time_to_next_instruction(tizen_maneuver, maneuver->time);

				/* Append to maneuver list */
				MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS append to maneuver list");
				maps_item_list_append(tizen_maneuver_list, (gpointer) tizen_maneuver, maps_route_maneuver_clone);
				maps_route_maneuver_destroy(tizen_maneuver);

				/* Fetching the next item from Maneuver list */
				MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS Fetching the next item from Maneuver list");
				maneuver_data = g_list_next(maneuver_data);
			} /* Loop over maneuvers */
			/*********************************************************************/

			/* Set maneuver list */
			maps_route_segment_set_maneuvers(tizen_segment, tizen_maneuver_list);
			maps_item_list_destroy(tizen_maneuver_list);

			/* Append to segment list */
			maps_item_list_append(tizen_segment_list, (gpointer) tizen_segment, maps_route_segment_clone);
			maps_route_segment_destroy(tizen_segment);

			/* Fetching the next item from segment list */
			MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS Fetching the next item from segment list");
			segment_data = g_list_next(segment_data);

		} /* Loop over segments */
		/*********************************************************************/

		/* Set segment list */
		maps_route_set_segments(tizen_route, tizen_segment_list);
		maps_item_list_destroy(tizen_segment_list);

		/* Set trip path */
		GList *trip_points = NULL;
		trip_points = g_list_first(route_info->shapePoints);
		maps_item_list_h tizen_route_path_list = NULL;
		maps_item_list_create(&tizen_route_path_list);
		MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS trip_points");
		while (trip_points) {
			coords_s *data = (coords_s *) trip_points->data;
			maps_coordinates_h shapeCoords;
			maps_coordinates_create(data->latitude, data->longitude, &shapeCoords);
			/* MAPS_LOGD(">>>>> __mapzen_route_cb: PROCESS trip shape=%f,%f", data->latitude, data->longitude); */
			maps_item_list_append(tizen_route_path_list, (gpointer) shapeCoords, maps_coordinates_clone);
			maps_coordinates_destroy(shapeCoords);
			trip_points = g_list_next(trip_points);
		}
		maps_route_set_path(tizen_route, tizen_route_path_list);
		maps_item_list_destroy(tizen_route_path_list);

		MAPS_LOGD(">>>>> PRE calldata_route->callback");
		bool b = calldata_route->callback((maps_error_e)convert_mapzen_error_to_maps_error(result), calldata_route->reqID, 0, 1, tizen_route, calldata_route->data);
		MAPS_LOGD(">>>>> POST calldata_route->callback");
		if (!b)
			return;
	} else {
		MAPS_LOGD(">>>>> PRE calldata_route->callback INVALID route_info");
		calldata_route->callback((maps_error_e)convert_mapzen_error_to_maps_error(result), calldata_route->reqID, 0, 0, NULL, calldata_route->data);
		MAPS_LOGD(">>>>> POST calldata_route->callback INVALID route_info");
	}

	MAPS_LOGD(">>>>> END __mapzen_route_cb");
}

EXPORT_API int maps_plugin_search_route(const maps_coordinates_h origin,
		const maps_coordinates_h destination, maps_preference_h preference,
		maps_service_search_route_cb callback, void *user_data, int *request_id) {
	if (!origin || !destination || !callback || !request_id)
		return MAPS_ERROR_INVALID_PARAMETER;

	MAPS_LOGD(">>>>> START maps_plugin_search_route: request_id=%d", *request_id);

	callback_info_route *calldata_route = (callback_info_route *)g_malloc0(sizeof(callback_info_route));
	if (calldata_route == NULL)
		MAPS_PRINT_ERROR_CODE_RETURN(MAPS_ERROR_OUT_OF_MEMORY);

	calldata_route->callback = callback;
	calldata_route->data = user_data;

	mapzen_route_req_s *route_req = (mapzen_route_req_s *)g_malloc0(sizeof(mapzen_route_req_s));
	if (route_req == NULL) {
		g_free(calldata_route);
		calldata_route = NULL;
		MAPS_PRINT_ERROR_CODE_RETURN(MAPS_ERROR_OUT_OF_MEMORY);
	}

	route_req->maps_key = g_strdup((gchar *) __provider_key);

	double origin_lat, origin_lon;
	double dest_lat, dest_lon;

	maps_coordinates_get_latitude(origin, &origin_lat);
	maps_coordinates_get_longitude(origin, &origin_lon);

	maps_coordinates_get_latitude(destination, &dest_lat);
	maps_coordinates_get_longitude(destination, &dest_lon);

	route_req->from.latitude = origin_lat;
	route_req->from.longitude = origin_lon;

	route_req->to.latitude = dest_lat;
	route_req->to.longitude = dest_lon;

	MAPS_LOGD("getting transport mode..");
	maps_route_transport_mode_e transport_mode;
	maps_preference_get_route_transport_mode(preference, &transport_mode);

	switch (transport_mode) {
	case MAPS_ROUTE_TRANSPORT_MODE_CAR:
		route_req->type = COSTING_AUTO;
		break;
	case MAPS_ROUTE_TRANSPORT_MODE_PEDESTRIAN:
		route_req->type = COSTING_PEDESTRIAN;
		break;
	case MAPS_ROUTE_TRANSPORT_MODE_BICYCLE:
		route_req->type = COSTING_BICYCLE;
		break;
	case MAPS_ROUTE_TRANSPORT_MODE_PUBLICTRANSIT:
		route_req->type = COSTING_MULTIMODAL;
		break;
	default:
		route_req->type = COSTING_AUTO;
		break;
	}

	/* Unit */
	maps_distance_unit_e unit;
	maps_preference_get_distance_unit(preference, &unit);

	switch (unit) {
	case MAPS_DISTANCE_UNIT_KM:
		route_req->unit = UNIT_KM;
		break;
	default:
		route_req->unit = UNIT_KM;
		break;
	}

	route_req->avoids = PENALTY_NONE;
	maps_route_feature_weight_e routeWeight;
	maps_preference_get_route_feature_weight(preference, &routeWeight);

	if (routeWeight == MAPS_ROUTE_FEATURE_WEIGHT_AVOID) {
		maps_route_feature_e routeFeature;
		maps_preference_get_route_feature(preference, &routeFeature);

		switch (routeFeature) {
		case MAPS_ROUTE_FEATURE_TOLL:
			route_req->avoids = PENALTY_TOLL_ROADS;
			break;
		case MAPS_ROUTE_FEATURE_MOTORWAY:
			route_req->avoids = PENALTY_LIMITED_ACCESS;
			break;
		case MAPS_ROUTE_FEATURE_BOATFERRY:
		case MAPS_ROUTE_FEATURE_RAILFERRY:
			route_req->avoids = PENALTY_USE_FERRY;
			break;
		case MAPS_ROUTE_FEATURE_DIRTROAD:
			route_req->avoids = PENALTY_USE_UNPAVED;
			break;
		default:
			route_req->avoids = PENALTY_NONE;
			break;
		}
	}

	route_req->way_points = NULL;

	*request_id = ++__request_id;
	calldata_route->reqID = __request_id;
	MAPS_LOGD(">>>>> PROCESS maps_plugin_search_route: calldata_route->reqID=%d", calldata_route->reqID);

	MAPS_LOGD(">>>>> PROCESS maps_plugin_search_route: PRE mapzen_start_route");
	int ret = mapzen_start_route(route_req, __mapzen_route_cb, __request_id, (void *)calldata_route);
	MAPS_LOGD(">>>>> PROCESS maps_plugin_search_route: POST mapzen_start_route");

	MAPS_LOGD(">>>>> END maps_plugin_search_route");
	return convert_mapzen_error_to_maps_error(ret);
}

EXPORT_API int maps_plugin_search_route_waypoints(
		const maps_coordinates_h *waypoint_list, int waypoint_num,
		maps_preference_h preference, maps_service_search_route_cb callback,
		void *user_data, int *request_id) {
	MAPS_LOGD(">>>>> START maps_plugin_search_route_waypoints");
	if (!waypoint_list || waypoint_num < 2 || !callback || !request_id)
		return MAPS_ERROR_INVALID_PARAMETER;

	callback_info_route *calldata_route = (callback_info_route *)g_malloc0(sizeof(callback_info_route));
	if (calldata_route == NULL)
		MAPS_PRINT_ERROR_CODE_RETURN(MAPS_ERROR_OUT_OF_MEMORY);

	calldata_route->callback = callback;
	calldata_route->data = user_data;

	mapzen_route_req_s *route_req = (mapzen_route_req_s *)g_malloc0(sizeof(mapzen_route_req_s));
	if (route_req == NULL) {
		g_free(calldata_route);
		calldata_route = NULL;
		MAPS_PRINT_ERROR_CODE_RETURN(MAPS_ERROR_OUT_OF_MEMORY);
	}

	route_req->maps_key = g_strdup((gchar *) __provider_key);

	route_req->from.latitude = 0.0;
	route_req->from.longitude = 0.0;

	route_req->to.latitude = 0.0;
	route_req->to.longitude = 0.0;

	MAPS_LOGD("getting transport mode..");
	maps_route_transport_mode_e transport_mode;
	maps_preference_get_route_transport_mode(preference, &transport_mode);

	switch (transport_mode) {
	case MAPS_ROUTE_TRANSPORT_MODE_CAR:
		route_req->type = COSTING_AUTO;
		break;
	case MAPS_ROUTE_TRANSPORT_MODE_PEDESTRIAN:
		route_req->type = COSTING_PEDESTRIAN;
		break;
	case MAPS_ROUTE_TRANSPORT_MODE_BICYCLE:
		route_req->type = COSTING_BICYCLE;
		break;
	case MAPS_ROUTE_TRANSPORT_MODE_PUBLICTRANSIT:
		route_req->type = COSTING_MULTIMODAL;
		break;
	default:
		route_req->type = COSTING_AUTO;
		break;
	}

	/* Unit */
	maps_distance_unit_e unit;
	maps_preference_get_distance_unit(preference, &unit);

	switch (unit) {
	case MAPS_DISTANCE_UNIT_KM:
		route_req->unit = UNIT_KM;
		break;
	default:
		route_req->unit = UNIT_KM;
		break;
	}

	route_req->avoids = PENALTY_NONE;
	maps_route_feature_weight_e routeWeight;
	maps_preference_get_route_feature_weight(preference, &routeWeight);

	if (routeWeight == MAPS_ROUTE_FEATURE_WEIGHT_AVOID) {
		maps_route_feature_e routeFeature;
		maps_preference_get_route_feature(preference, &routeFeature);

		switch (routeFeature) {
		case MAPS_ROUTE_FEATURE_TOLL:
			route_req->avoids = PENALTY_TOLL_ROADS;
			break;
		case MAPS_ROUTE_FEATURE_MOTORWAY:
			route_req->avoids = PENALTY_LIMITED_ACCESS;
			break;
		case MAPS_ROUTE_FEATURE_BOATFERRY:
		case MAPS_ROUTE_FEATURE_RAILFERRY:
			route_req->avoids = PENALTY_USE_FERRY;
			break;
		case MAPS_ROUTE_FEATURE_DIRTROAD:
			route_req->avoids = PENALTY_USE_UNPAVED;
			break;
		default:
			route_req->avoids = PENALTY_NONE;
			break;
		}
	}

	/* Waypoints */
	route_req->way_points = NULL;
	if (waypoint_num != 0) {
		int index = 0;
		double latitude = 0.0, longitude = 0.0;
		for (index = 0; index < waypoint_num; index++) {
			if (waypoint_list[index] != NULL) {
				maps_coordinates_get_latitude(waypoint_list[index], &latitude);
				maps_coordinates_get_longitude(waypoint_list[index], &longitude);

				coords_s *data = (coords_s *)g_malloc0(sizeof(coords_s));
				if (data) {
					data->latitude = latitude;
					data->longitude = longitude;

					MAPS_LOGD(">>>>> PROCESS maps_plugin_search_route_waypoints: ll[%d]=%f,%f", index, data->latitude, data->longitude);
					if (route_req->way_points == NULL)
						route_req->way_points = g_list_append(route_req->way_points, (gpointer) data);
					else
						route_req->way_points = g_list_insert_before(route_req->way_points, NULL, (gpointer) data);
				}
			}
		}
	}

	*request_id = ++__request_id;
	calldata_route->reqID = __request_id;

	MAPS_LOGD(">>>>> PROCESS maps_plugin_search_route_waypoints: PRE mapzen_start_route");
	int ret = mapzen_start_route(route_req, __mapzen_route_cb, __request_id, (void *)calldata_route);
	MAPS_LOGD(">>>>> PROCESS maps_plugin_search_route_waypoints: POST mapzen_start_route");

	MAPS_LOGD(">>>>> END maps_plugin_search_route_waypoints");
	return convert_mapzen_error_to_maps_error(ret);
}

static void __mapzen_place_search_cb(mapzen_error_e result, int request_id, GList *places, void *user_data)
{
	MAPS_LOGD("Got places result from ENGINE...");

	callback_info_place *calldata_place = (callback_info_place *) user_data;

	if (result != MAPZEN_ERROR_NONE || places == NULL) {
		MAPS_LOGD("Got places result from ENGINE...result is NULL");
		calldata_place->callback((maps_error_e) convert_mapzen_error_to_maps_error(result), calldata_place->reqID, 0, 0, NULL, calldata_place->data);
	} else {
		guint total_count = 0;
		int index = 0;
		total_count = g_list_length(places);
		if (total_count > 0) {
			maps_place_h place = NULL;
			MAPS_LOGD("Got places result from ENGINE... count -> %d", total_count);

			GList *temp_place = NULL;
			temp_place = g_list_first(places);

			while (temp_place) {
				maps_place_create(&place);
				mapzen_place_resp_s *mapzen_place = (mapzen_place_resp_s *) temp_place->data;
				maps_place_set_id(place, mapzen_place->place_id);
				maps_place_set_name(place, mapzen_place->display_name);

				/* Address */
				if (mapzen_place->address) {
					maps_address_h addr = NULL;
					maps_address_create(&addr);

					maps_address_set_street(addr, mapzen_place->address->street);
					maps_address_set_building_number(addr, mapzen_place->address->housenumber);
					maps_address_set_city(addr, mapzen_place->address->localadmin);
					maps_address_set_county(addr, mapzen_place->address->county);
					maps_address_set_state(addr, mapzen_place->address->region);
					maps_address_set_country(addr, mapzen_place->address->country);
					maps_address_set_country_code(addr, mapzen_place->address->country_a);
					maps_address_set_postal_code(addr, mapzen_place->address->postalcode);
					maps_address_set_freetext(addr, mapzen_place->display_name);

					maps_place_set_address(place, addr);
					maps_address_destroy(addr);
				} else {
					maps_place_set_address(place, NULL);
				}

				maps_coordinates_h coord;
				maps_coordinates_create(mapzen_place->coordinates.latitude, mapzen_place->coordinates.longitude, &coord);

				maps_place_set_location(place, coord);
				maps_coordinates_destroy(coord);

				GList *cats = NULL;
				cats = g_list_first(mapzen_place->categories);

				maps_item_list_h cat_list;
				maps_item_list_create(&cat_list);

				while (cats != NULL) {
					if (cats->data != NULL) {
						maps_place_category_h place_cat;
						maps_place_category_create(&place_cat);
						maps_place_category_set_name(place_cat, cats->data);
						maps_item_list_append(cat_list, (void *) place_cat, maps_place_category_clone);
						maps_place_category_destroy(place_cat);
					}
					cats = g_list_next(cats);
				}

				//if (g_list_length(cat_list) > 0) {
				int length = 0;
				maps_item_list_get_length(cat_list, &length);
				if (length > 0) {
					maps_place_set_categories(place, cat_list);
					maps_item_list_remove_all(cat_list, maps_place_category_destroy);
				}
				maps_item_list_destroy(cat_list);

				/* Don't have URLs yet.... */
				/*maps_place_image_h place_image = NULL; */
				/*maps_place_image_create(&place_image); */
				/*no url at this time. */
				/*maps_place_image_set_url(place_image, mapzen_place->icon_url); */
				/*maps_item_list_h image_list; */
				/*maps_item_list_create(&image_list); */
				/*maps_item_list_append(image_list, (void *) place_image, maps_place_image_clone); */
				/*maps_place_set_images(place, image_list); */

				/*maps_place_image_destroy(place_image); */
				/*maps_item_list_remove_all(image_list, maps_place_image_destroy); */
				/*maps_item_list_destroy(image_list); */

				bool b = calldata_place->callback((maps_error_e)convert_mapzen_error_to_maps_error(result), calldata_place->reqID, index, total_count, place, calldata_place->data);
				if (!b)
					return;

				index++;

				temp_place = temp_place->next;
			}
		} else {
			calldata_place->callback((maps_error_e)convert_mapzen_error_to_maps_error(result), calldata_place->reqID, index, total_count, NULL, calldata_place->data);
		}
	}
}

/*static void __mapzen_place_get_details_cb(mapzen_error_e result, int request_id, GList *places, void *user_data)*/
/*{*/
	/*MAPS_LOGD("Got places details result from ENGINE...");*/

	/*callback_info_place_details *calldata_place = (callback_info_place_details *) user_data;*/

	/*if (result != MAPZEN_ERROR_NONE || places == NULL) {*/
		/*MAPS_LOGD("Got places result from ENGINE...result is NULL");*/
		/*calldata_place->callback((maps_error_e)convert_mapzen_error_to_maps_error(result), calldata_place->reqID, NULL, calldata_place->data);*/
	/*} else {*/
		/*guint total_count = 0;*/
		/*total_count = g_list_length(places);*/
		/*[> Should only have one result <]*/
		/*if (total_count == 1) {*/
			/*maps_place_h place = NULL;*/
			/*MAPS_LOGD("Got places details result from ENGINE... count -> %d", total_count);*/

			/*GList *temp_place = NULL;*/
			/*temp_place = g_list_first(places);*/

			/*if (temp_place) {*/
				/*maps_place_create(&place);*/
				/*mapzen_place_resp_s *mapzen_place = (mapzen_place_resp_s *) temp_place->data;*/
				/*maps_place_set_id(place, mapzen_place->place_id);*/
				/*maps_place_set_name(place, mapzen_place->display_name);*/

				/*[> Address <]*/
				/*if (mapzen_place->address) {*/
					/*maps_address_h addr = NULL;*/
					/*maps_address_create(&addr);*/

					/*maps_address_set_street(addr, mapzen_place->address->street);*/
					/*maps_address_set_building_number(addr, mapzen_place->address->housenumber);*/
					/*maps_address_set_city(addr, mapzen_place->address->localadmin);*/
					/*maps_address_set_county(addr, mapzen_place->address->county);*/
					/*maps_address_set_state(addr, mapzen_place->address->region);*/
					/*maps_address_set_country(addr, mapzen_place->address->country);*/
					/*maps_address_set_country_code(addr, mapzen_place->address->country_a);*/
					/*maps_address_set_postal_code(addr, mapzen_place->address->postalcode);*/
					/*maps_address_set_freetext(addr, mapzen_place->display_name);*/

					/*maps_place_set_address(place, addr);*/
					/*maps_address_destroy(addr);*/
				/*} else {*/
					/*maps_place_set_address(place, NULL);*/
				/*}*/

				/*maps_coordinates_h coord;*/
				/*maps_coordinates_create(mapzen_place->coordinates.latitude, mapzen_place->coordinates.longitude, &coord);*/

				/*maps_place_set_location(place, coord);*/
				/*maps_coordinates_destroy(coord);*/

				/*GList *cats = NULL;*/
				/*cats = g_list_first(mapzen_place->categories);*/

				/*maps_item_list_h cat_list;*/
				/*maps_item_list_create(&cat_list);*/

				/*while (cats != NULL) {*/
					/*if (cats->data != NULL) {*/
						/*maps_place_category_h place_cat;*/
						/*maps_place_category_create(&place_cat);*/
						/*maps_place_category_set_name(place_cat, cats->data);*/
						/*maps_item_list_append(cat_list, (void *) place_cat, maps_place_category_clone);*/
						/*maps_place_category_destroy(place_cat);*/
					/*}*/
					/*cats = g_list_next(cats);*/
				/*}*/

				/*if (g_list_length(cat_list) > 0) {*/
				/*
				int length = 0;
				maps_item_list_get_length(cat_list, &length);
				if (length > 0) {
				*/
					/*maps_place_set_categories(place, cat_list);*/
					/*maps_item_list_remove_all(cat_list, maps_place_category_destroy);*/
				/*}*/
				/*maps_item_list_destroy(cat_list);*/

				/*[> Don't have URLs yet.... <]*/
				/*[>maps_place_image_h place_image = NULL; <]*/
				/*[>maps_place_image_create(&place_image); <]*/
				/*[>no url at this time. <]*/
				/*[>maps_place_image_set_url(place_image, mapzen_place->icon_url); <]*/
				/*[>maps_item_list_h image_list; <]*/
				/*[>maps_item_list_create(&image_list); <]*/
				/*[>maps_item_list_append(image_list, (void *) place_image, maps_place_image_clone); <]*/
				/*[>maps_place_set_images(place, image_list); <]*/

				/*[>maps_place_image_destroy(place_image); <]*/
				/*[>maps_item_list_remove_all(image_list, maps_place_image_destroy); <]*/
				/*[>maps_item_list_destroy(image_list); <]*/

				/*calldata_place->callback((maps_error_e)convert_mapzen_error_to_maps_error(result), calldata_place->reqID, place, calldata_place->data);*/
				/*return;*/
			/*}*/
		/*}*/
	/*}*/
	/*calldata_place->callback((maps_error_e)convert_mapzen_error_to_maps_error(result), calldata_place->reqID, NULL, calldata_place->data);*/
/*}*/

static void __maps_service_search_place_list_cb(mapzen_error_e result, int request_id, GList *places, void *user_data)
{
	MAPS_LOGD("Got places result from ENGINE...");

	callback_info_place_list *calldata_place = (callback_info_place_list *) user_data;

	if (result != MAPZEN_ERROR_NONE || places == NULL) {
		MAPS_LOGD("Got places result from ENGINE...result is NULL");
		calldata_place->callback((maps_error_e) convert_mapzen_error_to_maps_error(result), calldata_place->reqID, 0, NULL, calldata_place->data);
	} else {
		guint total_count = 0;
		total_count = g_list_length(places);
		if (total_count > 0) {
			maps_place_h place = NULL;
			MAPS_LOGD("Got places result from ENGINE... count -> %d", total_count);

			maps_place_list_h place_list = NULL;
			maps_place_list_create(&place_list);

			GList *temp_place = NULL;
			temp_place = g_list_first(places);

			while (temp_place) {
				maps_place_create(&place);
				mapzen_place_resp_s *mapzen_place = (mapzen_place_resp_s *) temp_place->data;
				maps_place_set_id(place, mapzen_place->place_id);
				maps_place_set_name(place, mapzen_place->display_name);

				/* Address */
				if (mapzen_place->address) {
					maps_address_h addr = NULL;
					maps_address_create(&addr);

					maps_address_set_street(addr, mapzen_place->address->street);
					maps_address_set_building_number(addr, mapzen_place->address->housenumber);
					maps_address_set_city(addr, mapzen_place->address->localadmin);
					maps_address_set_county(addr, mapzen_place->address->county);
					maps_address_set_state(addr, mapzen_place->address->region);
					maps_address_set_country(addr, mapzen_place->address->country);
					maps_address_set_country_code(addr, mapzen_place->address->country_a);
					maps_address_set_postal_code(addr, mapzen_place->address->postalcode);
					maps_address_set_freetext(addr, mapzen_place->display_name);

					maps_place_set_address(place, addr);
					maps_address_destroy(addr);
				} else {
					maps_place_set_address(place, NULL);
				}

				maps_coordinates_h coord;
				maps_coordinates_create(mapzen_place->coordinates.latitude, mapzen_place->coordinates.longitude, &coord);

				maps_place_set_location(place, coord);
				maps_coordinates_destroy(coord);

				GList *cats = NULL;
				cats = g_list_first(mapzen_place->categories);

				maps_item_list_h cat_list;
				maps_item_list_create(&cat_list);

				while (cats != NULL) {
					if (cats->data != NULL) {
						maps_place_category_h place_cat;
						maps_place_category_create(&place_cat);
						maps_place_category_set_name(place_cat, cats->data);
						maps_item_list_append(cat_list, (void *) place_cat, maps_place_category_clone);
						maps_place_category_destroy(place_cat);
					}
					cats = g_list_next(cats);
				}

				//if (g_list_length(cat_list) > 0) {
				int length = 0;
				maps_item_list_get_length(cat_list, &length);
				if (length > 0) {
					maps_place_set_categories(place, cat_list);
					maps_item_list_remove_all(cat_list, maps_place_category_destroy);
				}
				maps_item_list_destroy(cat_list);

				/* Don't have URLs yet.... */
				/*maps_place_image_h place_image = NULL; */
				/*maps_place_image_create(&place_image); */
				/*no url at this time. */
				/*maps_place_image_set_url(place_image, mapzen_place->icon_url); */
				/*maps_item_list_h image_list; */
				/*maps_item_list_create(&image_list); */
				/*maps_item_list_append(image_list, (void *) place_image, maps_place_image_clone); */
				/*maps_place_set_images(place, image_list); */

				/*maps_place_image_destroy(place_image); */
				/*maps_item_list_remove_all(image_list, maps_place_image_destroy); */
				/*maps_item_list_destroy(image_list); */

				maps_item_list_append(place_list, place, maps_place_clone);
				maps_place_destroy(place);
				place = NULL;
				temp_place = temp_place->next;
			}
			calldata_place->callback((maps_error_e)convert_mapzen_error_to_maps_error(result), calldata_place->reqID, total_count, place_list, calldata_place->data);

		} else {
			calldata_place->callback((maps_error_e)convert_mapzen_error_to_maps_error(result), calldata_place->reqID, total_count, NULL, calldata_place->data);
		}
	}
}

EXPORT_API int maps_plugin_search_place(const maps_coordinates_h position, int distance, const maps_place_filter_h filter, const maps_preference_h preference, maps_service_search_place_cb callback, void *user_data, int *request_id)
{
	if (!position || !filter || !callback || !request_id)
		return MAPS_ERROR_INVALID_PARAMETER;
	if (distance <= 0)
		return MAPS_ERROR_INVALID_PARAMETER;

	callback_info_place *calldata_place = (callback_info_place *)g_malloc0(sizeof(callback_info_place));
	if (calldata_place == NULL)
		MAPS_PRINT_ERROR_CODE_RETURN(MAPS_ERROR_OUT_OF_MEMORY);

	calldata_place->callback = callback;
	calldata_place->data = user_data;

	mapzen_search_req_s *place_req = (mapzen_search_req_s *)g_malloc0(sizeof(mapzen_search_req_s));
	if (place_req == NULL) {
		g_free(calldata_place);
		calldata_place = NULL;
		MAPS_PRINT_ERROR_CODE_RETURN(MAPS_ERROR_OUT_OF_MEMORY);
	}

	place_req->maps_key = g_strdup((gchar *) __provider_key);

	int max_result;
	maps_preference_get_max_results(preference, &max_result);

	if (max_result == 0)
		place_req->num_res = DEFAULT_NUM_RESULTS;
	else
		place_req->num_res = max_result;

	MAPS_LOGD("Place Result limit :: %d", place_req->num_res);

	char *categoryId = NULL;
	char *searchKeyword = NULL;

	maps_place_category_h category = NULL;
	maps_place_filter_get_category(filter, &category);
	maps_place_filter_get_keyword(filter, &searchKeyword);

	if (category) {
		maps_place_category_get_id(category, &categoryId);
		maps_place_category_destroy(category);
	}

	if (categoryId) {
		if (strlen(categoryId) == 0) {
			place_req->categories = NULL;
		} else {
			place_req->categories = g_strdup_printf("%s", categoryId);
		}
	}

	if (searchKeyword)
		place_req->search_string = g_strdup_printf("%s", searchKeyword);

	if (searchKeyword) {
		g_free(searchKeyword);
		searchKeyword = NULL;
	}

	if (categoryId) {
		g_free(categoryId);
		categoryId = NULL;
	}

	MAPS_LOGD(">>>>>>>> Place search categories :: %s <<<<<<<<<", place_req->categories);

	place_req->boundary = NULL;
	place_req->boundary = (mapzen_boundary_s *)g_malloc0(sizeof(mapzen_boundary_s));

	if (place_req->boundary) {
		place_req->boundary->type = MAPZEN_BOUNDARY_CIRCLE;
		double lat, lon;
		maps_coordinates_get_latitude(position, &lat);
		maps_coordinates_get_longitude(position, &lon);
		place_req->boundary->circle.center.latitude = lat;
		place_req->boundary->circle.center.longitude = lon;
		place_req->boundary->circle.radius = distance;
	}

	*request_id = ++__request_id;
	calldata_place->reqID = __request_id;

	int ret = mapzen_search_place(place_req, __mapzen_place_search_cb, __request_id, (void *) calldata_place);

	return convert_mapzen_error_to_maps_error(ret);
}

EXPORT_API int maps_plugin_search_place_by_area(const maps_area_h boundary, const maps_place_filter_h filter, const maps_preference_h preference, maps_service_search_place_cb callback, void *user_data, int *request_id)
{
	if (!boundary || !filter || !callback || !request_id)
		return MAPS_ERROR_INVALID_PARAMETER;

	callback_info_place *calldata_place = (callback_info_place *)g_malloc0(sizeof(callback_info_place));
	if (calldata_place == NULL)
		MAPS_PRINT_ERROR_CODE_RETURN(MAPS_ERROR_OUT_OF_MEMORY);

	calldata_place->callback = callback;
	calldata_place->data = user_data;

	mapzen_search_req_s *place_req = (mapzen_search_req_s *)g_malloc0(sizeof(mapzen_search_req_s));
	if (place_req == NULL) {
		g_free(calldata_place);
		calldata_place = NULL;
		MAPS_PRINT_ERROR_CODE_RETURN(MAPS_ERROR_OUT_OF_MEMORY);
	}

	place_req->maps_key = g_strdup((gchar *) __provider_key);

	int max_result;
	maps_preference_get_max_results(preference, &max_result);

	if (max_result == 0)
		place_req->num_res = DEFAULT_NUM_RESULTS;
	else
		place_req->num_res = max_result;

	MAPS_LOGD("Place Result limit :: %d", place_req->num_res);

	char *categoryId = NULL;
	char *searchKeyword = NULL;

	maps_place_category_h category = NULL;
	maps_place_filter_get_category(filter, &category);
	maps_place_filter_get_keyword(filter, &searchKeyword);

	if (category) {
		maps_place_category_get_id(category, &categoryId);
		maps_place_category_destroy(category);
	}

	if (categoryId)
		place_req->categories = g_strdup_printf("%s", categoryId);

	if (searchKeyword)
		place_req->search_string = g_strdup_printf("%s", searchKeyword);

	if (searchKeyword) {
		g_free(searchKeyword);
		searchKeyword = NULL;
	}

	if (categoryId) {
		g_free(categoryId);
		categoryId = NULL;
	}

	MAPS_LOGD(">>>>>>>> Place search categories :: %s <<<<<<<<<", place_req->categories);

	place_req->boundary = NULL;
	maps_area_s *bound = (maps_area_s *) boundary;

	if (bound->type != MAPS_AREA_NONE) {
		place_req->boundary = (mapzen_boundary_s *)g_malloc0(sizeof(mapzen_boundary_s));

		if (place_req->boundary != NULL) {
			if (bound->type == MAPS_AREA_CIRCLE) {
				place_req->boundary->type = MAPZEN_BOUNDARY_CIRCLE;
				place_req->boundary->circle.center.latitude = bound->circle.center.latitude;
				place_req->boundary->circle.center.longitude = bound->circle.center.longitude;
				place_req->boundary->circle.radius = bound->circle.radius;
			} else if (bound->type == MAPS_AREA_RECTANGLE) {
				place_req->boundary->type = MAPZEN_BOUNDARY_CIRCLE;
				gdouble centerLat = (bound->rect.top_left.latitude + bound->rect.bottom_right.latitude) * 0.5;
				gdouble centerLon = (bound->rect.top_left.longitude + bound->rect.bottom_right.longitude) * 0.5;
				//aprox radius calculation (Equirectangular approximation) - works great for small distances
				gdouble dLat = degrees_to_radians(bound->rect.top_left.latitude - bound->rect.bottom_right.latitude);
				gdouble dLon = degrees_to_radians(bound->rect.top_left.longitude - bound->rect.bottom_right.longitude);
				gdouble mLat = degrees_to_radians(bound->rect.top_left.latitude + bound->rect.bottom_right.latitude) * 0.5;
				gdouble radius = R_EARTH * 0.5 * sqrt( pow((dLon * cos(mLat)), 2.0) + pow(dLat, 2.0) );
				place_req->boundary->circle.center.latitude = centerLat;
				place_req->boundary->circle.center.longitude = centerLon;
				place_req->boundary->circle.radius = radius;
			}
		}
	}

	*request_id = ++__request_id;
	calldata_place->reqID = __request_id;

	int ret = mapzen_search_place(place_req, __mapzen_place_search_cb, __request_id, (void *) calldata_place);

	return convert_mapzen_error_to_maps_error(ret);
}

EXPORT_API int maps_plugin_search_place_list(const maps_area_h boundary, const maps_place_filter_h filter, const maps_preference_h preference, maps_service_search_place_list_cb callback, void *user_data, int *request_id)
{

	if (!boundary || !filter || !callback || !request_id)
		return MAPS_ERROR_INVALID_PARAMETER;

	callback_info_place_list *calldata_place = (callback_info_place_list *)g_malloc0(sizeof(callback_info_place_list));
	if (calldata_place == NULL)
		MAPS_PRINT_ERROR_CODE_RETURN(MAPS_ERROR_OUT_OF_MEMORY);

	calldata_place->callback = callback;
	calldata_place->data = user_data;

	mapzen_search_req_s *place_req = (mapzen_search_req_s *)g_malloc0(sizeof(mapzen_search_req_s));
	if (place_req == NULL) {
		g_free(calldata_place);
		calldata_place = NULL;
		MAPS_PRINT_ERROR_CODE_RETURN(MAPS_ERROR_OUT_OF_MEMORY);
	}

	place_req->maps_key = g_strdup((gchar *) __provider_key);

	int max_result;
	maps_preference_get_max_results(preference, &max_result);

	if (max_result == 0)
		place_req->num_res = DEFAULT_NUM_RESULTS;
	else
		place_req->num_res = max_result;

	MAPS_LOGD("Place Result limit :: %d", place_req->num_res);

	char *categoryId = NULL;
	char *searchKeyword = NULL;

	maps_place_category_h category = NULL;
	maps_place_filter_get_category(filter, &category);
	maps_place_filter_get_keyword(filter, &searchKeyword);

	if (category) {
		maps_place_category_get_id(category, &categoryId);
		maps_place_category_destroy(category);
	}

	if (categoryId) {
		if (strlen(categoryId) == 0) {
			place_req->categories = NULL;
		} else {
			place_req->categories = g_strdup_printf("%s", categoryId);
		}
	}

	if (searchKeyword)
		place_req->search_string = g_strdup_printf("%s", searchKeyword);

	if (searchKeyword) {
		g_free(searchKeyword);
		searchKeyword = NULL;
	}

	if (categoryId) {
		g_free(categoryId);
		categoryId = NULL;
	}

	MAPS_LOGD(">>>>>>>> Place search categories :: %s <<<<<<<<<", place_req->categories);

	place_req->boundary = NULL;
	maps_area_s *bound = (maps_area_s *) boundary;

	if (bound->type != MAPS_AREA_NONE) {
		place_req->boundary = (mapzen_boundary_s *)g_malloc0(sizeof(mapzen_boundary_s));

		if (place_req->boundary != NULL) {
			if (bound->type == MAPS_AREA_CIRCLE) {
				place_req->boundary->type = MAPZEN_BOUNDARY_CIRCLE;
				place_req->boundary->circle.center.latitude = bound->circle.center.latitude;
				place_req->boundary->circle.center.longitude = bound->circle.center.longitude;
				place_req->boundary->circle.radius = bound->circle.radius;
			} else if (bound->type == MAPS_AREA_RECTANGLE) {
				place_req->boundary->type = MAPZEN_BOUNDARY_RECT;
				place_req->boundary->rect.top_left.latitude = bound->rect.top_left.latitude;
				place_req->boundary->rect.top_left.longitude = bound->rect.top_left.longitude;
				place_req->boundary->rect.bottom_right.latitude = bound->rect.bottom_right.latitude;
				place_req->boundary->rect.bottom_right.longitude = bound->rect.bottom_right.longitude;
			}
		}
	}

	*request_id = ++__request_id;
	calldata_place->reqID = __request_id;

	int ret = mapzen_search_place_list(place_req, __maps_service_search_place_list_cb, __request_id, (void *) calldata_place);

	return convert_mapzen_error_to_maps_error(ret);
}

EXPORT_API int maps_plugin_get_place_details(const char *uri, maps_service_get_place_details_cb callback, void *user_data, int *request_id)
{
	return  MAPS_ERROR_NOT_SUPPORTED;

    /* Following is not currently supported by mapzen plugin */
	/*if (!uri || !callback || !request_id)
		return MAPS_ERROR_INVALID_PARAMETER;

	callback_info_place_details *calldata_place = (callback_info_place_details *)g_malloc0(sizeof(callback_info_place_details));
	if (calldata_place == NULL)
		MAPS_PRINT_ERROR_CODE_RETURN(MAPS_ERROR_OUT_OF_MEMORY);

	calldata_place->callback = callback;
	calldata_place->data = user_data;

	mapzen_get_details_req_s *place_req = (mapzen_get_details_req_s *)g_malloc0(sizeof(mapzen_get_details_req_s));
	if (place_req == NULL) {
		g_free(calldata_place);
		calldata_place = NULL;
		MAPS_PRINT_ERROR_CODE_RETURN(MAPS_ERROR_OUT_OF_MEMORY);
	}

	place_req->maps_key = g_strdup((gchar *) __provider_key);
	place_req->uri = g_strdup_printf("%s", uri);

	*request_id = ++__request_id;
	calldata_place->reqID = __request_id;

	int ret = mapzen_get_place_details(place_req, __mapzen_place_get_details_cb, __request_id, (void *) calldata_place);

	return convert_mapzen_error_to_maps_error(ret); */
}

EXPORT_API int maps_plugin_search_place_by_address(const char *address, const maps_area_h boundary, const maps_place_filter_h filter, const maps_preference_h preference, maps_service_search_place_cb callback, void *user_data, int *request_id)
{
	if (!address || !boundary || !filter || !callback || !request_id)
		return MAPS_ERROR_INVALID_PARAMETER;

	callback_info_place *calldata_place = (callback_info_place *)g_malloc0(sizeof(callback_info_place));
	if (calldata_place == NULL)
		MAPS_PRINT_ERROR_CODE_RETURN(MAPS_ERROR_OUT_OF_MEMORY);

	calldata_place->callback = callback;
	calldata_place->data = user_data;

	mapzen_search_req_s *place_req = (mapzen_search_req_s *)g_malloc0(sizeof(mapzen_search_req_s));
	if (place_req == NULL) {
		g_free(calldata_place);
		calldata_place = NULL;
		MAPS_PRINT_ERROR_CODE_RETURN(MAPS_ERROR_OUT_OF_MEMORY);
	}

	place_req->maps_key = g_strdup((gchar *) __provider_key);

	int max_result;
	maps_preference_get_max_results(preference, &max_result);

	if (max_result == 0)
		place_req->num_res = DEFAULT_NUM_RESULTS;
	else
		place_req->num_res = max_result;

	MAPS_LOGD("Place Result limit :: %d", place_req->num_res);

	char *categoryId = NULL;

	maps_place_category_h category = NULL;
	maps_place_filter_get_category(filter, &category);

	if (category) {
		maps_place_category_get_id(category, &categoryId);
		maps_place_category_destroy(category);
	}

	if (categoryId) {
		place_req->categories = g_strdup_printf("%s", categoryId);
		g_free(categoryId);
		categoryId = NULL;
	}

	place_req->search_string = g_strdup_printf("%s", address);

	MAPS_LOGD(">>>>>>>> Place search categories :: %s <<<<<<<<<", place_req->categories);

	place_req->boundary = NULL;
	maps_area_s *bound = (maps_area_s *) boundary;

	if (bound->type != MAPS_AREA_NONE) {
		place_req->boundary = (mapzen_boundary_s *)g_malloc0(sizeof(mapzen_boundary_s));

		if (place_req->boundary != NULL) {
			if (bound->type == MAPS_AREA_CIRCLE) {
				place_req->boundary->type = MAPZEN_BOUNDARY_CIRCLE;
				place_req->boundary->circle.center.latitude = bound->circle.center.latitude;
				place_req->boundary->circle.center.longitude = bound->circle.center.longitude;
				place_req->boundary->circle.radius = bound->circle.radius;
			} else if (bound->type == MAPS_AREA_RECTANGLE) {
				place_req->boundary->type = MAPZEN_BOUNDARY_RECT;
				place_req->boundary->rect.top_left.latitude = bound->rect.top_left.latitude;
				place_req->boundary->rect.top_left.longitude = bound->rect.top_left.longitude;
				place_req->boundary->rect.bottom_right.latitude = bound->rect.bottom_right.latitude;
				place_req->boundary->rect.bottom_right.longitude = bound->rect.bottom_right.longitude;
			}
		}
	}

	*request_id = ++__request_id;
	calldata_place->reqID = __request_id;

	int ret = mapzen_search_place(place_req, __mapzen_place_search_cb, __request_id, (void *) calldata_place);

	return convert_mapzen_error_to_maps_error(ret);
}

EXPORT_API int maps_plugin_cancel_request(int request_id)
{
	MAPS_LOGD("Plugin_Cancel_Request...");
	if (request_id < 0)
		return MAPS_ERROR_INVALID_PARAMETER;

	int ret = mapzen_cancel_request(request_id);

	return convert_mapzen_error_to_maps_error(ret);
}

EXPORT_API int maps_plugin_request_user_consent(const char *provider, maps_service_request_user_consent_cb callback, void *user_data)
{
	MAPS_LOGD("maps_plugin_request_user_consent. Mapzen plugin always gives user consent.");
	callback(true, provider, user_data);
	return MAPS_ERROR_NONE;
}

EXPORT_API int maps_plugin_request_user_consent_with_handle(maps_service_request_user_consent_with_handle_cb callback, void *user_data)
{
	MAPS_LOGD("maps_plugin_request_user_consent_with_handle. Mapzen plugin always gives user consent.");
	callback(true, user_data);
	return MAPS_ERROR_NONE;
}

EXPORT_API int maps_plugin_get_user_consent(bool *consent)
{
	MAPS_LOGD("maps_plugin_get_user_consent. Mapzen plugin always gives user consent.");
	*consent = true;
	return MAPS_ERROR_NONE;
}